<?php
namespace WCStepFilter;

/**
 * Post-Type Entity Trait
 *
 * @version 1.0.0
 *
 * @property string $postType
 * @property string $namespace
 */
trait PostTypeTrait
{
    /** Init class handlers */
    public function initPostType()
    {
        add_action('init', [$this, 'registerPostType']);
        add_action('add_meta_boxes', [$this, 'addMetaBoxes']);
        add_action('save_post_' . self::$postType, [$this, 'savePostAction']);
        add_filter(self::$namespace . '_settings_model', [$this, 'settingsModelFilter']);
    }

    /** Register entity post type */
    abstract public function registerPostType();

    /**
     * Get settings model array
     *
     * @return array
     */
    abstract public static function getSettingsModel();

    /**
     * Settings model filter
     *
     * @param array $model
     *
     * @return array
     */
    public static function settingsModelFilter($model)
    {
        $class = strtolower(basename(str_replace('\\', '/', __CLASS__)));
        $option = (array) get_option('woocommerce_step_filter_settings_models', []);

        if (empty(array_filter($option)) || !isset($option[$class])) {
            return $model;
        }

        foreach (array_filter((array) $option[$class]) as $key => $setting) {
            $setting = (array) $setting;

            if (!isset($setting['version_from']) || !$setting['version_from']) {
                continue;
            }

            if (isset($setting['version_to']) && $setting['version_to']
                && version_compare(WC_STEP_FILTER_VERSION, $setting['version_to'], '>=')
            ) {
                continue;
            }

            if (version_compare(WC_STEP_FILTER_VERSION, $setting['version_from'], '>=')) {
                $model[$key] = $setting;
            }
        }

        return $model;
    }

    /**
     * Get post settings
     *
     * @param integer $id
     * @param array $args
     *
     * @return array
     */
    public static function getSettings($id, $args = [])
    {
        $defaults = ['public' => false];
        $args = array_merge($defaults, $args);
        $output = [];

        foreach (self::getSettingsModel() as $settingModelKey => $settingModel) {
            if ($args['public'] && (!isset($settingModel['public']) || !$settingModel['public'])) {
                continue;
            }

            $output[$settingModelKey] = self::getSetting($id, $settingModelKey);
        }

        return apply_filters(self::$namespace . '_settings', $output, $id, $args);
    }

    /**
     * Get one of post settings
     *
     * @param integer $id
     * @param string $setting
     * @param mixed $default
     *
     * @return string|float|boolean|array
     */
    public static function getSetting($id, $setting, $default = null)
    {
        static $cache = [];

        if (isset($cache[$id], $cache[$id][$setting])) {
            return apply_filters(self::$namespace . '_setting', $cache[$id][$setting], $id, $setting);
        }

        $model = self::getSettingsModel();

        if (!isset($model[$setting])) {
            $cache[$id][$setting] = $default;

            return apply_filters(self::$namespace . '_setting', $default, $id, $setting);
        }

        $value = null;

        if (isset($model[$setting]['key'])) {
            $value = get_post_meta($id, $model[$setting]['key']);
        }

        if (empty($value)) {
            if ($default) {
                $value = $default;
            } elseif (isset($model[$setting]['default'])) {
                $value = $model[$setting]['default'];
            }
        } elseif (isset($value[0])) {
            $value = $value[0];
        }

        $value = Utils::handleSettingType($value, $model[$setting]['type']);

        if (!isset($cache[$id])) {
            $cache[$id] = [];
        }

        $cache[$id][$setting] = $value;

        return apply_filters(self::$namespace . '_setting', $value, $id, $setting);
    }

    /** Add meta-boxes */
    public function addMetaBoxes()
    {
        add_meta_box('settings', L10N::r('Settings'), [$this, 'metaBoxCallback'], self::$postType, 'normal');
    }

    /**
     * Post meta-box view
     *
     * @param \WP_Post $post
     */
    public function metaBoxCallback($post)
    {
        ?>
        <table class="form-table wcsf-settings-table">
            <?php foreach (self::getSettingsModel() as $modelItemKey => $modelItem) { ?>
                <tr class="wcsf-settings-table-row<?php
                    echo isset($setting['separate']) && $modelItem['separate'] ? ' separate' : '';
                    ?>"
                    data-component="wcsf-setting" data-key="<?php echo esc_attr($modelItemKey); ?>">
                    <th scope="row">
                        <label for="<?php echo esc_attr($modelItem['key']); ?>"><?php
                            echo wp_kses_post($modelItem['label']);
                            ?></label>
                    </th>
                    <td><?php
                        Admin::settingFieldView(
                            $modelItem,
                            ['values' => [$modelItem['key'] => self::getSetting($post->ID, $modelItemKey)]]
                        );
                        ?></td>
                </tr>
            <?php } ?>
        </table>
        <?php
    }

    /**
     * Save meta values
     *
     * @param integer $postId
     */
    public function savePostAction($postId)
    {
        if ((defined('DOING_AUTOSAVE') && DOING_AUTOSAVE)
            || !current_user_can('edit_post', $postId)
            || (isset($_REQUEST['action']) && $_REQUEST['action'] == 'inline-save')
        ) {
            return;
        }

        foreach (self::getSettingsModel() as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_post_meta($postId, $setting['key'], $_POST[$setting['key']]);
        }
    }

    /**
     * Returns array of ids and titles of posts
     *
     * @param array $args
     *
     * @return array
     */
    public static function getPostsIds($args = [])
    {
        static $cache;

        $saveCache = false;

        if (empty($args) && $cache) {
            return $cache;
        }

        if (empty($args)) {
            $saveCache = true;
        }

        $defaults = [
            'post_type' => self::$postType,
            'numberposts' => -1,
            'post_status' => 'publish'
        ];

        $args = array_replace($defaults, $args);
        $output = [];

        foreach (get_posts($args) as $post) {
            $output[$post->ID] = '#' . $post->ID . ': ' . $post->post_title;
        }

        if ($saveCache) {
            $cache = $output;
        }

        return $output;
    }
}
