<?php
namespace WCStepFilter;

/**
 * Taxonomy Entity Trait
 *
 * @version 1.0.0
 *
 * @property string $taxonomy
 * @property string $namespace
 */
trait TaxonomyTrait
{
    /** Init class handlers */
    public function initTaxonomy()
    {
        add_action('init', [$this, 'registerTaxonomy']);
        add_action(self::$taxonomy . '_edit_form_fields', [$this, 'outputCategoryFields'], 10, 2);
        add_action('edited_' . self::$taxonomy, [$this, 'saveCategoryFields'], 10, 2);
        add_action('parent_file', [$this, 'categoriesMenuItemHighlight']);
    }

    /** Register entity taxonomy */
    abstract public function registerTaxonomy();

    /**
     * Highlight categories menu item if it is active
     *
     * @param string $parentFile
     *
     * @return string
     */
    public function categoriesMenuItemHighlight($parentFile)
    {
        global $current_screen;

        $taxonomy = $current_screen->taxonomy;

        if ($taxonomy == self::$taxonomy) {
            $parentFile = 'edit.php?post_type=' . Filter::$postType;
        }

        return $parentFile;
    }

    /**
     * Get category settings model array
     *
     * @return array
     */
    abstract public static function getCategorySettingsModel();

    /**
     * Add meta-box to category edit page
     *
     * @param \WP_Term $term
     */
    public function outputCategoryFields($term)
    {
        foreach (self::getCategorySettingsModel() as $key => $setting) {
            ?>
            <tr class="wcsf-settings-table-row">
                <th scope="row" class="wcsf-settings-table-row-name">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                        echo wp_kses_post($setting['label']);
                        ?></label>
                </th>
                <td class="wcsf-settings-table-row-value"><?php
                    Admin::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => self::getCategorySetting($term->term_id, $key)]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save category meta fields
     *
     * @param integer $termId
     */
    public function saveCategoryFields($termId)
    {
        if (!$termId) {
            return;
        }

        foreach (self::getCategorySettingsModel() as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_term_meta($termId, $setting['key'], $_POST[$setting['key']]);
        }
    }

    /**
     * Get category setting
     *
     * @param integer $id
     * @param string $setting
     * @param mixed $default
     *
     * @return string|float|bool|array
     */
    public static function getCategorySetting($id, $setting, $default = null)
    {
        static $cache = [];

        if (isset($cache[$id], $cache[$id][$setting])) {
            return apply_filters(self::$namespace . '_category_setting', $cache[$id][$setting], $id, $setting);
        }

        $model = self::getCategorySettingsModel();

        if (!isset($model[$setting])) {
            $cache[$id][$setting] = $default;

            return apply_filters(self::$namespace . '_category_setting', $default, $id, $setting);
        }

        $value = null;

        if (isset($model[$setting]['key'])) {
            $value = get_term_meta($id, $model[$setting]['key']);
        }

        if (empty($value)) {
            if ($default) {
                $value = $default;
            } elseif (isset($model[$setting]['default'])) {
                $value = $model[$setting]['default'];
            }
        } elseif (isset($value[0])) {
            $value = $value[0];
        }

        $value = Utils::handleSettingType($value, $model[$setting]['type']);

        if (!isset($cache[$id])) {
            $cache[$id] = [];
        }

        $cache[$id][$setting] = $value;

        return apply_filters(self::$namespace . '_category_setting', $value, $id, $setting);
    }

    /**
     * Get list of categories ids and names
     *
     * @return array
     */
    public static function getCategoriesIds()
    {
        static $cache;

        if ($cache) {
            return apply_filters(self::$namespace . '_categories_ids', $cache);
        }

        $cache = $output = get_terms([
            'taxonomy' => self::$taxonomy,
            'fields' => 'id=>name',
            'hide_empty' => false,
        ]);

        return apply_filters(self::$namespace . '_categories_ids', $output);
    }
}
