<?php
namespace WCStepFilter;

/**
 * Core Class
 *
 * @class Core
 * @version 10.0.0
 */
class Core
{
    // <editor-fold desc="Properties">
    /**
     * Self instance variable
     * @var Core The single instance of the class
     */
    protected static $instance = null;

    /**
     * Is WooCommerce active
     * @var bool
     */
    public static $wcIsActive = false;

    /**
     * Admin part instance variable
     * @var Admin
     */
    protected $admin = null;

    /**
     * API class instance
     * @var object
     */
    public $api;

    /**
     * Storage class instance
     * @var object
     */
    public $storage;

    /**
     * Filter class instance
     * @var object
     */
    public $filter;

    /**
     * Question class instance
     * @var object
     */
    public $question;

    /**
     * Inquiry class instance
     * @var object
     */
    public $inquiry;

    /**
     * PDF class instance
     * @var object
     */
    public $pdf;

    /**
     * Template class instance variable
     * @var Template
     */
    public $template = null;

    /**
     * Product class instance variable
     * @var Product
     */
    public $product = null;
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Class Constructor */
    public function __construct()
    {
        $this->loadClasses();

        // actions
        add_action('plugins_loaded', [$this, 'pluginsLoadedAction']);
        add_action('woocommerce_init', [$this, 'wcInitAction'], 1);
        add_action('wp_loaded', [$this, 'requests']);
        add_action('plugins_loaded', [$this, 'loadTextDomain']);
        add_action('woocommerce_before_shop_loop', [$this, 'beforeShopProductsAction']);
        add_action('woocommerce_after_shop_loop', [$this, 'afterShopProductsAction']);
        add_action('woocommerce_no_products_found', [$this, 'beforeShopNoProductsAction'], 5);
        add_action('woocommerce_no_products_found', [$this, 'afterShopNoProductsAction'], 15);
        add_action('wcsf_before_output', [$this, 'beforeOutputAction']);
        add_action('wcsf_after_output', [$this, 'afterOutputAction']);

        if (isset($_GET['wcsf-results'])) {
            add_action('woocommerce_product_query', [$this, 'modifyProductsQuery'], 15);
            add_filter('woocommerce_layered_nav_link', [$this, 'layeredNavLinkFilter']);
            add_filter('woocommerce_widget_get_current_page_url', [$this, 'layeredNavLinkFilter']);

            // filtering WOOF plugin args
            add_filter('woof_products_query', [$this, 'modifyWoofQuery'], 15);
        }

        // scripts and styles
        add_action('wcsf_before', [$this, 'enqueueAssets']);
    }

    /**
     * Get single class instance
     *
     * @static
     * @return Core
     */
    public static function instance()
    {
        if (self::$instance === null) {
            self::$instance = new self();
        }

        return self::$instance;
    }

    /** Handles then plugins are loaded */
    public function pluginsLoadedAction()
    {
        self::$wcIsActive = class_exists('\WooCommerce');
    }

    /** Runs after woocommerce is loaded */
    public function wcInitAction()
    {
        add_action('wcsf_shortcode', [$this, 'shortCodeAction']);

        Utils::startSession();

        // start WC session variable if needed
        if (self::getSetting('store_session_in_db') && function_exists('WC')) {
            if (method_exists(WC(), 'initialize_session')) {
                WC()->initialize_session();
            }

            if (method_exists(WC()->session, 'set_customer_session_cookie')) {
                WC()->session->set_customer_session_cookie(true);
            }
        }

        // if is admin page
        if (is_admin()) {
            if (!class_exists('\\WCStepFilter\\Admin')) {
                require_once(__DIR__ . DIRECTORY_SEPARATOR . 'Admin.php');
            }

            $this->admin = new Admin();
        }
    }

    /**
     * Short-code call action
     *
     * @param array $attributes
     */
    public function shortCodeAction($attributes)
    {
        if (!isset($attributes['id']) || empty($attributes['id'])) {
            return;
        }

        $id = (int) $attributes['id'];

        if ($this->filter->getSetting($id, 'reset_filter_on_showing') && !isset($_GET['wcsf-submit'])
            && (!isset($_GET['wcsf-results']) || empty($_GET['wcsf-results']))
        ) {
            $this->filter->resetState($id);
        }

        if ($this->filter->getSetting($id, 'results_url') == '#'
            && isset($_GET['wcsf-value'], $_GET['wcsf-results']) && !empty($_GET['wcsf-value'])
            && $_GET['wcsf-results'] == $id
        ) {
            try {
                // set filter's value from get parameter
                $this->filter->setValue($id, $_GET['wcsf-value']);
            } catch (\Exception $exception) {
                $this->filter->addNotice(
                    $exception->getCode() ? $exception->getCode() : $this->filter->getActiveStepId($id),
                    ['message' => $exception->getMessage()]
                );
            }
        }
    }

    /**
     * Handles on output call
     *
     * @param array $args
     */
    public function beforeOutputAction($args)
    {
        if (isset($args['id']) && $args['id']) {
            $this->filter->setCurrentId((int) $args['id']);
        }
    }

    /** Handles after output call */
    public function afterOutputAction()
    {
        $this->filter->setCurrentId(null);
    }

    /** Load text domain */
    public function loadTextDomain()
    {
        load_plugin_textdomain(
            'woocommerce-step-filter',
            false,
            basename(WC_STEP_FILTER_PLUGIN_PATH) . '/languages/'
        );
    }

    /** Load main classes */
    public function loadClasses()
    {
        // include base slave classes
        $requiredClasses = [
            'Interfaces/PostTypeInterface',
            'Interfaces/TaxonomyInterface',
            'Abstract/PostTypeTrait',
            'Abstract/TaxonomyTrait',
            'Product/Attribute',
            'Product/Category',
            'Product/Tag',
            'Product',
            'L10N',
            'API',
            'Utils',
            'Styles',
            'Template',
            'Storage',
            'Filter',
            'Question',
            'Inquiry',
            'PDF'
        ];

        foreach ($requiredClasses as $requiredClass) {
            if (!class_exists('\\WCStepFilter\\' . $requiredClass)
                && file_exists(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php')
            ) {
                require_once(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php');
            }
        }

        $this->api = new API();
        $this->storage = new Storage();
        $this->inquiry = new Inquiry();
        $this->filter = new Filter();
        $this->question = new Question();
        $this->pdf = new PDF();
        $this->template = new Template();
        $this->product = new Product();
    }

    /** Styles and scripts enqueue */
    public function enqueueAssets()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';
        $stylesFolder = WC_STEP_FILTER_DEBUG ? 'scss' : 'css';
        $scriptsIncludingType = self::getSetting('scripts_including_type');
        $includedScripts = self::getSetting('included_scripts');
        $stylesIncludingType = self::getSetting('styles_including_type');

        wp_enqueue_script('jquery');

        if ($scriptsIncludingType == 'single' && !WC_STEP_FILTER_DEBUG) {
            wp_enqueue_script(
                'woocommerce-step-filter-scripts',
                WC_STEP_FILTER_PLUGIN_URL . 'assets/front/js/scripts.min.js',
                ['jquery'],
                WC_STEP_FILTER_VERSION,
                true
            );
        } elseif ($scriptsIncludingType == 'multiple' || WC_STEP_FILTER_DEBUG) {
            if (in_array('wNumb', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-wNumb',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/wNumb$suffix.js",
                    [],
                    '1.1.0',
                    true
                );
            }

            if (in_array('nouislider', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-nouislider',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider$suffix.js",
                    [],
                    '14.6.3',
                    true
                );
            }

            if (in_array('nouislider-launcher', $includedScripts)) {
                $dependency = ['jquery'];

                if (in_array('wNumb', $includedScripts)) {
                    $dependency[] = 'woocommerce-step-filter-wNumb';
                }

                if (in_array('nouislider', $includedScripts)) {
                    $dependency[] = 'woocommerce-step-filter-nouislider';
                }

                wp_enqueue_script(
                    'woocommerce-step-filter-nouislider-launcher',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider-launcher$suffix.js",
                    $dependency,
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('app', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-app',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/wcsf.jquery$suffix.js",
                    ['jquery'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('hooks', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-hooks',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/hooks$suffix.js",
                    ['woocommerce-step-filter-app'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }
        }

        switch ($stylesIncludingType) {
            case 'custom': {
                wp_enqueue_style(
                    'woocommerce-step-filter-full-custom',
                    WC_STEP_FILTER_UPLOADS_URL . 'app-full-custom.css',
                    [],
                    get_option('woocommerce_step_filter_styles_compiled_time', '')
                );

                break;
            }

            case 'full': {
                wp_enqueue_style(
                    'woocommerce-step-filter-full',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app-full$suffix.css",
                    [],
                    WC_STEP_FILTER_VERSION
                );

                break;
            }

            case 'basic': {
                wp_enqueue_style(
                    'woocommerce-step-filter',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app$suffix.css",
                    [],
                    WC_STEP_FILTER_VERSION
                );
            }
        }
    }
    // </editor-fold>

    // <editor-fold desc="HTML">
    /** Output shop extra HTML before products */
    public function beforeShopProductsAction()
    {
        if (isset($_GET['wcsf-results']) && !empty($_GET['wcsf-results'])) {
            $id = (int) $_GET['wcsf-results'];

            Template::html('shop/value', ['id' => $id], ['once' => true]);
            Template::html('shop/controls/index', ['id' => $id], ['once' => true]);
            Template::html('shop/before-products-description', ['id' => $id], ['once' => true]);
        }
    }

    /** Output shop extra HTML after products */
    public function afterShopProductsAction()
    {
        if (isset($_GET['wcsf-results']) && !empty($_GET['wcsf-results'])) {
            $id = (int) $_GET['wcsf-results'];

            Template::html('shop/after-products-description', ['id' => $id], ['once' => true]);
        }
    }

    /** Output shop extra HTML before no products */
    public function beforeShopNoProductsAction()
    {
        if (isset($_GET['wcsf-results']) && !empty($_GET['wcsf-results'])) {
            $id = (int) $_GET['wcsf-results'];

            Template::html('shop/value', ['id' => $id], ['once' => true]);
            Template::html('shop/controls/index', ['id' => $id], ['once' => true]);
            Template::html('shop/before-no-products-description', ['id' => $id], ['once' => true]);
        }
    }

    /** Output shop extra HTML after no products */
    public function afterShopNoProductsAction()
    {
        if (isset($_GET['wcsf-results']) && !empty($_GET['wcsf-results'])) {
            $id = (int) $_GET['wcsf-results'];

            Template::html('shop/after-no-products-description', ['id' => $id], ['once' => true]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Query">
    /** Add request actions */
    public function requests()
    {
        if (is_admin()) {
            return;
        }

        // get queries
        // reset the filter
        if (!isset($_POST['results']) && isset($_GET['wcsf-reset-state'])) {
            $this->filter->resetState((int) $_GET['wcsf-reset-state']);
        }

        // reset one value
        if (!isset($_POST['results']) && isset($_GET['wcsf-reset-value'], $_GET['wcsf-results'])) {
            $this->filter->resetValue((int) $_GET['wcsf-results'], (int) $_GET['wcsf-reset-value']);
        }

        // post queries
        if (!isset($_POST['wcsf'])) {
            return;
        }

        $id = isset($_POST['id']) ? (int) $_POST['id'] : false;

        // skip step
        if (isset($_POST['skip'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['skip']));

            return;
        }

        // back step
        if (isset($_POST['back'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['back']));

            return;
        }

        // save value
        if (isset($_POST['value']) && !empty($_POST['value'])) {
            try {
                $this->filter->setValue($id, (array) $_POST['value']);
            } catch (\Exception $exception) {
                $this->filter->addNotice(
                    $exception->getCode() ? $exception->getCode() : $this->filter->getActiveStepId($id),
                    ['message' => $exception->getMessage()]
                );
            }
        }

        // change step
        if (isset($_POST['stepId'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['stepId']));
        }

        // reset filter
        if (isset($_POST['reset'])) {
            $this->filter->resetState($id);
        }

        // show results page
        if (isset($_POST['results'])) {
            $redirectLink = $this->filter->getResultsUrl($id);
            $redirectLink = apply_filters('wcsf_results_redirect_link', $redirectLink, $id);
            $redirectCode = apply_filters('wcsf_results_redirect_code', 302, $id, $redirectLink);

            do_action('wcsf_before_results_redirect', $id, $redirectLink);

            wp_redirect($redirectLink, $redirectCode);

            // js version of redirect
            exit("<script>document.location = '$redirectLink';</script>");
        }
    }

    /**
     * Execute question formula by values
     *
     * @param object $question - question object
     * @param array $values - question values
     *
     * @return float
     *
     * @throws \Exception
     */
    public static function executeFormula($question, $values)
    {
        if (!class_exists('\Matex\Evaluator')) {
            require_once __DIR__ . '/../vendor/matex/Evaluator.php';
            require_once __DIR__ . '/../vendor/matex/Math.php';
        }

        $evaluator = new \Matex\Evaluator();
        $evaluator->variables = [];
        $evaluator->functions = \Matex\Math::functions;

        foreach ($question->formulaVariables as $variable) {
            $evaluator->variables[$variable['name']] = (float) $variable['value'];
        }

        foreach ($values as $fieldName => $fieldValue) {
            foreach ($question->formulaValueAliases as $alias) {
                if ($alias['name'] == $fieldName && $alias['value'] == $fieldValue) {
                    $fieldValue = $alias['alias'];

                    continue;
                }
            }

            foreach ($question->formulaNameAliases as $alias) {
                if ($alias['name'] == $fieldName) {
                    $fieldName = $alias['alias'];

                    continue;
                }
            }

            $evaluator->variables[$fieldName] = (float) $fieldValue;
        }

        return $evaluator->execute($question->formula);
    }

    /**
     * Get query args by the filter id and value
     *
     * @param integer $id
     * @param array $filterValue
     * @param array $extraQueryArgs - additional WP_Query arguments
     *
     * @return array
     */
    public static function getQueryArgsByFilter($id, $filterValue, $extraQueryArgs = [])
    {
        $taxQuery = [];
        $postIn = [];
        $postNotIn = [];
        $metaQuery = [];
        $extraQueryArgsSetting = Filter::getSetting($id, 'extra_query_args');
        $multiChoiceOperators = apply_filters(
            'wsf_multi_choice_operators',
            [
                -1 => 'NOT IN',
                1 => 'IN',
                2 => 'AND'
            ]
        );

        if (!empty($extraQueryArgsSetting)) {
            foreach ($extraQueryArgsSetting as $extraArg) {
                if (!is_array($extraArg) || empty($extraArg)
                    || ($extraArg['source'] == 'product' && empty($extraArg['product']))
                    || ($extraArg['source'] == 'category' && empty($extraArg['category']))
                    || ($extraArg['source'] == 'attribute'
                        && (empty($extraArg['attribute_name']) || empty($extraArg['attribute_value'])))
                ) {
                    continue;
                }

                switch ($extraArg['source']) {
                    case 'product': {
                        if ($extraArg['condition'] == '=') {
                            $postIn[] = (int) $extraArg['product'];
                        } else {
                            $postNotIn[] = (int) $extraArg['product'];
                        }

                        break;
                    }

                    case 'taxonomy': {
                        $taxQuery[] = [
                            'taxonomy' => $extraArg['taxonomy'],
                            'field' => 'id',
                            'terms' => (int) $extraArg['term'],
                            'operator' => (!isset($extraArg['condition']) || $extraArg['condition'] == '=')
                                ? 'IN'
                                : 'NOT IN'
                        ];

                        break;
                    }

                    case 'category': {
                        $taxQuery[] = [
                            'taxonomy' => 'product_cat',
                            'field' => 'id',
                            'terms' => (int) $extraArg['category'],
                            'operator' => (!isset($extraArg['condition']) || $extraArg['condition'] == '=')
                                ? 'IN'
                                : 'NOT IN'
                        ];

                        break;
                    }

                    case 'attribute': {
                        $taxQuery[] = [
                            'taxonomy' => 'pa_' . esc_sql($extraArg['attribute_name']),
                            'field' => 'id',
                            'terms' => (int) $extraArg['attribute_value'],
                            'operator' => (!isset($extraArg['condition']) || $extraArg['condition'] == '=')
                                ? 'IN'
                                : 'NOT IN'
                        ];
                    }
                }
            }
        }

        // loop for the each filter value
        foreach ((array) $filterValue as $questionId => $questionValue) {
            $question = Question::get($questionId);

            if (!is_object($question) || !property_exists($question, 'viewType') || $question->ignoreFiltering) {
                continue;
            }

            $viewType = $question->viewType;
            $filterType = $question->filterType;
            $relation = isset($question->relation) ? $question->relation : 'or';
            $formulaValue = 0;

            if ($viewType == 'formula') {
                try {
                    $formulaValue = self::executeFormula($question, $questionValue);
                } catch (\Exception $exception) {
                    echo '<p class="wcsf-question-formula-error" role="alert"><strong>'
                        . L10N::r('Formula execute error:')
                        . " <em>{$exception->getMessage()}</em></strong></p>";

                    continue;
                }
            }

            if (in_array($filterType, ['tag', 'category', 'attribute'])) {
                // attribute, category or tag
                if ($filterType == 'attribute') {
                    // filtering by attribute
                    if (!isset($question->attribute)) {
                        continue;
                    }

                    $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                } else {
                    // filtering by tag or category
                    $taxonomy = $filterType == 'tag' ? 'product_tag' : 'product_cat';
                }

                if (strpos($viewType, 'number') !== false) {
                    // numeric field type
                    $filterBy = [];
                    $terms = get_terms(['taxonomy' => $taxonomy]);
                    $valueSource = Question::getSetting($questionId, 'attribute_value_source');

                    foreach ($terms as $term) {
                        switch ($valueSource) {
                            default:
                            case 'name':
                                $termValue = (float) $term->name;
                                break;

                            case 'slug':
                                $termValue = (float) $term->slug;
                        }

                        $issetFrom = isset($questionValue['from']) && !empty($questionValue['from']);
                        $from = $issetFrom && $questionValue['from'] <= $termValue;
                        $issetTo = isset($questionValue['to']) && !empty($questionValue['to']);
                        $to = $issetTo && $questionValue['to'] >= $termValue;

                        // check by numeric with the first, the second and both values
                        if (($viewType == 'number' && !empty($questionValue))
                            || ($viewType == 'number-from' && $from)
                            || ($viewType == 'number-to' && $to)
                            || ($viewType == 'number-between'
                                && ($from && $to || !$issetTo && $from || !$issetFrom && $to))
                        ) {
                            $filterBy[] = $term->term_id;
                        }
                    }

                    if (!empty($filterBy)) {
                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => $filterBy,
                            'operator' => 'IN'
                        ];
                    }
                } else {
                    if (filter_var(get_option('woocommerce_hide_out_of_stock_items'), FILTER_VALIDATE_BOOLEAN)
                        && $filterType == 'attribute'
                        && $viewType != 'formula'
                    ) {
                        // skip products with out-of-stock variations by requested attributes
                        global $wpdb;

                        $visibilityTerms = wc_get_product_visibility_term_ids();
                        $include = array_filter($questionValue);
                        $productsIds = [];

                        if (strpos($viewType, 'multi-choice') !== false) {
                            $include = [];

                            foreach (array_filter($questionValue) as $termId => $choiceValue) {
                                if ($choiceValue > 0) {
                                    $include[] = $termId;
                                }
                            }
                        }

                        $terms = get_terms([
                            'taxonomy' => $taxonomy,
                            'include' => $include,
                            'fields' => 'id=>slug'
                        ]);

                        // using of an SQL query avoids cross wp_queries problems
                        $queryString =
                            "SELECT $wpdb->posts.post_parent, $wpdb->term_relationships.term_taxonomy_id "
                            . "FROM $wpdb->posts "
                            . "LEFT JOIN $wpdb->term_relationships "
                            . "ON ($wpdb->posts.ID = $wpdb->term_relationships.object_id) "
                            . "INNER JOIN $wpdb->postmeta ON ($wpdb->posts.ID = $wpdb->postmeta.post_id) "
                            . "WHERE $wpdb->posts.post_type = 'product_variation' "
                            . "AND $wpdb->posts.post_status IN ('publish', 'private') ";

                        if (!empty($terms)) {
                            $queryString .=
                                "AND ($wpdb->postmeta.meta_key = 'attribute_{$taxonomy}' "
                                . "AND $wpdb->postmeta.meta_value IN ("
                                . implode(',', array_map(
                                    function ($value) {
                                        return "'" . esc_sql($value) . "'";
                                    },
                                    $terms
                                )) . '))';
                        }

                        foreach ($wpdb->get_results($queryString) as $result) {
                            $productsIds[$result->post_parent][] = (int) $result->term_taxonomy_id;
                        }

                        foreach ($productsIds as $id => $variations) {
                            $counts = array_count_values($variations);

                            if (isset($counts[$visibilityTerms['outofstock']])
                                && $counts[$visibilityTerms['outofstock']] == count($variations)
                            ) {
                                $postNotIn[] = $id;
                            }
                        }
                    }

                    if ($viewType == 'formula') {
                        // formula field type
                        $filterBy = [];
                        $terms = get_terms(['taxonomy' => $taxonomy]);
                        $valueSource = Question::getSetting($questionId, 'attribute_value_source');

                        foreach ($terms as $term) {
                            switch ($valueSource) {
                                default:
                                case 'name':
                                    $termValue = (float) $term->name;
                                    break;

                                case 'slug':
                                    $termValue = (float) $term->slug;
                            }

                            $met = false;

                            switch ($question->formulaOperator) {
                                case '=':
                                    $met = $termValue == $formulaValue;
                                    break;

                                case '!=':
                                    $met = $termValue != $formulaValue;
                                    break;

                                case '<':
                                    $met = $termValue < (float) $formulaValue;
                                    break;

                                case '<=':
                                    $met = $termValue <= (float) $formulaValue;
                                    break;

                                case '>':
                                    $met = $termValue > (float) $formulaValue;
                                    break;

                                case '>=':
                                    $met = $termValue >= (float) $formulaValue;
                            }

                            if ($met) {
                                $filterBy[] = $term->term_id;
                            }
                        }

                        if (!empty($filterBy)) {
                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $filterBy,
                                'operator' => 'IN'
                            ];
                        }
                    } elseif (strpos($viewType, 'multi-choice') !== false) {
                        $in = [];

                        foreach (array_filter($questionValue) as $termId => $choiceValue) {
                            if ((int) $choiceValue == 0 || !isset($multiChoiceOperators[(int) $choiceValue])) {
                                continue;
                            } elseif ((int) $choiceValue == 1) {
                                // get all IN values in the array
                                $in[] = $termId;

                                continue;
                            }

                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $termId,
                                'operator' => $multiChoiceOperators[(int) $choiceValue]
                            ];
                        }

                        if (!empty($in)) {
                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $in,
                                'operator' => $multiChoiceOperators[1]
                            ];
                        }
                    } elseif (!empty(array_filter($questionValue))) {
                        // skip "doesn't matter" and not-defined value and filter by other terms
                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => array_filter($questionValue),
                            'operator' => $relation == 'or' ? 'IN' : 'AND'
                        ];
                    }
                }
            } elseif ($filterType == 'manual') {
                $manualTaxQuery = ['relation' => $relation];

                // filtering by a manual value
                foreach ($questionValue as $questionValueItem) {
                    if (!isset($question->manualValues[$questionValueItem])) {
                        continue;
                    }

                    $filterBy = $question->manualValues[$questionValueItem]['filter-by'];
                    $items = isset($question->manualValues[$questionValueItem]['items'])
                        ? $question->manualValues[$questionValueItem]['items']
                        : [];

                    $items = !empty($items) ? $items : [-1]; // if nothing is selected

                    if ($filterBy == 'any') {
                        continue;
                    } elseif ($filterBy == 'product') {
                        if (!isset($question->manualValues[$questionValueItem]['intersect'])
                            || (isset($question->manualValues[$questionValueItem]['intersect'])
                            && $question->manualValues[$questionValueItem]['intersect'])
                        ) {
                            $postIn = !empty($postIn) ? array_intersect($postIn, $items) : $items;
                        } else {
                            $postIn = array_unique(array_merge($postIn, $items));
                        }
                    } elseif ($filterBy == 'term') {
                        $termQueries = [];
                        $index = 0;

                        foreach ($items as $item) {
                            if (empty($item['source']) || empty($item['terms'])) {
                                continue;
                            }

                            if (isset($item['outer_relation']) && $item['outer_relation'] == 'and'
                                && end($items) != $item
                            ) {
                                $termQueries[$index]['relation'] = 'AND';
                                $termQueries[$index][] = [
                                    'taxonomy' => $item['source'],
                                    'field' => 'id',
                                    'terms' => array_filter((array) $item['terms'], function ($term) {
                                        return is_numeric($term);
                                    }),
                                    'operator' => isset($item['inner_relation']) && $item['inner_relation'] == 'or'
                                        ? 'IN'
                                        : 'AND'
                                ];
                            } else {
                                $termQueries[$index++][] = [
                                    'taxonomy' => $item['source'],
                                    'field' => 'id',
                                    'terms' => array_filter((array) $item['terms'], function ($term) {
                                        return is_numeric($term);
                                    }),
                                    'operator' => isset($item['inner_relation']) && $item['inner_relation'] == 'or'
                                        ? 'IN'
                                        : 'AND'
                                ];
                            }
                        }

                        $termQueries['relation'] = 'or';
                        $manualTaxQuery[] = $termQueries;
                    }
                }

                $taxQuery[] = $manualTaxQuery;
            } elseif (in_array($filterType, ['price', 'meta'])) {
                // filtering by price or meta
                $metaKey = $filterType == 'price' ? '_price' : $question->metaKey;
                $metaValue = null;
                $compare = '=';
                $type = 'CHAR';

                if ($viewType == 'formula') {
                    $type = 'NUMERIC';
                    $compare = $question->formulaOperator;
                    $metaValue = $formulaValue;
                } elseif (strpos($viewType, 'number') !== false) {
                    $type = 'NUMERIC';

                    switch ($viewType) {
                        case 'number':
                            $compare = '=';
                            $metaValue = !empty($questionValue) ? reset($questionValue) : null;

                            break;
                        case 'number-from':
                            $compare = '>=';

                            if (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $metaValue = $questionValue['from'];
                            }

                            break;
                        case 'number-to':
                            $compare = '<=';

                            if (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $metaValue = $questionValue['to'];
                            }

                            break;
                        case 'number-between':
                            if (isset($questionValue['from']) && !empty($questionValue['from'])
                                && isset($questionValue['to'])
                                && !empty($questionValue['to'])
                            ) {
                                $compare = 'BETWEEN';
                                $metaValue = array_values($questionValue);
                            } elseif (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $compare = '>=';
                                $metaValue = $questionValue['from'];
                            } elseif (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $compare = '<=';
                                $metaValue = $questionValue['to'];
                            }
                    }
                } elseif (strpos($viewType, 'multi-choice') !== false) {
                    $in = [];

                    foreach (array_filter($questionValue) as $metaValue => $choiceValue) {
                        if ($choiceValue == 1) {
                            // get all IN values in the array
                            $in[] = $metaValue;

                            continue;
                        }

                        if (!empty($metaValue)) {
                            $metaQuery[] = [
                                'key' => $metaKey,
                                'value' => $metaValue,
                                'compare' => $multiChoiceOperators[$choiceValue],
                                'type' => $type
                            ];
                        }
                    }

                    if (!empty($in)) {
                        $metaQuery[] = [
                            'key' => $metaKey,
                            'value' => $in,
                            'compare' => $multiChoiceOperators[1],
                            'type' => $type
                        ];
                    }

                    $metaValue = null;
                } else {
                    $compare = 'IN';
                    $metaValue = $questionValue;
                }

                if (!is_null($metaValue) && !empty($metaValue)) {
                    $metaQuery[] = [
                        'key' => $metaKey,
                        'value' => $metaValue,
                        'compare' => $compare,
                        'type' => $type
                    ];
                }
            }
        }

        return apply_filters(
            'wcsf_query_args_by_filter',
            [
                'tax_query' => $taxQuery,
                'meta_query' => $metaQuery,
                'post__in' => $postIn,
                'post__not_in' => $postNotIn
            ],
            $id,
            $filterValue,
            $extraQueryArgs
        );
    }

    /**
     * Filter the products on the shop page
     *
     * @param \WP_Query $query
     */
    public function modifyProductsQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return;
        }

        do_action('wcsf_before_modify_request', $query);

        $id = (int) $_GET['wcsf-results'];
        $value = (array) $_GET['wcsf-value'];
        $args = $this->getQueryArgsByFilter($id, $value);

        if (method_exists($query, 'set')) {
            if (!empty($args['tax_query'])) {
                $query->set('tax_query', array_merge($query->get('tax_query'), $args['tax_query']));
            }

            if (!empty($args['post__in'])) {
                $query->set('post__in', array_merge($query->get('post__in'), $args['post__in']));
            }

            if (!empty($args['post__not_in'])) {
                $query->set('post__not_in', array_merge($query->get('post__not_in'), $args['post__not_in']));
            }

            if (!empty($args['meta_query'])) {
                $query->set('meta_query', array_merge($query->get('meta_query'), $args['meta_query']));
            }

            $maxResults = $this->filter->getSetting($id, 'max_results_number');
            $onlyResultRedirect = $this->filter->getSetting($id, 'redirect_to_only_result_page');
            $noResultsRedirect = $this->filter->getSetting($id, 'empty_results_url');

            if ($maxResults || $onlyResultRedirect || $noResultsRedirect) {
                add_filter(
                    'found_posts',
                    function ($found, $instance) use ($id, $maxResults, $onlyResultRedirect, $noResultsRedirect) {
                        if ($instance->query_vars['post_type'] != 'product') {
                            return $found;
                        }

                        // override products number in the query
                        if ($maxResults && property_exists($instance, 'posts') && is_array($instance->posts)) {
                            $limit = $maxResults;
                            $instance->post_count = $maxResults;
                            $instance->found_posts = $maxResults;

                            if ($instance->query_vars['posts_per_page'] < $maxResults) {
                                $limit = abs($instance->query_vars['paged'] * $instance->query_vars['posts_per_page']
                                    - $maxResults - $instance->query_vars['posts_per_page']);
                            }

                            $instance->posts = array_slice($instance->posts, 0, $limit);
                            $found = min($found, $maxResults);
                        }

                        if (($onlyResultRedirect || $noResultsRedirect) && in_array($found, [0, 1])) {
                            $redirectLink = null;

                            if ($onlyResultRedirect && (int) $found == 1) {
                                $redirectLink = get_permalink(reset($instance->posts));
                            } elseif ($noResultsRedirect && (int) $found == 0) {
                                $redirectLink = $noResultsRedirect;

                                // if URL is absolute
                                if (strpos($redirectLink, $_SERVER['HTTP_HOST']) === false) {
                                    $redirectLink = home_url() . '/' . $redirectLink;
                                }
                            }

                            $redirectLink = apply_filters('wcsf_results_redirect_link', $redirectLink, $id);
                            $redirectCode = apply_filters('wcsf_results_redirect_code', 302, $id, $redirectLink);

                            wp_redirect($redirectLink, $redirectCode);

                            // js version of redirect
                            exit("<script>document.location = '$redirectLink';</script>");
                        }

                        return $found;
                    },
                    15,
                    2
                );
            }
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woocommerce_product_query', [$this, 'modifyProductsQuery']);
    }

    /**
     * Filter woocommerce widget filter links
     *
     * @param string $link
     *
     * @return string
     */
    public function layeredNavLinkFilter($link)
    {
        $keys = [
            'wcsf-results',
            'wcsf-value',
            'wcsf-referer'
        ];

        foreach ($keys as $key) {
            if (isset($_GET[$key])) {
                $link = add_query_arg($key, $_GET[$key], $link);
            }
        }

        return $link;
    }

    /**
     * Filter the WOOF plugin query
     *
     * @param array $query
     *
     * @return array
     */
    public function modifyWoofQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return $query;
        }

        do_action('wcsf_before_modify_request', $query);

        $id = (int) $_GET['wcsf-results'];
        $value = (array) $_GET['wcsf-value'];
        $args = $this->getQueryArgsByFilter($id, $value);
        $maxResults = $this->filter->getSetting($id, 'max_results_number');
        $query = array_replace($query, $args);

        // override products number in the query
        if ($maxResults) {
            $query['posts_per_page'] = $maxResults;

            add_filter(
                'found_posts',
                function ($found) use ($maxResults) {
                    return min($found, $maxResults);
                }
            );
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woof_products_query', [$this, 'modifyWoofQuery']);

        return $query;
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Get settings model array
     *
     * @return array
     */
    public static function getSettingsModel()
    {
        return apply_filters(
            'wcsf_global_settings_model',
            [
                'styles_including_type' => [
                    'name' => L10N::r('Styles including type'),
                    'default' => 'full',
                    'type' => 'select',
                    'desc' => L10N::r('For more info check the')
                        . ' <a href="' . WC_STEP_FILTER_PLUGIN_URL
                        . 'documentation/index.html#settings" target="_blank">'
                        . L10N::r('documentation') . '</a>',
                    'id' => 'woocommerce_step_filter_styles_including_type',
                    'key' => 'woocommerce_step_filter_styles_including_type',
                    'options' => [
                        'full' => L10N::r('Full styles file'),
                        'basic' => L10N::r('Basic styles file'),
                        'none' => L10N::r('None'),
                        'custom' => L10N::r('Custom full styles file')
                    ],
                    'section' => ''
                ],
                'scripts_including_type' => [
                    'name' => L10N::r('Scripts including type'),
                    'default' => 'single',
                    'type' => 'select',
                    'desc' => 'Include scripts separately or in a single file',
                    'id' => 'woocommerce_step_filter_scripts_including_type',
                    'key' => 'woocommerce_step_filter_scripts_including_type',
                    'options' => [
                        'single' => L10N::r('Single file'),
                        'multiple' => L10N::r('Multiple files')
                    ],
                    'section' => ''
                ],
                'included_scripts' => [
                    'name' => L10N::r('Included scripts'),
                    'default' => [
                        'app',
                        'hooks',
                        'wNumb',
                        'nouislider',
                        'nouislider-launcher'
                    ],
                    'type' => 'multiselect',
                    'desc' => 'Select which files will be included with the multiple including type',
                    'id' => 'woocommerce_step_filter_included_scripts',
                    'key' => 'woocommerce_step_filter_included_scripts',
                    'options' => [
                        'app' => 'app',
                        'hooks' => 'hooks',
                        'wNumb' => 'wNumb',
                        'nouislider' => 'noUiSlider',
                        'nouislider-launcher' => 'noUiSlider-launcher'
                    ],
                    'section' => ''
                ],
                'store_session_in_db' => [
                    'name' => L10N::r('Store session in the DB'),
                    'default' => 'no',
                    'type' => 'checkbox',
                    'desc' => L10N::r('Tick in a case of problems of the filter\'s state storing'),
                    'id' => 'woocommerce_step_filter_store_session_in_db',
                    'key' => 'woocommerce_step_filter_store_session_in_db',
                    'section' => ''
                ],
                'send_state_hash_ajax' => [
                    'name' => L10N::r('Send current state hash via AJAX'),
                    'default' => 'no',
                    'type' => 'checkbox',
                    'desc' => L10N::r('Might help in case of freezing of the filter\'s state with caching plugins'),
                    'id' => 'woocommerce_step_filter_send_state_hash_ajax',
                    'key' => 'woocommerce_step_filter_send_state_hash_ajax',
                    'section' => ''
                ],
                'custom_styles_minification' => [
                    'name' => L10N::r('Minify custom styles'),
                    'default' => 'yes',
                    'type' => 'checkbox',
                    'id' => 'woocommerce_step_filter_custom_styles_minification',
                    'key' => 'woocommerce_step_filter_custom_styles_minification',
                    'section' => 'custom_styles'
                ],
                'custom_styles_mode' => [
                    'name' => L10N::r('Mode'),
                    'default' => 'simple',
                    'type' => 'select',
                    'id' => 'woocommerce_step_filter_custom_styles_mode',
                    'key' => 'woocommerce_step_filter_custom_styles_mode',
                    'options' => [
                        'simple' => L10N::r('Simple'),
                        'advanced' => L10N::r('Advanced')
                    ],
                    'section' => 'custom_styles'
                ],
                'style_font_size' => [
                    'name' => L10N::r('Font size'),
                    'default' => '1rem',
                    'type' => 'text',
                    'id' => 'woocommerce_step_filter_style_font_size',
                    'key' => 'woocommerce_step_filter_style_font_size',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_primary' => [
                    'name' => L10N::r('Primary color'),
                    'default' => '#007bff',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_primary',
                    'key' => 'woocommerce_step_filter_style_color_primary',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_secondary' => [
                    'name' => L10N::r('Secondary color'),
                    'default' => '#6c757d',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_secondary',
                    'key' => 'woocommerce_step_filter_style_color_secondary',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_success' => [
                    'name' => L10N::r('Success color'),
                    'default' => '#28a745',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_success',
                    'key' => 'woocommerce_step_filter_style_color_success',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_info' => [
                    'name' => L10N::r('Info color'),
                    'default' => '#17a2b8',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_info',
                    'key' => 'woocommerce_step_filter_style_color_info',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_warning' => [
                    'name' => L10N::r('Warning color'),
                    'default' => '#ffc107',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_warning',
                    'key' => 'woocommerce_step_filter_style_color_warning',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_danger' => [
                    'name' => L10N::r('Danger color'),
                    'default' => '#dc3545',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_danger',
                    'key' => 'woocommerce_step_filter_style_color_danger',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_light' => [
                    'name' => L10N::r('Light color'),
                    'default' => '#f8f9fa',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_light',
                    'key' => 'woocommerce_step_filter_style_color_light',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'style_color_dark' => [
                    'name' => L10N::r('Dark color'),
                    'default' => '#343a40',
                    'type' => 'color',
                    'id' => 'woocommerce_step_filter_style_color_dark',
                    'key' => 'woocommerce_step_filter_style_color_dark',
                    'custom_attributes' => ['data-mode' => 'simple'],
                    'section' => 'custom_styles'
                ],
                'custom_scss' => [
                    'name' => L10N::r('Custom SCSS'),
                    'default' => str_replace(' !default', '', Styles::getSCSSVariablesString()),
                    'type' => 'textarea',
                    'desc' => L10N::r('You can overwrite any Bootstrap 5 variables or add custom styles here'),
                    'css' => 'height:400px;width:100%',
                    'id' => 'woocommerce_step_filter_custom_scss',
                    'key' => 'woocommerce_step_filter_custom_scss',
                    'custom_attributes' => ['data-mode' => 'advanced'],
                    'section' => 'custom_styles'
                ],
                'purchase_code' => [
                    'name' => L10N::r('Purchase code'),
                    'default' => '',
                    'type' => 'text',
                    'desc' => L10N::r('Verify your license key to unlock extra possibilities and enable auto-updates'),
                    'id' => 'woocommerce_step_filter_purchase_code',
                    'key' => 'woocommerce_step_filter_purchase_code',
                    'custom_attributes' => ['required' => 'required'],
                    'section' => 'license'
                ],
                'license_consent' => [
                    'name' => L10N::r('Consent'),
                    'default' => false,
                    'type' => 'checkbox',
                    'desc' => L10N::r('I agree that the license data will be transmitted to the license server'),
                    'id' => 'woocommerce_step_filter_license_consent',
                    'key' => 'woocommerce_step_filter_license_consent',
                    'custom_attributes' => ['required' => 'required'],
                    'section' => 'license'
                ]
            ]
        );
    }

    /**
     * Get one of settings
     *
     * @param string $setting
     * @param mixed $default
     *
     * @return string|boolean|array|float
     */
    public static function getSetting($setting, $default = null)
    {
        static $cache = [];

        if (isset($cache[$setting])) {
            return apply_filters('wcsf_global_setting', $cache[$setting], $setting);
        }

        $model = self::getSettingsModel();

        if (!isset($model[$setting])) {
            $cache[$setting] = $default;

            return apply_filters('wcsf_global_setting', $default, $setting);
        }

        $value = isset($model[$setting]['key']) ? get_option($model[$setting]['key'], null) : null;

        if (is_null($value)) {
            if ($default) {
                $value = $default;
            } elseif (isset($model[$setting]['default'])) {
                $value = $model[$setting]['default'];
            }
        }

        $value = Utils::handleSettingType($value, $model[$setting]['type']);
        $cache[$setting] = $value;

        return apply_filters('wcsf_global_setting', $value, $setting);
    }
    // </editor-fold>

    // <editor-fold desc="Deprecated">
    /**
     * Include php-template by the name.
     * First looking in the "theme folder/woocommerce-step-filter (WC_STEP_FILTER_THEME_TEMPLATES_DIR)"
     * Second looking in the "plugin folder/views"
     * Making extraction of the arguments as variables
     *
     * @param string $name
     * @param array  $arguments
     * @param array  $templateSettings
     *
     * @return string
     *
     * @deprecated 8.1.0
     */
    public static function getTemplatePart($name = '', $arguments = [], $templateSettings = [])
    {
        return Template::html($name, $arguments, $templateSettings);
    }

    /**
     * Get requested HTML part arguments
     *
     * @param array $defaults
     * @param array $settings
     *
     * @return array
     *
     * @deprecated 8.1.0
     */
    public static function getTemplatePartArgs($defaults = [], $settings = [])
    {
        return Template::getHTMLArgs($defaults, $settings);
    }
    // </editor-fold>
}
