<?php
namespace WCStepFilter;

/**
 * Filter Class
 *
 * @class Filter
 * @version 10.1.0
 */
class Filter implements PostTypeInterface
{
    use PostTypeTrait;

    // <editor-fold desc="Properties">
    /**
     * Current working filter ID
     * @var int
     */
    protected $currentId = null;

    /**
     * Post type string
     * @var string
     */
    public static $postType = 'wcsf';

    /**
     * Namespace string
     * @var string
     */
    public static $namespace = 'wcsf';

    /**
     * Settings model variable
     * @var array
     */
    public static $settingsModel;

    /**
     * Active steps session keys variable
     * @var string
     */
    public static $activeStepsSessionKey = 'wcsf-active-step';

    /**
     * Filter value session keys variable
     * @var string
     */
    public static $valuesSessionKey = 'wcsf-values';

    /**
     * Notices array
     * @var array
     */
    public $notices = [];

    /**
     * Ajax actions variable
     * @var array
     */
    public $ajaxActions = [];
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Class Constructor */
    public function __construct()
    {
        $this->initPostType();

        // admin
        add_action('admin_enqueue_scripts', [$this, 'enqueueAdmin']);
        add_filter('manage_' . self::$postType . '_posts_columns', [$this, 'columnsFilter']);
        add_action('manage_' . self::$postType . '_posts_custom_column', [$this, 'customColumnFilter'], 10, 2);

        $this->ajaxActions = apply_filters(
            'wcsf_ajax_actions',
            [
                'WCSFGetStep' => 'getStepHtmlAjax',
                'WCSFSubmit' => 'submitAjax',
                'WCSFSubmitAndGoTo' => 'submitAndGoToAjax',
                'WCSFSubmitAndGoNext' => 'submitAndGoNextAjax',
                'WCSFSubmitAndGoResults' => 'submitAndGoResultsAjax',
                'WCSFReset' => 'resetAjax',
                'WCSFGetPreliminaryResultsCount' => 'getPreliminaryResultsCountAjax',
            ]
        );

        foreach ($this->ajaxActions as $ajaxActionKey => $ajaxActionFunction) {
            add_action("wp_ajax_nopriv_{$ajaxActionKey}", [$this, $ajaxActionFunction]);
            add_action("wp_ajax_{$ajaxActionKey}", [$this, $ajaxActionFunction]);
        }
    }

    /**
     * Get current working filter ID
     *
     * @return int
     */
    public function getCurrentId()
    {
        return $this->currentId;
    }

    /**
     * Set current working filter ID
     *
     * @param int $currentId
     */
    public function setCurrentId($currentId)
    {
        $this->currentId = (int) $currentId;
    }

    /** Register entity post type */
    public function registerPostType()
    {
        register_post_type(
            self::$postType,
            [
                'label' => L10N::r('WooCommerce Step Filters'),
                'labels' => [
                    'name' => L10N::r('WooCommerce Step Filters'),
                    'singular_name' => L10N::r('WooCommerce Step Filter'),
                    'menu_name' => L10N::r('WooCommerce Step Filter'),
                    'add_new' => L10N::r('Add New Filter'),
                    'all_items' => L10N::r('Filters')
                ],
                'description' => L10N::r('This is where you can add new step filter items'),
                'public' => false,
                'show_ui' => true,
                'map_meta_cap' => true,
                'publicly_queryable' => false,
                'exclude_from_search' => true,
                'show_in_menu' => true,
                'hierarchical' => false,
                'rewrite' => false,
                'query_var' => false,
                'menu_icon' => 'dashicons-search',
                'supports' => [
                    'title',
                    'thumbnail',
                    'editor'
                ],
                'show_in_nav_menus' => false,
                'show_in_admin_bar' => true
            ]
        );
    }

    /**
     * Get settings model array
     *
     * @return array
     */
    public static function getSettingsModel()
    {
        $defaultResultsURL = '/';
        $controlsClassDescription = L10N::r('For more info check the')
            . ' <a href="' . WC_STEP_FILTER_PLUGIN_URL
            . 'documentation/index.html#filter-settings-controls" target="_blank">'
            . L10N::r('documentation') . '</a>';

        if (function_exists('wc_get_page_id')) {
            $defaultResultsURL = str_replace(home_url(), '', get_permalink(wc_get_page_id('shop')));
        }

        return apply_filters(
            'wcsf_settings_model',
            [
                // <editor-fold desc="Basic">
                'questions' => [
                    'label' => L10N::r('Select questions for using'),
                    'key' => '_questions',
                    'type' => 'multi-select',
                    'default' => [],
                    'values' => [],
                    'group' => L10N::r('Basic')
                ],
                'questions_categories' => [
                    'label' => L10N::r('Select questions categories for using'),
                    'key' => '_questions_categories',
                    'type' => 'multi-select',
                    'default' => [],
                    'values' => [],
                    'group' => L10N::r('Basic')
                ],
                'source' => [
                    'label' => L10N::r('Source of the steps'),
                    'key' => '_source',
                    'type' => 'select',
                    'default' => 'posts',
                    'values' => [
                        'posts' => L10N::r('Posts'),
                        'categories' => L10N::r('Categories')
                    ],
                    'group' => L10N::r('Basic')
                ],
                'mode' => [
                    'label' => L10N::r('Mode'),
                    'key' => '_mode',
                    'type' => 'select',
                    'default' => 'free-walk',
                    'values' => [
                        'free-walk' => L10N::r('Free walk'),
                        'step-by-step' => L10N::r('Step by step'),
                        'single-step' => L10N::r('Single step'),
                        'sequence' => L10N::r('Sequence')
                    ],
                    'group' => L10N::r('Basic')
                ],
                'filter_by_queried_object' => [
                    'label' => L10N::r('Filter by the current opened term'),
                    'key' => '_filter_by_queried_object',
                    'type' => 'checkbox',
                    'default' => true,
                    'description' => L10N::r('Apply current opened term page as a filter extra query argument'),
                    'group' => L10N::r('Basic')
                ],
                'extra_query_args' => [
                    'label' => L10N::r('Extra query arguments'),
                    'key' => '_extra_query_args',
                    'type' => 'data-table',
                    'default' => [],
                    'values' => [
                        'source' => [
                            'label' => L10N::r('Source'),
                            'key' => 'source',
                            'type' => 'select',
                            'default' => 'product',
                            'values' => [
                                'product' => L10n::r('Product'),
                                'category' => L10n::r('Category'),
                                'attribute' => L10n::r('Attribute')
                            ]
                        ],
                        'product' => [
                            'label' => L10N::r('Product'),
                            'key' => 'product',
                            'type' => 'wc-product-search',
                            'multiple' => false,
                            'default' => []
                        ],
                        'category' => [
                            'label' => L10N::r('Category'),
                            'key' => 'category',
                            'type' => 'wc-terms-search',
                            'multiple' => false,
                            'default' => []
                        ],
                        'attribute_name' => [
                            'label' => L10N::r('Attribute name'),
                            'key' => 'attribute_name',
                            'type' => 'select',
                            'default' => '',
                            'values' => []
                        ],
                        'attribute_value' => [
                            'label' => L10N::r('Attribute value'),
                            'key' => 'attribute_value',
                            'type' => 'ajax-select',
                            'target-parent' => '[data-component="wcsf-data-table-item"]',
                            'target-selector' => '[data-key="attribute_name"] select',
                            'action' => 'wcsfGetAttributeValuesSelectOptions',
                            'default' => '',
                        ],
                        'condition' => [
                            'label' => L10N::r('Condition'),
                            'key' => 'condition',
                            'type' => 'select',
                            'default' => '=',
                            'values' => [
                                '=' => '=',
                                '!=' => '!='
                            ]
                        ]
                    ],
                    'inModal' => true,
                    'description' => L10N::r('Additional query arguments for filtering'),
                    'group' => L10N::r('Basic')
                ],
                // </editor-fold>
                // <editor-fold desc="Results">
                'results_url' => [
                    'label' => L10N::r('Results URL'),
                    'key' => '_results_url',
                    'type' => 'text',
                    'default' => $defaultResultsURL,
                    // phpcs:disable
                    'description' => L10N::r('Use # symbol to stay on the same page (for example, for using the filter as a widget)'),
                    // phpcs:enable
                    'group' => L10N::r('Results')
                ],
                'empty_results_url' => [
                    'label' => L10N::r('Empty results URL'),
                    'key' => '_empty_results_url',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Open another page in a case of empty results'),
                    'group' => L10N::r('Results')
                ],
                'redirect_to_only_result_page' => [
                    'label' => L10N::r('Instant redirect to the only result page'),
                    'key' => '_redirect_to_only_result_page',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Not works for within results step'),
                    'group' => L10N::r('Results')
                ],
                'max_results_number' => [
                    'label' => L10N::r('Maximum products number in results'),
                    'key' => '_max_results_number',
                    'type' => 'number',
                    'default' => 0,
                    'min' => 0,
                    'description' => L10N::r('Zero is equal infinity'),
                    'group' => L10N::r('Results'),
                ],
                'reset_filter_on_showing' => [
                    'label' => L10N::r('Reset filter on showing'),
                    'key' => '_reset_filter_on_showing',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Recommended for the widget filters'),
                    'group' => L10N::r('Results'),
                    'separate' => true
                ],
                'show_results_within' => [
                    'label' => L10N::r('Show results within the filter'),
                    'key' => '_show_results_within',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('Results')
                ],
                'results_within_type' => [
                    'label' => L10N::r('Results within type'),
                    'key' => '_results_within_type',
                    'type' => 'select',
                    'values' => [
                        'step' => L10N::r('Separate step'),
                        'under-steps' => L10N::r('Under steps')
                    ],
                    'default' => 'step',
                    'description' => L10N::r('Doesn\'t matter for single step mode'),
                    'group' => L10N::r('Results')
                ],
                'results_within_products_per_page' => [
                    'label' => L10N::r('Products per page for results within'),
                    'key' => '_results_within_products_per_page',
                    'type' => 'number',
                    'default' => 0,
                    'description' => L10N::r('Zero is equal the default WooCommerce setting'),
                    'group' => L10N::r('Results')
                ],
                'results_step_title' => [
                    'label' => L10N::r('Results step title'),
                    'key' => '_results_tab_title',
                    'type' => 'text',
                    'default' => L10N::r('Results'),
                    'group' => L10N::r('Results')
                ],
                'results_step_thumbnail' => [
                    'label' => L10N::r('Results step thumbnail'),
                    'key' => '_results_tab_thumbnail',
                    'type' => 'thumbnail',
                    'default' => '',
                    'group' => L10N::r('Results'),
                    'separate' => true
                ],
                'show_preliminary_results_count' => [
                    'label' => L10N::r('Show preliminary results count'),
                    'key' => '_show_preliminary_results_count',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('In a tooltip'),
                    'group' => L10N::r('Results')
                ],
                'preliminary_results_count_text' => [
                    'label' => L10N::r('Preliminary results count text'),
                    'key' => '_preliminary_results_count_text',
                    'type' => 'text',
                    'default' => L10N::r('Products found: %d'),
                    'description' => L10N::r('%d - products count variable')
                        . '; ' . L10N::r('%min - products min price variable')
                        . '; ' . L10N::r('%max - products max price variable'),
                    'group' => L10N::r('Results')
                ],
                'preliminary_results_count_text_one' => [
                    'label' => L10N::r('One preliminary result count text'),
                    'key' => '_preliminary_results_count_text_one',
                    'type' => 'text',
                    'default' => L10N::r('One product at %max'),
                    'group' => L10N::r('Results')
                ],
                'preliminary_results_count_text_zero' => [
                    'label' => L10N::r('Zero preliminary results count text'),
                    'key' => '_preliminary_results_count_text_zero',
                    'type' => 'text',
                    'default' => L10N::r('No products found'),
                    'group' => L10N::r('Results')
                ],
                // </editor-fold>
                // <editor-fold desc="Behavior">
                'collect_inquiries' => [
                    'label' => L10N::r('Collect inquiries for statistic'),
                    'key' => '_collect_inquiries',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('Behavior'),
                    'separate' => true
                ],
                'scrolling_top_on_update' => [
                    'label' => L10N::r('Scrolling to the top of the step on form update'),
                    'key' => '_scrolling_top_on_update',
                    'type' => 'checkbox',
                    'default' => true,
                    'group' => L10N::r('Behavior')
                ],
                'scrolling_up_gap' => [
                    'label' => L10N::r('The gap on scrolling up (px)'),
                    'key' => '_scrolling_up_gap',
                    'type' => 'number',
                    'default' => 0,
                    'group' => L10N::r('Behavior'),
                    'separate' => true
                ],
                'nav_action' => [
                    'label' => L10N::r('Navigation action'),
                    'key' => '_nav_action',
                    'type' => 'select',
                    'default' => 'submit-and-go-to',
                    'values' => [
                        'submit-and-go-to' => L10N::r('Submit and go to step'),
                        'get-step' => L10N::r('Go to step')
                    ],
                    'group' => L10N::r('Behavior')
                ],
                // </editor-fold>
                // <editor-fold desc="View">
                'nav_view' => [
                    'label' => L10N::r('Navigation view'),
                    'key' => '_nav_view',
                    'type' => 'select',
                    'default' => 'buttons',
                    'values' => [],
                    'group' => L10N::r('View')
                ],
                'sticky_nav' => [
                    'label' => L10N::r('Sticky navigation'),
                    'key' => '_sticky_nav',
                    'type' => 'select',
                    'default' => 'always',
                    'values' => [
                        'never' => L10N::r('Never'),
                        'xxs' => L10N::r('XXS'),
                        'xs' => L10N::r('XS'),
                        'sm' => L10N::r('S'),
                        'md' => L10N::r('M'),
                        'lg' => L10N::r('L'),
                        'xl' => L10N::r('XL'),
                        'xxl' => L10N::r('XXL'),
                        'always' => L10N::r('Always')
                    ],
                    'description' => L10N::r('Select required screen size'),
                    'group' => L10N::r('View')
                ],
                'sticky_nav_offset_top' => [
                    'label' => L10N::r('Sticky navigation offset top'),
                    'key' => '_sticky_nav_offset_top',
                    'type' => 'number',
                    'default' => 0,
                    'description' => L10N::r('px'),
                    'group' => L10N::r('View'),
                    'separate' => true
                ],
                'show_header' => [
                    'label' => L10N::r('Show header'),
                    'key' => '_show_header',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('View')
                ],
                'sticky_header' => [
                    'label' => L10N::r('Sticky header'),
                    'key' => '_sticky_header',
                    'type' => 'select',
                    'default' => 'always',
                    'values' => [
                        'never' => L10N::r('Never'),
                        'xxs' => L10N::r('XXS'),
                        'xs' => L10N::r('XS'),
                        'sm' => L10N::r('S'),
                        'md' => L10N::r('M'),
                        'lg' => L10N::r('L'),
                        'xl' => L10N::r('XL'),
                        'xxl' => L10N::r('XXL'),
                        'always' => L10N::r('Always')
                    ],
                    'description' => L10N::r('Select required screen size'),
                    'group' => L10N::r('View')
                ],
                'sticky_header_offset_top' => [
                    'label' => L10N::r('Sticky header offset top'),
                    'key' => '_sticky_header_offset_top',
                    'type' => 'number',
                    'default' => 0,
                    'description' => L10N::r('px'),
                    'group' => L10N::r('View'),
                    'separate' => true
                ],
                'show_footer' => [
                    'label' => L10N::r('Show footer'),
                    'key' => '_show_footer',
                    'type' => 'checkbox',
                    'default' => true,
                    'group' => L10N::r('View')
                ],
                'sticky_footer' => [
                    'label' => L10N::r('Sticky footer'),
                    'key' => '_sticky_footer',
                    'type' => 'select',
                    'default' => 'always',
                    'values' => [
                        'never' => L10N::r('Never'),
                        'xxs' => L10N::r('XXS'),
                        'xs' => L10N::r('XS'),
                        'sm' => L10N::r('S'),
                        'md' => L10N::r('M'),
                        'lg' => L10N::r('L'),
                        'xl' => L10N::r('XL'),
                        'xxl' => L10N::r('XXL'),
                        'always' => L10N::r('Always')
                    ],
                    'description' => L10N::r('Select required screen size'),
                    'group' => L10N::r('View'),
                    'separate' => true
                ],
                'enable_description_step' => [
                    'label' => L10N::r('Enable description step'),
                    'key' => '_enable_description_tab',
                    'type' => 'checkbox',
                    'default' => true,
                    'group' => L10N::r('View')
                ],
                'description_step_title' => [
                    'label' => L10N::r('Description step title'),
                    'key' => '_description_tab_title',
                    'type' => 'text',
                    'default' => L10N::r('Welcome'),
                    'group' => L10N::r('View'),
                    'separate' => true
                ],
                'enable_value' => [
                    'label' => L10N::r('Enable value list'),
                    'key' => '_enable_value',
                    'type' => 'checkbox',
                    'default' => true,
                    'description' => L10N::r('Output list of the answers'),
                    'group' => L10N::r('View')
                ],
                'enable_shop_value' => [
                    'label' => L10N::r('Enable shop value list'),
                    'key' => '_enable_shop_value',
                    'type' => 'checkbox',
                    'default' => true,
                    'description' =>
                        L10N::r('Output list of the answers. Appears on the shop page while results showing.'),
                    'group' => L10N::r('View')
                ],
                // </editor-fold>
                // <editor-fold desc="Results PDF">
                'results_pdf_header_height' => [
                    'label' => L10N::r('Results PDF header height'),
                    'key' => '_results_pdf_header_height',
                    'type' => 'number',
                    'default' => 2,
                    'min' => 0,
                    'description' => L10N::r('cm'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_header_content' => [
                    'label' => L10N::r('Results PDF header content'),
                    'key' => '_results_pdf_header_content',
                    'type' => 'editor',
                    'default' => get_bloginfo('name'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_top_description' => [
                    'label' => L10N::r('Results PDF top description'),
                    'key' => '_results_pdf_top_description',
                    'type' => 'editor',
                    'default' => L10N::r('Your filtering results'),
                    'description' => L10N::r('PDF shortcodes for using:')
                        . '<br>'
                        . '[wcsf-result-pdf-page-number], [wcsf-result-pdf-page-total], [wcsf-result-pdf-new-page]',
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_footer_height' => [
                    'label' => L10N::r('Results PDF footer height'),
                    'key' => '_results_pdf_footer_height',
                    'type' => 'number',
                    'default' => 2,
                    'min' => 0,
                    'description' => L10N::r('cm'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_footer_content' => [
                    'label' => L10N::r('Results PDF footer content'),
                    'key' => '_results_pdf_footer_content',
                    'type' => 'editor',
                    'default' => '[wcsf-result-pdf-page-number] / [wcsf-result-pdf-page-total]',
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_bottom_description' => [
                    'label' => L10N::r('Results PDF bottom description'),
                    'key' => '_results_pdf_bottom_description',
                    'type' => 'editor',
                    'default' => 'Get more info: ' . get_bloginfo('url'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_additional_css' => [
                    'label' => L10N::r('Results PDF additional CSS'),
                    'key' => '_results_pdf_additional_css',
                    'type' => 'textarea',
                    'default' => 'body {font-size: 12pt}'
                        . PHP_EOL . '.header-inner, .footer-inner {padding: 1em; background-color: #f5f5f5;}',
                    'group' => L10N::r('Results PDF')
                ],
                // </editor-fold>
                // <editor-fold desc="Controls">
                'header_controls' => [
                    'label' => L10N::r('Header controls'),
                    'key' => '_header_controls',
                    'type' => 'multi-select',
                    'default' => [
                        'reset',
                        'back',
                        'preliminary-results',
                        'spacer',
                        'start',
                        'skip',
                        'next',
                        'results',
                        'results-pdf'
                    ],
                    'values' => [
                        'preliminary-results' => L10N::r('Preliminary results'),
                        'spacer' => L10N::r('Free space'),
                        'spacer-2' => L10N::r('Free space'),
                        'start' => L10N::r('Start'),
                        'back' => L10N::r('Back'),
                        'skip' => L10N::r('Skip'),
                        'next' => L10N::r('Next'),
                        'reset' => L10N::r('Reset'),
                        'results' => L10N::r('Results'),
                        'results-pdf' => L10N::r('Results pdf')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'footer_controls' => [
                    'label' => L10N::r('Footer controls'),
                    'key' => '_footer_controls',
                    'type' => 'multi-select',
                    'default' => [
                        'reset',
                        'back',
                        'preliminary-results',
                        'spacer',
                        'start',
                        'skip',
                        'next',
                        'results',
                        'results-pdf'
                    ],
                    'values' => [
                        'preliminary-results' => L10N::r('Preliminary results'),
                        'spacer' => L10N::r('Free space'),
                        'spacer-2' => L10N::r('Free space'),
                        'start' => L10N::r('Start'),
                        'back' => L10N::r('Back'),
                        'skip' => L10N::r('Skip'),
                        'next' => L10N::r('Next'),
                        'reset' => L10N::r('Reset'),
                        'results' => L10N::r('Results'),
                        'results-pdf' => L10N::r('Results pdf')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'shop_controls' => [
                    'label' => L10N::r('Shop controls'),
                    'key' => '_shop_controls',
                    'type' => 'multi-select',
                    'default' => [
                        'spacer',
                        'back',
                        'repeat',
                        'reset'
                    ],
                    'values' => [
                        'spacer' => L10N::r('Free space'),
                        'spacer-2' => L10N::r('Free space'),
                        'back' => L10N::r('Back'),
                        'repeat' => L10N::r('Repeat'),
                        'reset' => L10N::r('Reset')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'controls_preliminary_results_text' => [
                    'label' => L10N::r('Preliminary results text'),
                    'key' => '_controls_preliminary_results_text',
                    'type' => 'text',
                    'default' => L10N::r('%d products found from %min to %max'),
                    'description' => L10N::r('%d - products count variable')
                        . '; ' . L10N::r('%min - products min price variable')
                        . '; ' . L10N::r('%max - products max price variable'),
                    'group' => L10N::r('Controls')
                ],
                'controls_preliminary_results_text_one' => [
                    'label' => L10N::r('One preliminary result text'),
                    'key' => '_controls_preliminary_results_text_one',
                    'type' => 'text',
                    'default' => L10N::r('One product at %max'),
                    'group' => L10N::r('Controls')
                ],
                'controls_preliminary_results_text_zero' => [
                    'label' => L10N::r('Zero preliminary results text'),
                    'key' => '_controls_preliminary_results_text_zero',
                    'type' => 'text',
                    'default' => L10N::r('No products found'),
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'start_button_text' => [
                    'label' => L10N::r('"Start" button text'),
                    'key' => '_start_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Start'),
                    'group' => L10N::r('Controls')
                ],
                'start_button_class' => [
                    'label' => L10N::r('"Start" button class'),
                    'key' => '_start_button_class',
                    'type' => 'text',
                    'default' => 'btn-primary',
                    'description' => $controlsClassDescription,
                    'separate' => true,
                    'group' => L10N::r('Controls')
                ],
                'back_button_text' => [
                    'label' => L10N::r('"Back" button text'),
                    'key' => '_back_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Back'),
                    'group' => L10N::r('Controls')
                ],
                'back_button_class' => [
                    'label' => L10N::r('"Back" button class'),
                    'key' => '_back_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-left hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'skip_button_text' => [
                    'label' => L10N::r('"Skip" button text'),
                    'key' => '_skip_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Skip'),
                    'group' => L10N::r('Controls')
                ],
                'skip_button_class' => [
                    'label' => L10N::r('"Skip" button class'),
                    'key' => '_skip_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'show_next_button' => [
                    'label' => L10N::r('Show "Next" button'),
                    'key' => '_show_next_button',
                    'type' => 'select',
                    'default' => 'until-last-step',
                    'values' => [
                        'until-last-step' => L10N::r('Until the last step'),
                        'always' => L10N::r('Always')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'next_button_text' => [
                    'label' => L10N::r('"Next" button text'),
                    'key' => '_next_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Next'),
                    'group' => L10N::r('Controls')
                ],
                'next_button_class' => [
                    'label' => L10N::r('"Next" button class'),
                    'key' => '_next_button_class',
                    'type' => 'text',
                    'default' => 'btn-primary show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'reset_button_text' => [
                    'label' => L10N::r('"Reset" button text'),
                    'key' => '_reset_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Reset'),
                    'group' => L10N::r('Controls')
                ],
                'reset_button_class' => [
                    'label' => L10N::r('"Reset" button class'),
                    'key' => '_reset_button_class',
                    'type' => 'text',
                    'default' => 'btn-warning show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'show_results_button' => [
                    'label' => L10N::r('Show "Results" button'),
                    'key' => '_show_results_button',
                    'type' => 'select',
                    'default' => 'on-answer',
                    'values' => [
                        'on-answer' => L10N::r('Then have an answer'),
                        'last-step' => L10N::r('On the last step'),
                        'always' => L10N::r('Always')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'results_button_text' => [
                    'label' => L10N::r('"Results" button text'),
                    'key' => '_results_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Show results'),
                    'group' => L10N::r('Controls')
                ],
                'results_button_class' => [
                    'label' => L10N::r('"Results" button class'),
                    'key' => '_results_button_class',
                    'type' => 'text',
                    'default' => 'btn-success show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'show_results_pdf_button' => [
                    'label' => L10N::r('Show "Results PDF" button'),
                    'key' => '_show_results_pdf_button',
                    'type' => 'select',
                    'default' => 'on-answer',
                    'values' => [
                        'on-answer' => L10N::r('Then have an answer'),
                        'last-step' => L10N::r('On the last step'),
                        'always' => L10N::r('Always')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'results_pdf_button_text' => [
                    'label' => L10N::r('"Results PDF" button text'),
                    'key' => '_results_pdf_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Get PDF'),
                    'group' => L10N::r('Controls')
                ],
                'results_pdf_button_class' => [
                    'label' => L10N::r('"Results PDF" button class'),
                    'key' => '_results_pdf_button_class',
                    'type' => 'text',
                    'default' => 'btn-info show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'shop_reset_button_text' => [
                    'label' => L10N::r('Shop "Reset" button text'),
                    'key' => '_shop_reset_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Reset'),
                    'group' => L10N::r('Controls')
                ],
                'shop_reset_button_class' => [
                    'label' => L10N::r('Shop "Reset" button class'),
                    'key' => '_shop_reset_button_class',
                    'type' => 'text',
                    'default' => 'btn-warning show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'shop_back_button_text' => [
                    'label' => L10N::r('Shop "Back" button text'),
                    'key' => '_shop_back_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Back'),
                    'group' => L10N::r('Controls')
                ],
                'shop_back_button_url' => [
                    'label' => L10N::r('Shop "Back" button URL'),
                    'key' => '_shop_back_button_url',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Keep blank to get previous page automatically'),
                    'group' => L10N::r('Controls')
                ],
                'shop_back_button_class' => [
                    'label' => L10N::r('Shop "Back" button class'),
                    'key' => '_shop_back_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-left hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ],
                'shop_repeat_button_text' => [
                    'label' => L10N::r('Shop "Repeat" button text'),
                    'key' => '_shop_repeat_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Repeat'),
                    'group' => L10N::r('Controls')
                ],
                'shop_repeat_button_url' => [
                    'label' => L10N::r('Shop "Repeat" button URL'),
                    'key' => '_shop_repeat_button_url',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Keep blank to get previous page automatically'),
                    'group' => L10N::r('Controls')
                ],
                'shop_repeat_button_class' => [
                    'label' => L10N::r('Shop "Repeat" button class'),
                    'key' => '_shop_repeat_button_class',
                    'type' => 'text',
                    'default' => 'btn-info show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls'),
                    'separate' => true
                ]
                // </editor-fold>
            ]
        );
    }
    // </editor-fold>

    // <editor-fold desc="Admin">
    /** Admin styles and scripts enqueue */
    public function enqueueAdmin()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';

        wp_enqueue_script(
            'woocommerce-step-filter-filter',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/filter$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        if (defined('WC_VERSION')
            && ((isset($_GET['post']) && get_post_type((int) $_GET['post']) == self::$postType)
                || (isset($_GET['post_type']) && $_GET['post_type'] == self::$postType))
        ) {
            wp_register_script(
                'select2',
                \WC()->plugin_url() . "/assets/js/select2/select2.full$suffix.js",
                'jquery',
                '4.0.3'
            );

            wp_register_script(
                'selectWoo',
                WC()->plugin_url() . "/assets/js/selectWoo/selectWoo.full$suffix.js",
                'jquery',
                '1.0.0'
            );

            wp_register_script(
                'wc-enhanced-select',
                WC()->plugin_url() . "/assets/js/admin/wc-enhanced-select$suffix.js",
                ['jquery', 'selectWoo'],
                WC_VERSION
            );

            wp_enqueue_script('select2');
            wp_enqueue_script('selectWoo');
            wp_enqueue_script('wc-enhanced-select');
            wp_enqueue_style('woocommerce_admin_styles', \WC()->plugin_url() . '/assets/css/admin.css', [], WC_VERSION);
        }
    }

    /**
     * Post meta-box view
     *
     * @param \WP_Post $post
     */
    public function metaBoxCallback($post)
    {
        $postSettingsModelGroups = [];
        $attributes = wc_get_attribute_taxonomies();
        $shortCodes = [
            'shortcode' => [
                'group' => L10N::r('Basic'),
                'label' => L10N::r('Shortcode for using'),
                'key' => '_shortcode',
                'type' => 'text',
                'readonly' => true,
                'default' => '[woocommerce-step-filter id="' . $post->ID . '"]'
            ]
        ];

        $settingsModel = self::getSettingsModel();

        foreach ($attributes as $attribute) {
            $settingsModel['extra_query_args']['values']['attribute_name']['values'][$attribute->attribute_name]
                = $attribute->attribute_label;
        }

        $attributeKeys = array_keys($settingsModel['extra_query_args']['values']['attribute_name']['values']);
        $settingsModel['extra_query_args']['values']['attribute_name']['default'] = reset($attributeKeys);
        $settingsModel = $shortCodes + $settingsModel;
        $settingsModel['questions']['values'] = Question::getPostsIds();
        $settingsModel['questions_categories']['values'] = Question::getCategoriesIdsList();
        $settingsModel['nav_view']['values'] = Template::getNavList();

        foreach ($settingsModel as $settingKey => $setting) {
            $postSettingsModelGroups[L10N::r($setting['group'])][$settingKey] = $setting;
        }

        $activeGroup = isset($_GET['activeGroup'], $postSettingsModelGroups[$_GET['activeGroup']])
            ? esc_sql($_GET['activeGroup'])
            : L10N::r('Basic');
        ?>
        <div data-component="wcsf-settings-groups">
            <section class="wcsf-filter wcsf-settings-group" data-component="wcsf-filter wcsf-settings-group">
                <?php foreach ($postSettingsModelGroups as $group => $settingsModel) { ?>
                    <fieldset>
                        <a class="button button-large wcsf-settings-group-toggle" href="#" role="button"
                            data-component="wcsf-settings-group-toggle"
                            data-id="<?php echo esc_attr($group); ?>"><?php echo wp_kses_post($group); ?></a>
                        <table class="form-table wcsf-settings-table wcsf-settings-group-content"
                            aria-expanded="<?php echo var_export($activeGroup == $group, true); ?>"
                            data-component="wcsf-settings-group-content"
                            data-id="<?php echo esc_attr($group); ?>">
                            <?php foreach ($settingsModel as $settingKey => $setting) { ?>
                                <tr class="wcsf-settings-table-row form-field<?php
                                    echo isset($setting['separate']) && $setting['separate'] ? ' separate' : '';
                                    ?>"
                                    data-component="wcsf-setting" data-key="<?php echo esc_attr($settingKey); ?>">
                                    <th scope="row">
                                        <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                                            echo wp_kses_post($setting['label']);
                                            ?></label>
                                    </th>
                                    <td><?php
                                        Admin::settingFieldView(
                                            $setting,
                                            ['values' => [$setting['key'] => self::getSetting($post->ID, $settingKey)]]
                                        );
                                        ?></td>
                                </tr>
                            <?php } ?>
                        </table>
                    </fieldset>
                <?php } ?>
            </section>
        </div>
        <?php
    }

    /**
     * Post list columns filter
     *
     * @param array $columns
     *
     * @return array $columns
     */
    public function columnsFilter($columns)
    {
        $columns['shortcode'] = L10N::r('ShortCode');

        return $columns;
    }

    /**
     * Post list line cell
     *
     * @param array $columns
     * @param integer $postId
     */
    public function customColumnFilter($columns, $postId)
    {
        if ($columns == 'shortcode') {
            echo '[woocommerce-step-filter id="' . (int) $postId . '"]';
        }
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Get questions list setting for any source type
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getQuestionsList($id)
    {
        switch (self::getSetting($id, 'source')) {
            case 'categories':
                $output = [];
                $stepsIds = self::getSetting($id, 'questions_categories');

                foreach ($stepsIds as $stepId) {
                    $questions = Question::getPostsIds([
                        'tax_query' => [
                            [
                                'taxonomy' => Question::$taxonomy,
                                'terms' => $stepId
                            ]
                        ]
                    ]);

                    $output = array_replace(
                        $output,
                        Question::getList($id, array_keys($questions))
                    );
                }

                break;

            default:
            case 'posts':
                $output = Question::getList($id, self::getSetting($id, 'questions'));
        }

        return apply_filters('wcsf_questions_list', $output, $id);
    }
    // </editor-fold>

    // <editor-fold desc="Steps">
    /**
     * Get filter steps array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getSteps($id)
    {
        static $cache = [];

        if (isset($cache[$id])) {
            return apply_filters('wcsf_steps', $cache[$id], $id);
        }

        switch (self::getSetting($id, 'source')) {
            case 'categories':
                $output = [];
                $stepsIds = self::getSetting($id, 'questions_categories');

                foreach (array_filter($stepsIds) as $stepId) {
                    $term = get_term($stepId, Question::$taxonomy);

                    if ($term instanceof \WP_Error) {
                        continue;
                    }

                    $questions = Question::getPostsIds([
                        'orderby' => 'menu_order',
                        'order' => 'ASC',
                        'tax_query' => [
                            [
                                'taxonomy' => Question::$taxonomy,
                                'terms' => $stepId
                            ]
                        ]
                    ]);

                    $output[$stepId] = (object) [
                        'id' => $stepId,
                        'name' => $term->name,
                        'description' => $term->description,
                        'thumbnail' => Question::getCategorySetting($stepId, 'thumbnail'),
                        'layout' => Question::getCategorySetting($stepId, 'layout'),
                        'gridColumns' => Question::getCategorySetting($stepId, 'grid_columns'),
                        'children' => Question::getList($id, array_keys($questions))
                    ];
                }

                break;
            default:
            case 'posts':
                $output = Question::getList($id, self::getSetting($id, 'questions'));
        }

        if (self::getSetting($id, 'enable_description_step')) {
            $output = array_replace(
                [
                    'description' => (object) [
                        'id' => 'description',
                        'name' => self::getSetting($id, 'description_step_title'),
                        'description' => get_post_field('post_content', $id),
                        'thumbnail' => get_post_thumbnail_id($id)
                    ]
                ],
                $output
            );
        }

        if (self::getSetting($id, 'show_results_within')
            && self::getSetting($id, 'results_within_type') == 'step'
        ) {
            $output['results'] = (object) [
                'id' => 'results',
                'name' => self::getSetting($id, 'results_step_title'),
                'thumbnail' => self::getSetting($id, 'results_step_thumbnail')
            ];
        }

        $cache[$id] = $output;

        return apply_filters('wcsf_steps', $output, $id);
    }

    /**
     * Get filter steps ids array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getStepsIds($id)
    {
        static $cache = [];

        if (isset($cache[$id])) {
            return apply_filters('wcsf_steps_ids', $cache[$id], $id);
        }

        $output = [];

        foreach (self::getSteps($id) as $step) {
            $output[] = $step->id;
        }

        $cache[$id] = $output;

        return apply_filters('wcsf_steps_ids', $output, $id);
    }

    /**
     * Get current progress level from zero to the last step
     *
     * @param integer $id
     *
     * @return string|null
     */
    public static function getStepsProgress($id)
    {
        $activeStepId = self::getActiveStepId($id);
        $output = 0;

        foreach (self::getStepsIds($id) as $stepId) {
            if ($stepId == $activeStepId) {
                return $output;
            }

            $output += 1;
        }

        return $output;
    }

    /**
     * Get filter active step id
     *
     * @param integer $id
     *
     * @return integer
     */
    public static function getActiveStepId($id)
    {
        $output = Storage::get(self::$activeStepsSessionKey, $id);

        if (!$output) {
            $stepsIds = self::getStepsIds($id);
            $output = reset($stepsIds);
        }

        return apply_filters('wcsf_active_step_id', $output, $id);
    }

    /**
     * Get filter active step
     *
     * @param integer $id
     *
     * @return object|null
     */
    public static function getActiveStep($id)
    {
        $steps = self::getSteps($id);
        $activeStepId = self::getActiveStepId($id);
        $mode = self::getSetting($id, 'mode');
        $metActiveItem = false;
        $output = null;

        foreach ($steps as $step) {
            if ($mode == 'step-by-step' && $metActiveItem) {
                // in the step-by-step mode reset all next steps after the active
                self::resetValue($id, $step->id);
            }

            if ($step->id == $activeStepId) {
                $metActiveItem = true;
                $output = $step;
            }
        }

        return apply_filters('wcsf_active_step', $output, $id);
    }

    /**
     * Get filter the next active step
     *
     * @param integer $id
     *
     * @return integer|string
     */
    public static function getNextActiveStepId($id)
    {
        $stepsIds = self::getStepsIds($id);
        $activeStepId = self::getActiveStepId($id);
        $foundActive = false;

        foreach ($stepsIds as $stepId) {
            if ($foundActive) {
                return apply_filters('wcsf_next_active_step_id', $stepId, $id);
            }

            if ($stepId == $activeStepId) {
                $foundActive = true;
            }
        }

        return apply_filters('wcsf_next_active_step_id', null, $id);
    }

    /**
     * Get filter the prev active step
     *
     * @param integer $id
     *
     * @return integer|string
     */
    public static function getPrevActiveStepId($id)
    {
        $stepsIds = self::getStepsIds($id);
        $activeStepId = self::getActiveStepId($id);
        $prevStep = null;

        foreach ($stepsIds as $stepId) {
            if ($stepId == $activeStepId) {
                return apply_filters('wcsf_prev_active_step_id', $prevStep, $id);
            }

            $prevStep = $stepId;
        }

        return apply_filters('wcsf_prev_active_step_id', end($stepsIds), $id);
    }

    /**
     * Set active filter step to the session variable
     *
     * @param integer $id
     * @param integer|string $step
     */
    public static function setActiveStep($id, $step)
    {
        Storage::set(self::$activeStepsSessionKey, $id, $step);
    }

    /**
     * Reset filter or question value. Set the first step as active if needed
     *
     * @param integer $id
     * @param integer $questionId
     */
    public static function resetState($id, $questionId = null)
    {
        self::resetValue($id, $questionId);

        if (!$questionId) {
            $stepsIds = self::getStepsIds($id);

            self::setActiveStep($id, reset($stepsIds));
        }
    }
    // </editor-fold>

    // <editor-fold desc="Notices">
    /**
     * Add notice by type into a queue
     *
     * @param string $stepId
     * @param array $massageData
     */
    public function addNotice($stepId, $massageData)
    {
        $this->notices[$stepId][] = $massageData;
    }

    /**
     * Return the queue of notices
     *
     * @param string $stepId - try to get messages from one step by id or output all of messages
     *
     * @return array
     */
    public function getNotices($stepId = null)
    {
        if ($stepId) {
            // return step's messages array or nothing
            return isset($this->notices[$stepId]) ? $this->notices[$stepId] : [];
        } else {
            // return all steps messages
            return array_reduce($this->notices, 'array_merge', []);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Value">
    /**
     * Save the filter value
     * Value should be an array like [questionId => [value1 => 123, value2 = 345]]
     *
     * @param integer $id
     * @param array $value
     *
     * @throws \Exception
     */
    public static function setValue($id, $value)
    {
        foreach ((array) $value as $questionId => $valueItem) {
            foreach ($valueItem as $stepValueKey => $stepValue) {
                // clear empty items, but not zero
                if ($stepValue == '') {
                    unset($valueItem[$stepValueKey]);
                }
            }

            $minValuesSelected = Question::getSetting($questionId, 'min_values_selected');
            $maxValuesSelected = Question::getSetting($questionId, 'max_values_selected');

            if ($minValuesSelected && count($valueItem) < $minValuesSelected) {
                $message = str_replace(
                    [
                        '%limit%',
                        '%value%'
                    ],
                    [
                        $minValuesSelected,
                        count($valueItem)
                    ],
                    Question::getSetting($questionId, 'min_values_selected_message')
                );

                throw new \Exception($message, $questionId);
            } elseif ($maxValuesSelected && count($valueItem) > $maxValuesSelected) {
                $message = str_replace(
                    [
                        '%limit%',
                        '%value%'
                    ],
                    [
                        $maxValuesSelected,
                        count($valueItem)
                    ],
                    Question::getSetting($questionId, 'max_values_selected_message')
                );

                throw new \Exception($message, $questionId);
            }

            Storage::set(self::$valuesSessionKey, $id, $valueItem, $questionId);
            Question::passValueToBond($id, $questionId, $valueItem);
        }
    }

    /**
     * Reset the filter by step id or fully
     *
     * @param integer $id
     * @param integer $questionId
     */
    public static function resetValue($id, $questionId = null)
    {
        if ($questionId) {
            Storage::set(self::$valuesSessionKey, $id, [], $questionId);
            Question::passValueToBond($id, $questionId, []);
        } else {
            Storage::set(self::$valuesSessionKey, $id, []);
        }
    }

    /**
     * Get the filter step value
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getValue($id)
    {
        $output = (array) Storage::get(self::$valuesSessionKey, $id);
        $output = self::removeBorderlineNumberValues($id, $output);

        return apply_filters('wcsf_value', array_filter($output), $id);
    }

    /**
     * Remove number view values which equals to their min/max values
     *
     * @param integer $id
     * @param array $values
     *
     * @return array
     */
    public static function removeBorderlineNumberValues($id, $values)
    {
        foreach (array_keys($values) as $questionId) {
            $question = Question::get($questionId);
            $boundary = [];

            if (!is_object($question) || !property_exists($question, 'viewType')
                || strpos($question->viewType, 'number') === false
            ) {
                continue;
            }

            // get min/max values
            switch ($question->filterType) {
                case 'price':
                    $boundary = Utils::getBoundaryPrices();
                    break;

                case 'attribute':
                    $boundary = Utils::getBoundaryAttributeValues($question->attribute);
                    break;

                case 'meta':
                    $boundary = Utils::getBoundaryMetaValues($question->metaKey);
            }

            switch ($question->viewType) {
                case 'number-from':
                    if (isset($values[$questionId]['from']) && $values[$questionId]['from'] == $boundary['min']) {
                        unset($values[$questionId]);
                    }

                    break;
                case 'number-to':
                    if (isset($values[$questionId]['to']) && $values[$questionId]['to'] == $boundary['min']) {
                        unset($values[$questionId]);
                    }

                    break;
                case 'number-between':
                    if (isset($values[$questionId]['from']) && $values[$questionId]['from'] == $boundary['min']
                        && isset($values[$questionId]['to']) && $values[$questionId]['to'] == $boundary['max']
                    ) {
                        unset($values[$questionId]);
                    }
            }
        }

        return apply_filters('wcsf_filtered_borderline_number_values', $values, $id);
    }
    
    /**
     * Get the filter step value list
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getValueList($id)
    {
        $value = self::getValue($id);
        $stepsIds = self::getStepsIds($id);
        $output = [];

        // sorting the values according the steps
        $value = array_replace(array_flip($stepsIds), $value);

        foreach ($value as $questionId => $valueItem) {
            if (!is_array($valueItem)) {
                continue;
            }

            $question = Question::get($questionId);

            if (!is_object($question) || !property_exists($question, 'viewType')) {
                continue;
            }

            $values = [];
            $anyProductValueText = Question::getSetting($questionId, 'any_product_value_text');
            $listItem = (object) [
                'id' => $questionId,
                'name' => wp_kses_post($question->name),
                'value' => null,
                'values' => []
            ];

            if ($question->viewType == 'formula') {
                foreach ($valueItem as $name => $item) {
                    $values[] = "$name: $item";
                }

                $listItem->value = implode(', ', $values);
                $listItem->values = $values;

                if (Question::getSetting($questionId, 'output_formula_value')) {
                    try {
                        $formulaValue = Core::executeFormula($question, $valueItem);
                        $listItem->value .= str_replace(
                            ['%value%'],
                            [$formulaValue],
                            Question::getSetting($questionId, 'formula_value_text')
                        );

                        $listItem->values[] = $formulaValue;
                    } catch (\Exception $exception) {
                        $listItem->value .= ' = ' . $exception->getMessage();
                        $listItem->values[] = $exception->getMessage();
                    }
                }
            } elseif (strpos($question->viewType, 'number') !== false) {
                $unit = Question::getSetting($questionId, 'numeric_unit');
                $fromText = Question::getSetting($questionId, 'from_value_text');
                $toText = Question::getSetting($questionId, 'to_value_text');

                // get numeric values
                if ($question->viewType == 'number' && !empty($valueItem)) {
                    $values[] = reset($valueItem) . $unit;
                }

                if (isset($valueItem['from']) && in_array($question->viewType, ['number-from', 'number-between'])) {
                    $values[] = "$fromText $valueItem[from]$unit";
                }

                if (isset($valueItem['to']) && in_array($question->viewType, ['number-to', 'number-between'])) {
                    $values[] = "$toText $valueItem[to]$unit";
                }

                $listItem->value = implode(' ', $values);
                $listItem->values = $values;
            } else {
                if ($question->filterType == 'manual') {
                    // get manual names
                    foreach ($valueItem as $valueItemPart) {
                        if (isset($question->manualValues[$valueItemPart])) {
                            $values[] = $question->manualValues[$valueItemPart]['name'];
                        }
                    }
                } elseif (in_array($question->filterType, ['tag', 'category', 'attribute'])) {
                    // define taxonomy
                    switch ($question->filterType) {
                        default:
                        case 'category':
                            $taxonomy = 'product_cat';
                            break;

                        case 'tag':
                            $taxonomy = 'product_tag';
                            break;

                        case 'attribute':
                            $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                    }

                    // get terms values
                    foreach ($valueItem as $valueItemKey => $valueItemPart) {
                        // get multi-choice term name
                        if (strpos($question->viewType, 'multi-choice') !== false) {
                            if (in_array((int) $valueItemPart, [-1, 1, 2])) {
                                $term = get_term($valueItemKey, $taxonomy);

                                switch ((int) $valueItemPart) {
                                    case -1:
                                        $values[] = "<s class=\"wcsf-value-item-value-excluded\">$term->name</s>";
                                        break;

                                    case 1:
                                        $values[] = "<span class=\"wcsf-value-item-value-possible\">$term->name</span>";
                                        break;

                                    case 2:
                                    default:
                                        $values[] = "<b class=\"wcsf-value-item-value-required\">$term->name</b>";
                                }
                            }

                            continue;
                        }

                        // get "Doesn't matter" text
                        if ($valueItemPart == 0) {
                            $values[] = $anyProductValueText;

                            continue;
                        }

                        // get term name
                        $term = get_term($valueItemPart, $taxonomy);

                        if ($term) {
                            $values[] = $term->name;
                        }
                    }
                } elseif ($question->filterType == 'meta') {
                    foreach ($valueItem as $valueItemKey => $valueItemPart) {
                        // get multi-choice term name
                        if (strpos($question->viewType, 'multi-choice') !== false) {
                            if (in_array((int) $valueItemPart, [-1, 1, 2])) {
                                switch ((int) $valueItemPart) {
                                    case -1:
                                        $values[] = "<s class=\"wcsf-value-item-value-excluded\">$valueItemKey</s>";
                                        break;

                                    case 1:
                                        $values[] =
                                            "<span class=\"wcsf-value-item-value-possible\">$valueItemKey</span>";
                                        break;

                                    case 2:
                                    default:
                                        $values[] = "<b class=\"wcsf-value-item-value-required\">$valueItemKey</b>";
                                }
                            }

                            continue;
                        }

                        // get "Doesn't matter" text
                        if ($valueItemPart == '0') {
                            $values[] = $anyProductValueText;

                            continue;
                        }

                        $values[] = $valueItemPart;
                    }
                }

                $listItem->value = implode(', ', $values);
                $listItem->values = $values;
            }

            if (!empty($listItem->value)) {
                $output[] = $listItem;
            }
        }

        return apply_filters('wcsf_value_list', $output, $id);
    }
    // </editor-fold>

    // <editor-fold desc="Nav">
    /**
     * Check previous step existence
     *
     * @param integer $id
     *
     * @return boolean
     */
    public static function canGoBack($id)
    {
        $stepsIds = self::getStepsIds($id);

        return reset($stepsIds) != self::getActiveStepId($id);
    }

    /**
     * Check next step existence
     *
     * @param integer $id
     *
     * @return boolean
     */
    public static function canGoForward($id)
    {
        $stepsIds = self::getStepsIds($id);
        $activeStepId = self::getActiveStepId($id);

        return end($stepsIds) != $activeStepId && $activeStepId != 'results';
    }

    /**
     * Get filter navigation items array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getNavItems($id)
    {
        $steps = self::getSteps($id);
        $activeStepId = self::getActiveStepId($id);
        $nextStepId = self::getNextActiveStepId($id);
        $mode = self::getSetting($id, 'mode');
        $navAction = self::getSetting($id, 'nav_action');
        $isStrictWalk = $mode == 'step-by-step';
        $activeNavItem = null;
        $previousNavItem = null;
        $output = [];

        foreach ($steps as $step) {
            $item = [
                'id' => $step->id,
                'name' => $step->name,
                'class' => '',
                'component' => "wcsf-{$navAction}",
                'type' => 'submit',
                'state' => 'default',
                'disabled' => false,
                'thumbnail' => $step->thumbnail
            ];

            // step-by-step mode
            if ($activeStepId == $step->id) {
                // active step
                $activeNavItem = $step->id;
                $item['type'] = 'button';
                $item['state'] = 'active';
                $item['class'] = 'active';
                $item['component'] = 'wcsf-none';
            } elseif ($nextStepId == $step->id) {
                // next active step
                $item['state'] = 'default';
                $item['class'] = 'next-active';
            } else {
                // other steps
                $item['state'] = $activeNavItem && $isStrictWalk ? 'disabled' : 'default';
                $item['class'] = $activeNavItem && $isStrictWalk ? 'disabled' : ($activeNavItem ? 'default' : 'past');
                $item['disabled'] = $activeNavItem && $isStrictWalk;
                $item['component'] = $activeNavItem && $isStrictWalk ? "wcsf-$navAction" : 'wcsf-get-step';
            }

            $output[] = $item;
        }

        return apply_filters('wcsf_nav_items', $output, $id);
    }

    /**
     * Get pagination items array for products loop
     *
     * @param array $args
     *
     * @return array
     */
    public static function getProductsPaginationItems($args)
    {
        $output = [];
        $defaults = [
            'productsQuery' => [],
            'extraQueryArgs' => ['paged' => 1]
        ];

        $args = array_replace_recursive($defaults, $args);

        if (!$args['productsQuery'] || empty($args['productsQuery'])) {
            return [];
        }

        $paged = is_array($args['extraQueryArgs']) && isset($args['extraQueryArgs']['paged'])
            ? (int) $args['extraQueryArgs']['paged']
            : 1;

        $paginationArgs = apply_filters(
            'wcsf_products_pagination_args',
            [
                'format' => '?paged=%#%',
                'base' => '%_%',
                'total' => $args['productsQuery']->max_num_pages,
                'current' => max($paged, get_query_var('paged')),
                'show_all' => false,
                'end_size' => 1,
                'mid_size' => 2,
                'prev_next' => true,
                'prev_text'    => '&larr;',
                'next_text'    => '&rarr;',
                'type' => 'array'
            ],
            $args
        );

        $links = paginate_links($paginationArgs);

        foreach ((array) $links as $link) {
            // add custom classes
            $link = str_replace('page-numbers', 'page-numbers page-link', $link);

            // replace empty href
            $link = str_replace('href=""', 'href="?paged=1"', $link);
            $link = str_replace("href=''", 'href="?paged=1"', $link);

            preg_match_all('/<a[^>]+href=([\'"])(?<href>.+?)\1[^>]*>/i', $link, $result);

            if (!empty($result) && !empty($result['href'][0])) {
                $href = $result['href'][0];
                $linkParts = parse_url($href);

                parse_str($linkParts['query'], $linkPartsQuery);

                $paged = isset($linkPartsQuery['paged']) ? $linkPartsQuery['paged'] : 1;

                // add custom attributes
                $link = str_replace(
                    ' href=',
                    ' data-component="wcsf-products-pagination-link wcsf-get-step" data-step-id="results" data-page="'
                    . $paged . '" href=',
                    $link
                );
            }

            $output[] = (object) [
                'class' => strpos($link, 'current') !== false ? 'active' : '',
                'innerHtml' => $link
            ];
        }

        return apply_filters('wcsf_products_pagination_items', $output, $args);
    }

    /**
     * Get link URL for the shop "repeat" control
     *
     * @param integer $id
     * @param string $url
     *
     * @return string
     */
    public static function getShopRepeatUrl($id, $url)
    {
        $query = [];
        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        $query['wcsf-reset-state'] = $id;

        // build query string
        $parts['query'] = http_build_query($query);

        // stringify URL parts
        $output = Utils::buildURL($parts);

        return apply_filters('wcsf_shop_repeat_url', $output, $url, $parts);
    }
    // </editor-fold>
    
    // <editor-fold desc="Results">
    /**
     * Get filter results URL with query args
     *
     * @param integer $id
     * @param array $queryArgs - extra query args
     *
     * @return string
     */
    public static function getResultsUrl($id, $queryArgs = [])
    {
        $url = self::getSetting($id, 'results_url');
        $query = [];

        // if the settings is empty
        if (!$url) {
            $url = get_permalink(wc_get_page_id('shop'));
        }

        // will stay on the same page
        if ($url == '#') {
            $url = isset($queryArgs['wcsf-referer']) ? $queryArgs['wcsf-referer']
                : (isset($queryArgs['referer']) ? $queryArgs['referer'] : Utils::getCurrentURL());
        } else {
            $noResultsRedirectLink = self::getSetting($id, 'empty_results_url');
            $redirectToOnlyResultPage = self::getSetting($id, 'redirect_to_only_result_page');

            if ($noResultsRedirectLink || $redirectToOnlyResultPage) {
                $preliminaryResults = self::getPreliminaryResultsQuery($id);

                if ($redirectToOnlyResultPage && $preliminaryResults->post_count == 1) {
                    $url = get_permalink(reset($preliminaryResults->posts));
                } elseif ($noResultsRedirectLink && $preliminaryResults->post_count == 0) {
                    $url = $noResultsRedirectLink;
                }
            }
        }

        // if URL is absolute
        if (strpos($url, $_SERVER['HTTP_HOST']) === false) {
            $url = home_url() . '/' . $url;
        }

        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        // redirect to the product archive
        if (!empty($query['page_id']) && '' === get_option('permalink_structure')
            && wc_get_page_id('shop') == $query['page_id']
        ) {
            $url = get_post_type_archive_link('product');
            $parts = parse_url($url);
            $query = [];

            // parse query args
            if (isset($parts['query'])) {
                parse_str($parts['query'], $query);
            }
        }

        // add questions extra parameters
        foreach (self::getValue($id) as $questionId => $value) {
            $viewType = Question::getSetting($questionId, 'view_type');
            $filterType = Question::getSetting($questionId, 'filter_type');
            $additionalParam = Question::getSetting($questionId, 'additional_get_param');

            if (!$additionalParam || !trim($additionalParam) || strpos($viewType, 'multi-choice') !== false) {
                continue;
            }

            if (in_array($filterType, ['attribute', 'category', 'tag'])) {
                $additionalGetParamTermSource = Question::getSetting($questionId, 'additional_get_param_term_source');

                if (in_array($additionalGetParamTermSource, ['slug', 'name'])) {
                    switch ($filterType) {
                        default:
                        case 'category':
                            $taxonomy = 'product_cat';
                            break;

                        case 'tag':
                            $taxonomy = 'product_tag';
                            break;

                        case 'attribute':
                            $attributeId = Question::getSetting($questionId, 'attribute');
                            $attribute = wc_get_attribute($attributeId);
                            $taxonomy = $attribute->slug;
                    }

                    foreach ($value as &$valueItem) {
                        $term = get_term($valueItem, $taxonomy);

                        switch ($additionalGetParamTermSource) {
                            case 'slug':
                                $valueItem = $term->slug;
                                break;

                            case 'name':
                                $valueItem = $term->name;
                        }
                    }
                }
            }

            if (in_array($viewType, ['checkbox', 'image-checkbox', 'chips-checkbox', 'button-checkbox'])) {
                $query[$additionalParam] = $value;

                continue;
            }

            if ($viewType == 'number-between') {
                $additionalParam = explode('|', $additionalParam);

                if (count($additionalParam) > 1) {
                    foreach ($value as $valueItem) {
                        $query[array_shift($additionalParam)] = trim($valueItem);
                    }
                } else {
                    $query[array_shift($additionalParam)] = $value;
                }

                continue;
            }

            $query[$additionalParam] = reset($value);
        }

        // blend in required args
        $query['wcsf-results'] = $id;
        $query['wcsf-value'] = self::getValue($id);
        $query['wcsf-referer'] = Utils::getCurrentURL();

        // unset reset query
        if (isset($query['wcsf-reset-value'])) {
            unset($query['wcsf-reset-value']);
        }

        if (isset($query['wcsf-reset-state'])) {
            unset($query['wcsf-reset-state']);
        }

        // woof plugin query emulate
        if (class_exists('\\WOOF')) {
            $query['swoof'] = 1;
        }

        $query = array_replace($query, $queryArgs);

        // apply outer filters
        $query = apply_filters('wcsf_results_redirect_url_query', $query, $id);

        // build query string
        $parts['query'] = http_build_query($query);

        return apply_filters('wcsf_results_redirect_url', Utils::buildURL($parts), $id, $parts);
    }

    /**
     * Get filter reset URL
     *
     * @param integer $id
     * @param array $queryArgs - extra query args
     *
     * @return string
     */
    public static function getResetUrl($id, $queryArgs = [])
    {
        $query = [];
        $url = isset($queryArgs['wcsf-referer']) ? $queryArgs['wcsf-referer']
            : (isset($queryArgs['referer']) ? $queryArgs['referer'] : Utils::getCurrentURL());

        // if the settings is empty
        if (!$url) {
            $url = get_permalink(wc_get_page_id('shop'));
        }

        // if URL is absolute
        if (strpos($url, $_SERVER['HTTP_HOST']) === false) {
            $url = home_url() . '/' . $url;
        }

        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        // add reset state
        $query['wcsf-reset-state'] = $id;

        $query = array_replace($query, $queryArgs);

        // clear the query
        unset($query['wcsf-reset-value']);
        unset($query['wcsf-results']);
        unset($query['wcsf-value']);
        unset($query['wcsf-referer']);

        // apply outer filters
        $query = apply_filters('wcsf_results_reset_url_query', $query, $id);

        // build query string
        $parts['query'] = http_build_query($query);

        return apply_filters('wcsf_reset_redirect_url', Utils::buildURL($parts), $id, $parts);
    }

    /**
     * Get one filter value reset URL
     *
     * @param integer $id
     * @param integer $valueId
     *
     * @return string
     */
    public static function getResetValueUrl($id, $valueId)
    {
        $url = isset($_REQUEST['wcsf-referer']) ? $_REQUEST['wcsf-referer']
            : (isset($_REQUEST['referer']) ? $_REQUEST['referer'] : Utils::getCurrentURL());

        $parts = parse_url($url);
        $query = [];

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        // clear the referer
        if (isset($query['referer'])) {
            unset($query['referer']);
        }

        $queryCopy = $query;

        // clear the query
        unset($queryCopy['wcsf-value'][$valueId]);

        // reset filter id
        $queryCopy['wcsf-results'] = $id;

        // add reset key
        $queryCopy['wcsf-reset-value'] = $valueId;

        // build query string
        $parts['query'] = http_build_query($queryCopy);

        return Utils::buildURL($parts);
    }

    /**
     * Get preliminary results query by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $extraQueryArgs - additional WP_Query arguments
     *
     * @return \WP_Query
     */
    public static function getPreliminaryResultsQuery($id, $valuePart = [], $extraQueryArgs = [])
    {
        $defaultQueryArgs = [
            'post_type' => 'product',
            'fields' => 'ids',
            'no_found_rows' => true,
            'nopaging' => true
        ];

        $value = self::getValue($id);
        $value = array_replace_recursive($value, $valuePart);
        $queryArgs = Core::getQueryArgsByFilter($id, $value, $extraQueryArgs);
        $queryArgs = array_replace_recursive($defaultQueryArgs, $queryArgs, $extraQueryArgs);
        $queryArgs = apply_filters('wcsf_preliminary_results_query_args', $queryArgs, $id, $valuePart, $extraQueryArgs);
        $output = new \WP_Query($queryArgs);

        return apply_filters('wcsf_preliminary_results_query', $output, $id, $valuePart, $extraQueryArgs);
    }

    /**
     * Get preliminary products IDs by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $extraQueryArgs - additional WP_Query arguments
     *
     * @return array
     */
    public static function getPreliminaryResultsIds($id, $valuePart = [], $extraQueryArgs = [])
    {
        $query = self::getPreliminaryResultsQuery($id, $valuePart, $extraQueryArgs);

        return $query->posts;
    }

    /**
     * Get preliminary products count by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $extraQueryArgs - additional WP_Query arguments
     *
     * @return integer
     */
    public static function getPreliminaryResultsCount($id, $valuePart = [], $extraQueryArgs = [])
    {
        $query = self::getPreliminaryResultsQuery($id, $valuePart, $extraQueryArgs);

        return (int) $query->post_count;
    }

    /**
     * Get preliminary products count
     *
     * @param integer $id
     * @param object $results
     *
     * @return array
     */
    public static function getPreliminaryResultsMessages($id, $results)
    {
        $boundaryPrices = Utils::getBoundaryPrices($results->posts);
        $replacements = [
            '%d' => (int) $results->post_count,
            '%min' => wc_price($boundaryPrices['min']),
            '%max' => wc_price($boundaryPrices['max'])
        ];

        switch ((int) $results->post_count) {
            case 0:
                $postFix = '_zero';
                break;

            case 1:
                $postFix = '_one';
                break;

            default:
                $postFix = '';
        }

        $output = [
            'tooltip' => self::getSetting($id, 'preliminary_results_count_text' . $postFix),
            'controls' => self::getSetting($id, 'controls_preliminary_results_text' . $postFix)
        ];

        foreach ($output as &$message) {
            $message = str_replace(array_keys($replacements), array_values($replacements), $message);
        }

        unset($message);

        return apply_filters('wcsf_preliminary_results_messages', $output);
    }

    /**
     * Get products request arguments
     *
     * @param integer $id
     * @param array $args - additional WP_Query arguments
     *
     * @return array
     */
    public static function getProductsQueryArgs($id, $args = [])
    {
        $defaults = [
            'post_type' => 'product',
            'posts_per_page' => wc_get_default_products_per_row() * wc_get_default_product_rows_per_page(),
            'orderby' => $args['orderby'],
            'order' => 'ASC',
            'paged' => $args['paged'],
            'post_status' => 'publish',
            'tax_query' => []
        ];

        if (filter_var(get_option('woocommerce_hide_out_of_stock_items'), FILTER_VALIDATE_BOOLEAN)) {
            $defaults['tax_query'][] = [
                'taxonomy' => 'product_visibility',
                'field' => 'name',
                'terms' => 'outofstock',
                'operator' => 'NOT IN'
            ];
        }

        $args = array_replace_recursive($defaults, (array) $args);

        // if order is included to orderBy
        $orderParts = explode('-', $args['orderby']);

        if (isset($orderParts[1])) {
            $args['order'] = $orderParts[1];
        }

        $wcQuery = new \WC_Query();
        $output = array_merge_recursive(
            $args,
            Core::getQueryArgsByFilter($id, self::getValue($id), $args),
            $wcQuery->get_catalog_ordering_args($orderParts[0], $args['order'])
        );

        return apply_filters('wcsf_products_query_args', $output, $id, $args);
    }
    // </editor-fold>

    // <editor-fold desc="AJAX">
    /**
     * Return an AJAX reply and exit
     *
     * @param array $data
     * @param array $postData
     */
    public static function ajaxReply($data, $postData = [])
    {
        if (isset($postData['id']) && $postData['id']) {
            $data['stepId'] = self::getActiveStepId($postData['id']);

            if (Core::getSetting('send_state_hash_ajax') && !empty($postData)) {
                $data['stateHash'] = md5(serialize([
                    'value' => self::getValue($postData['id']),
                    'step' => self::getActiveStepId($postData['id'])
                ]));
            }
        }

        Utils::sendJSON($data);
    }

    /** Save value and get form template by ajax */
    public function submitAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null
        ];

        $args = array_replace($defaults, $postData);
        $args['value'] = isset($args['value']) ? (array) $args['value'] : [$args['stepId'] => []];
        $args['resetButtonUrl'] = self::getResetUrl($args['id'], $args);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            self::ajaxReply(
                [
                    'hasError' => true,
                    'message' => $exception->getMessage(),
                    'content' => Template::html('form', $args, ['echo' => false])
                ],
                $postData
            );
        }

        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Save value and get a step form template by ajax */
    public function submitAndGoToAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $args['resetButtonUrl'] = self::getResetUrl($args['id'], $args);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            self::ajaxReply(
                [
                    'hasError' => true,
                    'message' => $exception->getMessage(),
                    'content' => Template::html('form', $args, ['echo' => false])
                ],
                $postData
            );
        }

        // change step
        if ($args['stepId']) {
            self::setActiveStep($args['id'], $args['stepId']);
        }

        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Save value and get the next form template by ajax */
    public function submitAndGoNextAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'referer' => '',
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $args['resetButtonUrl'] = self::getResetUrl($args['id'], $args);
        $showResultsWithin = self::getSetting($args['id'], 'show_results_within');

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            self::ajaxReply(
                [
                    'hasError' => true,
                    'message' => $exception->getMessage(),
                    'content' => Template::html('form', $args, ['echo' => false])
                ],
                $postData
            );
        }

        // get steps after value update
        $stepsIds = self::getStepsIds($args['id']);

        if (!$showResultsWithin && end($stepsIds) == $args['stepId']) {
            self::ajaxReply(
                ['finalRedirectUrl' => self::getResultsUrl($args['id'], ['wcsf-referer' => $args['referer']])],
                $postData
            );
        }

        if (end($stepsIds) != $args['stepId']) {
            // set active step from the step id and increment it
            self::setActiveStep($args['id'], $args['stepId']);
            self::setActiveStep($args['id'], self::getNextActiveStepId($args['id']));
        }

        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Save value and get the results by ajax */
    public function submitAndGoResultsAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $stepsIds = self::getStepsIds($args['id']);
        $showResultsWithin = self::getSetting($args['id'], 'show_results_within');
        $args['resetButtonUrl'] = self::getResetUrl($args['id'], $args);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            self::ajaxReply(
                [
                    'hasError' => true,
                    'message' => $exception->getMessage(),
                    'content' => Template::html('form', $args, ['echo' => false])
                ],
                $postData
            );
        }

        // set last step as active
        self::setActiveStep($args['id'], end($stepsIds));

        if (!$showResultsWithin) {
            self::ajaxReply(
                ['finalRedirectUrl' => self::getResultsUrl($args['id'], ['wcsf-referer' => $args['referer']])],
                $postData
            );
        }

        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Save value and get the form template by ajax */
    public function getStepHtmlAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null
        ];

        $args = array_replace($defaults, $postData);
        $args['resetButtonUrl'] = self::getResetUrl($args['id'], $args);

        // change step
        if ($args['stepId']) {
            self::setActiveStep($args['id'], $args['stepId']);
        }

        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Reset filter by ajax */
    public function resetAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'resetStepId' => null
        ];

        $args = array_replace($defaults, $postData);

        if (!$args['id']) {
            return;
        }

        // don't need to pass any value from the form
        if (isset($args['value'])) {
            unset($args['value']);
        }

        self::resetState($args['id'], $args['resetStepId']);
        self::ajaxReply(['content' => Template::html('form', $args, ['echo' => false])], $postData);
    }

    /** Get preliminary results count by ajax */
    public function getPreliminaryResultsCountAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'value' => [],
            'extraQueryArgs' => []
        ];

        $args = array_replace($defaults, $postData);
        $results = self::getPreliminaryResultsQuery($args['id'], (array) $args['value'], $args['extraQueryArgs']);

        self::ajaxReply(
            array_replace(
                [
                    'count' => (int) $results->post_count,
                    'boundaryPrices' => Utils::getBoundaryPrices($results->posts)
                ],
                self::getPreliminaryResultsMessages($args['id'], $results)
            ),
            $postData
        );
    }
    // </editor-fold>
}
