<?php
namespace WCStepFilter;

/**
 * Inquiry Class
 *
 * @class Inquiry
 * @version 2.0.0
 */
class Inquiry
{
    // <editor-fold desc="Properties">
    /**
     * Filter inquiry post type name
     * @var string
     */
    public static $postTypeName = 'wcsf_inquiry';

    /**
     * Multi-choice question values
     *
     * @var array
     */
    public static $multiChoicePossibleValues;
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Class Constructor */
    public function __construct()
    {
        add_action('init', [$this, 'onInit'], 5);
        add_action('init', [$this, 'outputCSVFilesOnInit'], 100);
        add_action('wcsf_before_results_redirect', [$this, 'beforeResultsRedirectAction']);
        add_action('wcsf_products_query', [$this, 'productsQueryAction'], 10, 3);

        // ajax
        add_action('wp_ajax_WCSFGetInquiriesListTable', [$this, 'outputListTableAjax']);

        // admin
        add_action('admin_menu', [$this, 'registerMenuItem']);
        add_filter('pre_get_posts', [$this, 'preGetPostsFilter']);
    }

    /** On init action */
    public function onInit()
    {
        self::$multiChoicePossibleValues = apply_filters(
            'wsf_multi_choice_possible_values',
            [
                0 => L10N::r('Undefined'),
                1 => L10N::r('Possible'),
                2 => L10N::r('Required'),
                -1 => L10N::r('Excluded')
            ]
        );

        $args = [
            'label' => L10N::r('WooCommerce Step Filter Inquiries'),
            'labels' => [
                'name' => L10N::r('WooCommerce Step Filter Inquiries'),
                'singular_name' => L10N::r('WooCommerce Step Filter Inquiry'),
                'menu_name' => L10N::r('Inquiry'),
                'all_items' => L10N::r('Inquiries')
            ],
            'public' => false,
            'show_ui' => true,
            'map_meta_cap' => true,
            'publicly_queryable' => false,
            'exclude_from_search' => true,
            'show_in_menu' => false,
            'hierarchical' => false,
            'rewrite' => false,
            'query_var' => false,
            'supports' => [],
            'show_in_nav_menus' => false,
            'show_in_admin_bar' => false
        ];

        register_post_type(self::$postTypeName, $args);

        // purge posts
        if (isset($_POST['woocommerce-step-filter'])
            && wp_verify_nonce($_POST['woocommerce-step-filter'], basename(__FILE__))
        ) {
            if (isset($_POST['purge-filter-inquiries']) && !empty($_POST['purge-filter-inquiries'])) {
                foreach (self::getCollection($_REQUEST) as $item) {
                    wp_delete_post($item->id);
                }
            }
        }
    }

    /** Output CSVs on init if necessary */
    public function outputCSVFilesOnInit()
    {
        // export CSVs
        if (isset($_GET['wcsf-download-inquiries-list']) && $_GET['wcsf-download-inquiries-list']) {
            $name = get_the_title((int) $_GET['wcsf-download-inquiries-list']);

            if (isset($_REQUEST['date-from'])) {
                $name .= '-' . (string) $_REQUEST['date-from'];
            }

            if (isset($_REQUEST['date-to'])) {
                $name .= '-' . (string) $_REQUEST['date-to'];
            }

            self::outputCSVFile(
                self::getListCSV((int) $_GET['wcsf-download-inquiries-list'], $_REQUEST),
                $name . '-' . date('c') . '.csv'
            );
        } elseif (isset($_GET['wcsf-download-inquiries-statistic']) && $_GET['wcsf-download-inquiries-statistic']) {
            $name = get_the_title((int) $_GET['wcsf-download-inquiries-statistic']);

            if (isset($_REQUEST['date-from'])) {
                $name .= '-' . (string) $_REQUEST['date-from'];
            }

            if (isset($_REQUEST['date-to'])) {
                $name .= '-' . (string) $_REQUEST['date-to'];
            }

            self::outputCSVFile(
                self::getStatisticCSV((int) $_GET['wcsf-download-inquiries-statistic'], $_REQUEST),
                $name . '-' . date('c') . '.csv'
            );
        }
    }

    /**
     * Fires before redirect to the results page
     *
     * @param integer $filterId
     */
    public function beforeResultsRedirectAction($filterId)
    {
        if (empty($filterId) || !Filter::getSetting($filterId, 'collect_inquiries')) {
            return;
        }

        $value = Filter::getValue($filterId);

        if (empty($value)) {
            return;
        }

        self::add(
            $filterId,
            $value,
            [
                'productsIds' => Filter::getPreliminaryResultsIds($filterId),
                'referer' => isset($_REQUEST['wcsf-referer'])
                    ? $_REQUEST['wcsf-referer']
                    : (isset($_REQUEST['referer']) ? $_REQUEST['referer'] : '')
            ]
        );
    }

    /**
     * Products within filter query
     *
     * @param integer $filterId
     * @param array $arguments
     * @param \WP_Query $productsQuery
     *
     * @return int|\WP_Error
     */
    public function productsQueryAction($filterId, $arguments, $productsQuery)
    {
        if (empty($filterId) || !Filter::getSetting($filterId, 'collect_inquiries')) {
            return 0;
        }

        $value = Filter::getValue($filterId);

        if (empty($value)) {
            return 0;
        }

        return self::add(
            $filterId,
            $value,
            [
                'productsIds' => wp_list_pluck($productsQuery->posts, 'ID'),
                'referer' => isset($_REQUEST['wcsf-referer'])
                    ? $_REQUEST['wcsf-referer']
                    : (isset($_REQUEST['referer']) ? $_REQUEST['referer'] : '')
            ]
        );
    }

    /**
     * Add an inquiry record
     *
     * @param integer $filterId
     * @param array $value
     * @param array $args
     *
     * @return int|\WP_Error
     */
    public static function add($filterId, $value, $args = [])
    {
        $defaults = [
            'referer' => '',
            'productsIds' => []
        ];

        $args = array_replace($defaults, $args);
        $data = [
            'post_author' => 1, // bind all records to the first user
            'post_title' => (string) $args['referer'],
            'post_type' => self::$postTypeName,
            'post_parent' => (int) $filterId,
            'post_content' => serialize($value),
            'post_excerpt' => serialize($args['productsIds']),
            'menu_order' => count($args['productsIds'])
        ];

        $data = apply_filters('wcsf_add_inquiry_data', $data, $filterId, $value, $args);

        do_action('wcsf_before_add_inquiry', $data, $filterId, $value, $args);

        return wp_insert_post($data);
    }

    /**
     * Update inquiry record
     *
     * @param integer $id
     * @param array $args
     *
     * @return int|\WP_Error
     */
    public static function update($id, $args)
    {
        $defaults = [
            'filterId' => null,
            'value' => null,
            'resultsIds' => null,
            'resultsCount' => null
        ];

        $args = array_replace($defaults, $args);
        $data = ['ID' => $id];

        if (!is_null($args['filterId'])) {
            $data['post_parent'] = (int) $args['filterId'];
        }

        if (!is_null($args['value'])) {
            $data['post_content'] = serialize($args['value']);
        }

        if (!is_null($args['resultsIds'])) {
            $data['post_excerpt'] = serialize($args['resultsIds']);
        }

        if (!is_null($args['resultsCount'])) {
            $data['menu_order'] = (int) $args['resultsCount'];
        }

        $data = apply_filters('wcsf_update_inquiry_data', $data, $id, $args);

        return wp_update_post($data);
    }

    /**
     * Update inquiry record
     *
     * @param array $data
     * @param string $name
     */
    public static function outputCSVFile($data, $name = 'file.csv')
    {
        header('Content-Type: text/csv');
        header('Content-Disposition: attachment; filename=' . $name);

        $output = fopen('php://output', 'wb');

        foreach ($data as $row) {
            fputcsv($output, $row, "\t");
        }

        fclose($output);
        exit;
    }
    // </editor-fold>

    // <editor-fold desc="Admin">
    /** Register settings page menu item action */
    public function registerMenuItem()
    {
        add_submenu_page(
            'edit.php?post_type=' . Filter::$postType,
            L10N::r('WooCommerce Step Filter Inquiries'),
            L10N::r('Inquiries'),
            'manage_options',
            'woocommerce-step-filter-inquiries',
            [$this, 'adminPageCallback']
        );
    }

    /**
     * Filter pre get post query
     *
     * @param \WP_Query $query
     *
     * @return \WP_Query
     */
    public function preGetPostsFilter($query)
    {
        global $pagenow;

        if (!is_admin() || 'edit.php' !== $pagenow || !isset($_GET['wcsf_ids']) || empty($_GET['wcsf_ids'])) {
            return $query;
        }

        $query->set('post__in', wp_parse_id_list($_GET['wcsf_ids']));

        return $query;
    }

    /** Plugins global settings page view */
    public function adminPageCallback()
    {
        $filters = Filter::getPostsIds();
        $filterId = isset($_REQUEST['filter-id']) ? (int) $_REQUEST['filter-id'] : null;
        $inquiries = [];

        if ($filterId) {
            $inquiries = self::getCollection($_REQUEST);
        }
        ?>
        <div class="wrap" data-component="wcsf-inquire">
            <h1><?php echo get_admin_page_title(); ?></h1>
            <form action="?page=woocommerce-step-filter-inquiries" method="get">
                <input type="hidden" name="post_type" value="wcsf">
                <input type="hidden" name="page"
                    value="<?php echo isset($_GET['paged']) ? $_GET['paged'] : 'woocommerce-step-filter-inquiries'; ?>">
                <table class="form-table">
                    <tbody>
                        <tr>
                            <th scope="row"><label for="wcsf-filter-id"><?php L10N::e('Filter'); ?></label></th>
                            <td>
                                <select name="filter-id" id="wcsf-filter-id"><?php
                                    foreach ($filters as $id => $name) {
                                        echo '<option value="' . esc_attr($id) . '" '
                                            . selected($filterId, $id, false) . '>'
                                            . esc_html($name) . '</option>';
                                    }
                                    ?></select>
                            </td>
                        </tr>
                        <tr>
                            <th scope="row">
                                <label for="wcsf-date-from"><?php L10N::e('Date from'); ?></label>
                            </th>
                            <td>
                                <input type="date" id="wcsf-date-from" name="date-from" value="<?php
                                    echo esc_attr(isset($_REQUEST['date-from']) ? $_REQUEST['date-from'] : '');
                                    ?>">
                            </td>
                        </tr>
                        <tr>
                            <th scope="row">
                                <label for="wcsf-date-to"><?php L10N::e('Date to'); ?></label>
                            </th>
                            <td>
                                <input type="date" id="wcsf-date-to" name="date-to" value="<?php
                                    echo esc_attr(isset($_REQUEST['date-to']) ? $_REQUEST['date-to'] : '');
                                    ?>">
                            </td>
                        </tr>
                    </tbody>
                </table>
                <p class="submit">
                    <button class="button button-primary" type="submit"><?php L10N::e('Get inquiries'); ?></button>
                </p>
                <?php if ($filterId) { ?>
                    <p><?php printf(L10N::r('Records found: %d'), count($inquiries)); ?></p>
                <?php } ?>
            </form>
            <?php
            if ($filterId) {
                $this->outputStatistic($filterId, self::getStatistic($filterId, $inquiries));
                $this->outputList($filterId, $inquiries);
            }
            ?>
        </div>
        <?php
    }

    /**
     * Output list HTML
     *
     * @param integer $filterId
     * @param array $inquiries
     */
    public function outputList($filterId, $inquiries)
    {
        if (empty($inquiries)) {
            return;
        }
        ?>
        <section class="wcsf-inquiry-list" data-component="wcsf-inquiry-list">
            <h2 class="wcsf-inquiry-list-title">
                <?php L10N::e('Inquires list'); ?>
                <small><a href="?<?php
                    echo $_SERVER['QUERY_STRING'];
                    echo esc_attr('&wcsf-download-inquiries-list=' . $filterId);
                    ?>"><?php echo L10N::r('Download CSV'); ?></a></small>
            </h2>
            <?php
            $this->outputListTable(
                $filterId,
                array_slice($inquiries, 0, apply_filters('wcsf_inquiries_list_per_page', 20)),
                count($inquiries)
            );
            ?>
            <form action="#" method="post">
                <?php wp_nonce_field(basename(__FILE__), 'woocommerce-step-filter'); ?>
                <table class="wp-list-table widefat">
                    <tr>
                        <td>
                            <button class="button" data-component="wcsf-inquiries-purge"
                                data-confirm-message="<?php
                                echo esc_attr(L10N::r('Purge statistic for the selected filter and date range?'));
                                ?>"
                                name="purge-filter-inquiries" value="<?php echo esc_attr($filterId); ?>"><?php
                                L10N::e('Purge statistic');
                                ?></button>
                            <input type="hidden" name="filter-id" value="<?php echo esc_attr($filterId); ?>">
                            <input type="hidden" name="date-from" value="<?php
                                echo esc_attr(isset($_REQUEST['date-from']) ? $_REQUEST['date-from'] : '');
                                ?>">
                            <input type="hidden" name="date-to" value="<?php
                                echo esc_attr(isset($_REQUEST['date-to']) ? $_REQUEST['date-to'] : '');
                                ?>">
                        </td>
                    </tr>
                </table>
            </form>
        </section>
        <?php
    }

    /**
     * Output list table HTML
     *
     * @param integer $filterId
     * @param array $inquiries - collection
     * @param integer $total
     */
    public function outputListTable($filterId, $inquiries, $total = 0)
    {
        $questions = Filter::getQuestionsList($filterId);
        $pagination = self::getListPagination([
            'paged' => isset($_REQUEST['paged']) ? (int) $_REQUEST['paged'] : 1,
            'total' => ceil($total / apply_filters('wcsf_inquiries_list_per_page', 20))
        ]);
        ?>
        <table class="wcsf-inquiry-list-table wp-list-table widefat striped"
            data-component="wcsf-inquiry-list-table"
            data-filter-id="<?php echo esc_attr($filterId); ?>"
            data-total="<?php echo esc_attr($total); ?>"
            data-ajax-url="<?php echo admin_url('admin-ajax.php'); ?>">
            <thead>
                <tr>
                    <th><?php L10N::e('ID'); ?></th>
                    <th><?php L10N::e('Date'); ?></th>
                    <th><?php L10N::e('Results count'); ?></th>
                    <th><?php L10N::e('Values'); ?></th>
                </tr>
            </thead>
            <tbody>
                <?php foreach ($inquiries as $inquiry) { ?>
                    <tr class="wcsf-inquiry-list-item">
                        <td><?php
                            echo $inquiry->link ? '<a href="' . esc_url($inquiry->link) . '" target="_blank">' : '';
                            echo wp_kses_post($inquiry->id);
                            echo $inquiry->link ? '</a>' : '';
                            ?></td>
                        <td><?php echo wp_kses_post($inquiry->date); ?></td>
                        <td>
                            <a target="_blank"
                                href="<?php
                                echo get_admin_url(
                                    null,
                                    'edit.php?post_type=product&wcsf_ids=' . implode(',', $inquiry->resultsIds)
                                );
                                ?>"><?php echo wp_kses_post($inquiry->resultsCount); ?></a>
                        </td>
                        <td>
                            <dl class="wcsf-inquiry-list-item-data">
                                <?php
                                if (!empty($inquiry->values)) {
                                    foreach ($inquiry->values as $questionId => $values) {
                                        if (!isset($questions[$questionId])) {
                                            continue;
                                        }

                                        $output = [];
                                        $question = $questions[$questionId];
                                        $isNumericView = strpos($question->viewType, 'number') !== false;
                                        $isMultiChoiceView = strpos($question->viewType, 'multi-choice') !== false;

                                        echo '<div class="wcsf-inquiry-list-item-data-item">';
                                        echo '<dt class="wcsf-inquiry-list-item-data-item-name">'
                                            . wp_kses_post($question->name) . ':</dt> ';

                                        foreach ($values as $key => $value) {
                                            foreach (Question::getFields($filterId, $questionId) as $fieldKey => $field) {
                                                if ((!$isNumericView && !$isMultiChoiceView && $field->value != $value)
                                                    || ($isNumericView && $key != $fieldKey)
                                                    || ($isMultiChoiceView && $field->value != $key)
                                                ) {
                                                    continue;
                                                }

                                                $valueOutput = $field->name;

                                                if ($isNumericView) {
                                                    $valueOutput = $value;
                                                } elseif ($isMultiChoiceView) {
                                                    $valueOutput = "<b>$field->name" . '</b>: <span>'
                                                        . self::$multiChoicePossibleValues[$value] . '</span>';
                                                }

                                                $output[] = '<dd class="wcsf-inquiry-list-item-data-item-value">'
                                                    . wp_kses_post($valueOutput) . '</dd>';
                                            }
                                        }

                                        if ($isNumericView) {
                                            echo implode(' - ', $output);
                                        } else {
                                            echo implode(', ', $output);
                                        }

                                        echo '</div>';
                                    }
                                }
                                ?>
                            </dl>
                        </td>
                    </tr>
                <?php } ?>
            </tbody>
            <?php if (!empty($pagination)) { ?>
                <tfoot>
                    <tr>
                        <td colspan="4">
                            <nav class="wcsf-inquiry-list-pagination">
                                <ul class="wcsf-inquiry-list-pagination-list pagination"
                                    data-component="wcsf-inquiry-list-pagination">
                                    <?php foreach ($pagination as $item) { ?>
                                        <li class="<?php
                                        echo esc_attr(trim("wcsf-inquiry-list-pagination-list-item {$item['class']}"));
                                        ?>"><?php echo $item['innerHtml']; ?></li>
                                    <?php } ?>
                                </ul>
                            </nav>
                        </td>
                    </tr>
                </tfoot>
            <?php } ?>
        </table>
        <?php
    }

    /** Output list table HTML via Ajax */
    public function outputListTableAjax()
    {
        $filterId = isset($_REQUEST['filter-id']) ? (int) $_REQUEST['filter-id'] : null;
        $total = isset($_REQUEST['total']) ? (int) $_REQUEST['total'] : null;
        $args = $_REQUEST;
        $args['numberposts'] = apply_filters('wcsf_inquiries_list_per_page', 20);
        $this->outputListTable($filterId, self::getCollection($args), $total);

        exit;
    }

    /**
     * Get pagination items array
     *
     * @param array $args
     *
     * @return array
     */
    public static function getListPagination($args)
    {
        $defaults = [
            'paged' => 1,
            'total' => 10
        ];

        $args = array_merge($defaults, $args);
        $output = [];
        $paginationArgs = apply_filters(
            'wcsf_inquiries_list_pagination_args',
            [
                'format' => '?paged=%#%',
                'base' => '%_%',
                'total' => $args['total'],
                'current' => $args['paged'],
                'show_all' => false,
                'end_size' => 1,
                'mid_size' => 2,
                'prev_next' => true,
                'prev_text' => L10N::r('« Previous'),
                'next_text' => L10N::r('Next »'),
                'type' => 'array'
            ],
            $args
        );

        $links = paginate_links($paginationArgs);

        foreach ((array) $links as $link) {
            // add custom classes
            $link = str_replace(['page-numbers', 'current'], ['page-numbers button', 'current active'], $link);

            // replace empty href
            $link = str_replace('href=""', 'href="?paged=1"', $link);
            $link = str_replace("href=''", 'href="?paged=1"', $link);

            $output[] = [
                'class' => strpos($link, 'current') !== false ? 'active' : '',
                'innerHtml' => $link
            ];
        }

        return apply_filters('wcsf_inquiries_list_pagination', $output, $args);
    }

    /**
     * Output questions statistic HTML
     *
     * @param integer $filterId
     * @param array $statistic - @see getStatistic
     */
    public function outputStatistic($filterId, $statistic)
    {
        if (empty($statistic)) {
            return;
        }
        ?>
        <section class="wcsf-inquiry-questions">
            <h2 class="wcsf-inquiry-questions-title">
                <?php echo L10N::r('Inquiries statistic'); ?>
                <small><a href="?<?php
                    echo $_SERVER['QUERY_STRING'];
                    echo esc_attr('&wcsf-download-inquiries-statistic=' . $filterId);
                    ?>"><?php echo L10N::r('Download CSV'); ?></a></small>
            </h2>
            <?php foreach ($statistic as $question) { ?>
                <article class="wcsf-inquiry-question">
                    <h3 class="wcsf-inquiry-question-title"><?php echo wp_kses_post($question->name); ?></h3>
                    <div class="wcsf-inquiry-question-fields-wrapper">
                        <div class="wcsf-inquiry-question-fields">
                            <?php foreach ($question->values as $value) { ?>
                                <div class="wcsf-inquiry-question-fields-item">
                                    <div class="wcsf-inquiry-question-fields-item-inner"
                                        style="<?php echo esc_attr("height:{$value->percent}%;"); ?>">
                                        <div class="wcsf-inquiry-question-fields-item-content"
                                            title="<?php echo esc_attr("{$value->name} ({$value->count})"); ?>">
                                            <span class="wcsf-inquiry-question-fields-item-name"><?php
                                                echo wp_kses_post($value->name);
                                                ?></span>
                                            /
                                            <span class="wcsf-inquiry-question-fields-item-percent"><?php
                                                echo wp_kses_post(round($value->percent, 2));
                                                ?>%</span>
                                            <small class="wcsf-inquiry-question-fields-item-count"><?php
                                                echo wp_kses_post("({$value->count})");
                                                ?></small>
                                        </div>
                                    </div>
                                </div>
                            <?php } ?>
                        </div>
                    </div>
                </article>
            <?php } ?>
        </section>
        <?php
    }
    // </editor-fold>

    // <editor-fold desc="Get">
    /**
     * Return inquiry data object
     *
     * @param object $inquiry - WP query post
     *
     * @return object
     */
    public static function getData($inquiry)
    {
        $output = (object) [
            'id' => (int) $inquiry->ID,
            'link' => (string) $inquiry->post_title,
            'date' => $inquiry->post_date,
            'filterId' => (int) $inquiry->post_parent,
            'resultsCount' => (int) $inquiry->menu_order,
            'resultsIds' => (array) unserialize($inquiry->post_excerpt),
            'values' => (array) unserialize($inquiry->post_content)
        ];

        return apply_filters('wcsf_inquiry_data', $output, $inquiry);
    }

    /**
     * Return collection of inquires
     *
     * @param array $args
     *
     * @return array
     */
    public static function getCollection($args)
    {
        $defaults = [
            'filter-id' => null,
            'date-from' => null,
            'date-to' => null,
            'paged' => isset($_REQUEST['paged']) ? (int) $_REQUEST['paged'] : 1,
            'numberposts' => -1,
            'total' => null
        ];

        $args = array_replace($defaults, $args);

        $queryArgs = [
            'post_parent' => $args['filter-id'],
            'paged' => $args['paged'],
            'total' => $args['total'],
            'numberposts' => $args['numberposts']
        ];

        if (isset($args['date-from']) && $args['date-from']) {
            $queryArgs['date_query'][] = [
                'inclusive' => true,
                'after' => $args['date-from']
            ];
        }

        if (isset($args['date-to']) && $args['date-to']) {
            $queryArgs['date_query'][] = [
                'inclusive' => true,
                'before' => $args['date-to']
            ];
        }

        return apply_filters('wcsf_inquires_collection', self::getList($queryArgs), $args);
    }

    /**
     * Return array of inquires from DB
     *
     * @param array $args
     *
     * @return array
     */
    public static function getList($args)
    {
        $defaults = [
            'post_type' => self::$postTypeName,
            'post_status' => 'any',
            'numberposts' => 30
        ];

        $args = array_replace($defaults, $args);
        $inquires = get_posts($args);
        $output = [];

        foreach ($inquires as $inquiry) {
            $output[$inquiry->ID] = self::getData($inquiry);
        }

        return apply_filters('wcsf_inquires_list', $output, $args);
    }

    /**
     * Return array of inquires ready to CSV
     *
     * @param integer $filterId
     * @param array $args
     *
     * @return array
     */
    public static function getListCSV($filterId, $args = [])
    {
        $questions = Filter::getQuestionsList($filterId);
        $inquiries = self::getCollection($args);
        $output = [
            [
                'ID',
                'Date',
                'Results',
                'Values',
                'Link'
            ]
        ];

        foreach ($inquiries as $inquiry) {
            $valueRow = [];

            if (!empty($inquiry->values)) {
                foreach ($inquiry->values as $questionId => $values) {
                    if (!isset($questions[$questionId])) {
                        continue;
                    }

                    $question = $questions[$questionId];
                    $isNumericView = strpos($question->viewType, 'number') !== false;
                    $isMultiChoiceView = strpos($question->viewType, 'multi-choice') !== false;
                    $itemRow = [];

                    foreach ($values as $key => $value) {
                        foreach (Question::getFields($filterId, $questionId) as $fieldKey => $field) {
                            if ((!$isNumericView && !$isMultiChoiceView && $field->value != $value)
                                || ($isNumericView && $key != $fieldKey)
                                || ($isMultiChoiceView && $field->value != $key)
                            ) {
                                continue;
                            }

                            $valueOutput = $field->name;

                            if ($isNumericView) {
                                $valueOutput = $value;
                            } elseif ($isMultiChoiceView) {
                                $valueOutput = '[' . $field->name . ':' . self::$multiChoicePossibleValues[$value] . ']';
                            }

                            if (!empty($valueOutput)) {
                                $itemRow[] = $valueOutput;
                            }
                        }
                    }

                    if (!empty($itemRow)) {
                        if ($isNumericView) {
                            $valueRow[] = $question->name . ':' . implode(' - ', $itemRow);
                        } else {
                            $valueRow[] = $question->name . ':' . implode(' & ', $itemRow);
                        }
                    }
                }
            }

            $output[] = [
                $inquiry->id,
                $inquiry->date,
                $inquiry->resultsCount,
                implode(', ', $valueRow),
                $inquiry->link
            ];
        }

        return apply_filters('wcsf_inquires_list_csv', $output, $filterId, $args);
    }

    /**
     * Returns object of the filter's statistic data
     *
     * @param integer $filterId
     * @param array $inquiries - collection of inquiries
     *
     * @return array
     */
    public static function getStatistic($filterId, $inquiries = [])
    {
        $output = [];

        foreach (Filter::getQuestionsList($filterId) as $question) {
            $fields = Question::getFields($filterId, $question->id);
            $isNumericView = strpos($question->viewType, 'number') !== false;
            $isMultiChoiceView = strpos($question->viewType, 'multi-choice') !== false;
            $outputItems = [];

            if ($isNumericView) {
                foreach ($fields as $key => $field) {
                    $outputItems[$question->id . '-' . $key] = (object) [
                        'name' => $question->name . ': ' . $field->label,
                        'values' => []
                    ];
                }
            } elseif ($isMultiChoiceView) {
                foreach ($fields as $field) {
                    $outputItems[$question->id . '-' . $field->name] = (object) [
                        'name' => $question->name . ': ' . $field->name,
                        'values' => []
                    ];
                }
            } else {
                $outputItems[$question->id] = (object) [
                    'name' => $question->name,
                    'values' => []
                ];

                // fill by all exist fields
                foreach ($fields as $field) {
                    $outputItems[$question->id]->values[$field->value] = 0;
                }
            }

            foreach ($fields as $field) {
                foreach ($inquiries as $inquiry) {
                    // have no this question in the inquiry answer
                    if (!isset($inquiry->values[$question->id])) {
                        continue;
                    }

                    $inquiryValues = $inquiry->values[$question->id];

                    if ($isNumericView) {
                        if (is_array($inquiryValues)) {
                            foreach ($inquiryValues as $key => $inquiryValue) {
                                if (!isset($outputItems[$question->id . '-' . $key]->values[$inquiryValue])) {
                                    $outputItems[$question->id . '-' . $key]->values[$inquiryValue] = 0;
                                }

                                $outputItems[$question->id . '-' . $key]->values[$inquiryValue]++;
                            }
                        }
                    } elseif ($isMultiChoiceView) {
                        if (!isset($inquiryValues[$field->value])) {
                            continue;
                        }

                        foreach (self::$multiChoicePossibleValues as $key => $label) {
                            if (!isset($outputItems[$question->id . '-' . $field->name]->values[$label])) {
                                $outputItems[$question->id . '-' . $field->name]->values[$label] = 0;
                            }
                        }

                        $label = self::$multiChoicePossibleValues[$inquiryValues[$field->value]];
                        $outputItems[$question->id . '-' . $field->name]->values[$label]++;
                    } elseif (in_array($field->value, $inquiryValues)) {
                        $outputItems[$question->id]->values[$field->value]++;
                    }
                }
            }

            foreach ($outputItems as $key => &$outputItem) {
                if (!property_exists($outputItem, 'name')) {
                    unset($outputItems[$key]);

                    continue;
                }

                // add other properties
                $outputItem->fields = $fields;
                $outputItem->totalValuesCount = array_sum($outputItem->values);

                foreach ($outputItem->values as $value => $count) {
                    $name = $value;
                    $percent = $outputItem->totalValuesCount != 0
                        ? ($count / $outputItem->totalValuesCount) * 100
                        : 0;

                    foreach ($outputItem->fields as $field) {
                        if ($field->value === $value) {
                            $name = $field->name;

                            break;
                        }
                    }

                    $outputItem->values[$value] = (object) [
                        'count' => $count,
                        'name' => $name,
                        'percent' => $percent
                    ];
                }

                // sort numeric questions
                if ($isNumericView) {
                    ksort($outputItem->values);
                }

                unset($outputItem);
            }

            $output = array_merge($output, $outputItems);
        }

        return apply_filters('wcsf_inquires_statistic', $output, $filterId, $inquiries);
    }

    /**
     * Return array of inquires statistic ready to CSV
     *
     * @param integer $filterId
     * @param array $args
     *
     * @return array
     */
    public static function getStatisticCSV($filterId, $args = [])
    {
        $inquiries = self::getCollection($args);
        $statistic = self::getStatistic($filterId, $inquiries);
        $output = [
            [
                'Question',
                'Values'
            ]
        ];

        foreach ($statistic as $question) {
            $valueRow = [];

            foreach ($question->values as $value) {
                $valueRow[] = '[' . $value->name . ': ' . round($value->percent, 2) . '% (' . $value->count . ')]';
            }

            $output[] = [
                $question->name,
                implode(', ', $valueRow)
            ];
        }

        return apply_filters('wcsf_inquires_statistic_csv', $output, $filterId, $args);
    }

    /**
     * Returns array of ids and titles of posts
     *
     * @param array $args
     *
     * @return array
     */
    public static function getPostsIdsList($args = [])
    {
        $defaults = [
            'post_type' => self::$postTypeName,
            'numberposts' => -1,
            'post_status' => 'publish'
        ];

        $args = array_replace($defaults, $args);
        $output = [];

        foreach (get_posts($args) as $post) {
            $output[$post->ID] = $post->post_title;
        }

        return $output;
    }
    // </editor-fold>
}
