<?php
namespace WCStepFilter;

/**
 * Question Class
 *
 * @class Question
 * @version 10.0.0
 */
class Question implements PostTypeInterface, TaxonomyInterface
{
    use PostTypeTrait;
    use TaxonomyTrait;

    // <editor-fold desc="Properties">
    /**
     * Current working question ID
     * @var string
     */
    protected $currentId = null;

    /**
     * Post type string
     * @var string
     */
    public static $postType = 'wcsf_question';

    /**
     * Taxonomy string
     * @var string
     */
    public static $taxonomy = 'wcsf_question_cat';

    /**
     * Namespace string
     * @var string
     */
    public static $namespace = 'wcsf_question';
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Class Constructor */
    public function __construct()
    {
        $this->initPostType();
        $this->initTaxonomy();

        // admin
        add_action('admin_enqueue_scripts', [$this, 'enqueueAdmin']);
        add_action('admin_menu', [$this, 'registerCategoriesPageMenuItem']);
        add_action('wp_ajax_wcsfAddQuestionValue', [$this, 'addQuestionValueAjax']);
        add_action('wp_ajax_wcsfGetQuestionValuesSelectOptions', [$this, 'getValuesSelectOptionsAjax']);
    }

    /**
     * Get current working question ID
     * 
     * @return string
     */
    public function getCurrentId()
    {
        return $this->currentId;
    }

    /**
     * Set current working question ID
     * 
     * @param string $currentId
     */
    public function setCurrentId($currentId)
    {
        $this->currentId = $currentId;
    }

    /** Register entity post type */
    public function registerPostType()
    {
        register_post_type(
            self::$postType,
            [
                'label' => L10N::r('WooCommerce Step Filter Questions'),
                'labels' => [
                    'name' => L10N::r('WooCommerce Step Filter Questions'),
                    'singular_name' => L10N::r('WooCommerce Step Filter Question'),
                    'menu_name' => L10N::r('Question'),
                    'all_items' => L10N::r('Questions')
                ],
                'description' => L10N::r('This is where you can add new questions for step filter'),
                'public' => false,
                'show_ui' => true,
                'map_meta_cap' => true,
                'publicly_queryable' => false,
                'exclude_from_search' => true,
                'show_in_menu' => 'edit.php?post_type=' . Filter::$postType,
                'hierarchical' => false,
                'rewrite' => false,
                'query_var' => false,
                'supports' => [
                    'title',
                    'thumbnail',
                    'editor',
                    'page-attributes'
                ],
                'show_in_nav_menus' => false,
                'show_in_admin_bar' => true
            ]
        );
    }

    /** Register entity taxonomy */
    public function registerTaxonomy()
    {
        register_taxonomy(
            self::$taxonomy,
            [self::$postType],
            [
                'label' => L10N::r('WooCommerce Step Filter Question Categories'),
                'labels' => [
                    'name' => L10N::r('WooCommerce Step Filter Question Categories'),
                    'singular_name' => L10N::r('WooCommerce Step Filter Question Category'),
                    'menu_name' => L10N::r('Question Category'),
                    'all_items' => L10N::r('Question Categories')
                ],
                'hierarchical' => true,
                'public' => true,
                'show_ui' => true,
                'show_admin_column' => true,
                'show_in_nav_menus' => false,
                'show_in_menu' => true,
                'query_var' => true,
                'rewrite' => false
            ]
        );
    }
    // </editor-fold>

    // <editor-fold desc="Admin">
    /** Register settings page menu item action */
    public function registerCategoriesPageMenuItem()
    {
        add_submenu_page(
            'edit.php?post_type=' . Filter::$postType,
            L10N::r('Question Categories'),
            L10N::r('Question Categories'),
            'edit_pages',
            'edit-tags.php?taxonomy=' . self::$taxonomy
        );
    }

    /** Admin styles and scripts enqueue */
    public function enqueueAdmin()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';

        wp_enqueue_script(
            'woocommerce-step-filter-question',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/question$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        if (defined('WC_VERSION')
            && ((isset($_GET['post']) && get_post_type((int) $_GET['post']) == self::$postType)
                || (isset($_GET['post_type']) && $_GET['post_type'] == self::$postType))
        ) {
            wp_register_script(
                'select2',
                \WC()->plugin_url() . "/assets/js/select2/select2.full$suffix.js",
                ['jquery'],
                '4.0.3'
            );

            wp_register_script(
                'selectWoo',
                WC()->plugin_url() . "/assets/js/selectWoo/selectWoo.full$suffix.js",
                ['jquery'],
                '1.0.0'
            );

            wp_register_script(
                'wc-enhanced-select',
                WC()->plugin_url() . "/assets/js/admin/wc-enhanced-select$suffix.js",
                ['jquery', 'selectWoo'],
                WC_VERSION
            );

            wp_enqueue_script('select2');
            wp_enqueue_script('selectWoo');
            wp_enqueue_script('wc-enhanced-select');
            wp_enqueue_style('woocommerce_admin_styles', \WC()->plugin_url() . '/assets/css/admin.css', [], WC_VERSION);
        }
    }

    /**
     * Post meta-box view
     *
     * @param \WP_Post $post
     */
    public function metaBoxCallback($post)
    {
        global $_wp_additional_image_sizes;

        $manualValues = self::getSetting($post->ID, 'manual_values');
        $attributes = wc_get_attribute_taxonomies();
        $settingsModel = self::getSettingsModel();

        // unset settings model for custom output
        unset($settingsModel['manual_values']);

        // blend in required values
        $settingsModel['bond_questions']['values'] = self::getPostsIds(['exclude' => $post->ID]);
        $imageSizes = [];

        foreach ($_wp_additional_image_sizes as $imageSizeName => $imageSize) {
            $crop = $imageSize['crop'] ? ' crop' : '';
            $imageSizes[] = $imageSizeName != "{$imageSize['width']}x{$imageSize['height']}$crop"
                ? "$imageSizeName ({$imageSize['width']}x{$imageSize['height']}$crop)"
                : $imageSizeName;
        }

        $settingsModel['thumbnail_size']['description'] .= '<details><summary>'
            . L10N::r('More') . '</summary>'
            . implode(', ', $imageSizes) . '</details>';

        $settingsModel['dependencies']['values']['question']['values'] = array_replace(
            ['' => ''],
            $settingsModel['bond_questions']['values']
        );

        $settingsModel['category_parent']['values'] = array_replace(
            [0 => ''],
            Utils::getTermsTree(
                ['taxonomy' => 'product_cat'],
                [
                    'flatten' => true,
                    'showLevelInName' => true,
                    'nameOnly' => true
                ]
            )
        );

        $settingsModel['category_parent_by_question']['values'] = array_replace(
            [0 => ''],
            $settingsModel['bond_questions']['values']
        );

        foreach ($attributes as $attribute) {
            $settingsModel['attribute']['values'][$attribute->attribute_id] = $attribute->attribute_label;
        }

        $attributeKeys = array_keys($settingsModel['attribute']['values']);
        $settingsModel['attribute']['default'] = reset($attributeKeys);
        ?>
        <div class="wcsf-question"
            data-component="wcsf-question"
            data-post-id="<?php echo esc_attr($post->ID); ?>"
            data-ajax-url="<?php echo admin_url('admin-ajax.php'); ?>">
            <table class="wcsf-settings-table form-table wcsf-data-table">
                <?php foreach ($settingsModel as $settingKey => $setting) { ?>
                    <tr class="wcsf-settings-table-row form-field<?php
                        echo isset($setting['separate']) && $setting['separate'] ? ' separate' : '';
                        ?>"
                        data-component="wcsf-setting" data-key="<?php echo esc_attr($settingKey); ?>">
                        <th scope="row" class="wcsf-settings-table-row-name">
                            <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                                echo wp_kses_post($setting['label']);
                                ?></label>
                        </th>
                        <td class="wcsf-settings-table-row-value"><?php
                            Admin::settingFieldView(
                                $setting,
                                ['values' => [$setting['key'] => self::getSetting($post->ID, $settingKey)]]
                            );
                            ?></td>
                    </tr>
                <?php } ?>
                <tr class="wcsf-question-manual" data-component="wcsf-setting" data-key="manual">
                    <th><?php L10N::e('Manual values'); ?></th>
                    <td>
                        <button class="button-primary"
                            data-component="wcsf-question-add-value"><?php L10N::e('Add a value'); ?></button>
                        <br><br>
                        <div class="wcsf-question-manual-values">
                            <table class="wcsf-question-manual-values-table wp-list-table widefat striped">
                                <thead class="wcsf-question-manual-values-table-heading">
                                    <tr class="wcsf-question-manual-values-table-heading-row">
                                        <th class="wcsf-question-manual-values-table-heading-cell is-name"><?php
                                            L10N::e('Name');
                                            ?></th>
                                        <th class="wcsf-question-manual-values-table-heading-cell is-settings"><?php
                                            L10N::e('Settings');
                                            ?></th>
                                        <th class="wcsf-question-manual-values-table-heading-cell is-remove"><?php
                                            L10N::e('Remove');
                                            ?></th>
                                    </tr>
                                </thead>
                                <tbody data-component="wcsf-question-manual-values">
                                    <?php
                                    if (is_array($manualValues)) {
                                        foreach ($manualValues as $valueKey => $value) {
                                            $this->renderManualQuestionValue(
                                                array_replace(
                                                    (array) $value,
                                                    [
                                                        'postId' => $post->ID,
                                                        'index' => $valueKey
                                                    ]
                                                )
                                            );
                                        }
                                    } else {
                                        $this->renderManualQuestionValue(['postId' => $post->ID]);
                                    }
                                    ?>
                                </tbody>
                            </table>
                        </div>
                    </td>
                </tr>
            </table>
        </div>
        <?php
    }

    /** Return manual value html */
    public function addQuestionValueAjax()
    {
        $this->renderManualQuestionValue($_GET);

        exit();
    }

    /**
     * Return table of manual values html
     *
     * @param array $args
     */
    public function renderManualQuestionValue($args = [])
    {
        $defaults = [
            'postId' => 0,
            'index' => 0,
            'name' => '',
            'thumbnail' => '',
            'pre-checked' => 0,
            'dependencies' => [],
            'filter-by' => 'any'
        ];

        $args = array_replace($defaults, $args);
        $settingsModel = self::getSettingsModel();
        $dependenciesValuesModel = $settingsModel['dependencies']['values'];
        $dependenciesValuesModel['question']['values'] = array_replace(
            ['' => ''],
            self::getPostsIds(['exclude' => [$args['postId']]])
        );
        ?>
        <tr data-component="wcsf-question-manual-value" data-index="<?php echo esc_attr($args['index']); ?>">
            <td>
                <input type="text"
                    name="_manual_values[<?php echo esc_attr($args['index']); ?>][name]"
                    class="regular-text wcsf-question-manual-values-item-name"
                    placeholder="<?php echo esc_attr(L10N::r('Some value here')); ?>"
                    value="<?php echo esc_attr($args['name']); ?>">
            </td>
            <td>
                <a href="#wcsf-data-table-modal-settings-<?php echo esc_attr($args['index']); ?>"
                    class="button-primary wcsf-data-table-item-open-modal"
                    data-component="wcsf-open-modal"><?php L10N::e('Settings'); ?></a>
                <div id="wcsf-data-table-modal-settings-<?php echo esc_attr($args['index']); ?>"
                    class="wcsf-modal"
                    data-component="wcsf-modal">
                    <div class="wcsf-modal-dialog">
                        <a href="#close" title="<?php echo esc_attr(L10N::r('Close')); ?>"
                            class="wcsf-modal-close" data-component="wcsf-modal-close">&times;</a>
                        <div class="wcsf-modal-dialog-body">
                            <table class="wcsf-settings-table form-table wp-list-table striped">
                                <tr>
                                    <th scope="row"><label for="<?php
                                        echo esc_attr("_manual_values[{$args['index']}][thumbnail]");
                                        ?>"><?php
                                            L10N::e('Thumbnail');
                                            ?></label></th>
                                    <td>
                                        <?php
                                        Admin::settingFieldView(
                                            [
                                                'label' => L10N::r('Thumbnail'),
                                                'key' => esc_attr("_manual_values[{$args['index']}][thumbnail]"),
                                                'type' => 'thumbnail',
                                                'default' => ''
                                            ],
                                            [
                                                'values' => [
                                                    "_manual_values[{$args['index']}][thumbnail]" => $args['thumbnail']
                                                ]
                                            ]
                                        );
                                        ?>
                                    </td>
                                </tr>
                                <tr>
                                    <th scope="row"><label for="_manual_values[<?php
                                        echo esc_attr($args['index']);
                                        ?>][pre-checked]"><?php L10N::e('Pre-checked'); ?></label></th>
                                    <td>
                                        <input type="hidden"
                                            name="_manual_values[<?php echo esc_attr($args['index']); ?>][pre-checked]"
                                            value="0">
                                        <input type="checkbox"
                                            id="_manual_values[<?php echo esc_attr($args['index']); ?>][pre-checked]"
                                            name="_manual_values[<?php echo esc_attr($args['index']); ?>][pre-checked]"
                                            class="regular-text wcsf-question-manual-values-item-pre-checked"
                                            value="1" <?php checked($args['pre-checked']); ?>>
                                    </td>
                                </tr>
                                <tr>
                                    <td colspan="2">
                                        <strong><?php echo L10N::r('Sources'); ?></strong>
                                        <ul>
                                            <?php
                                            $this->renderAnyProductField($args);
                                            $this->renderProductSelectFields($args);
                                            $this->renderTermsSelectFields($args);
                                            ?>
                                        </ul>
                                    </td>
                                </tr>
                                <tr>
                                    <td colspan="2">
                                        <strong><?php echo L10N::r('Dependencies'); ?></strong><br><br>
                                        <?php
                                        // phpcs:disable
                                        Admin::settingFieldView(
                                            [
                                                'label' => L10N::r('Dependencies'),
                                                'key' => esc_attr("_manual_values[{$args['index']}][dependencies]"),
                                                'type' => 'data-table',
                                                'values' => $dependenciesValuesModel,
                                                'default' => [],
                                                'description' => L10N::r('The value will be shown only in the case of the meeting of a dependency')
                                            ],
                                            ['values' => [esc_attr("_manual_values[{$args['index']}][dependencies]") => $args['dependencies']]]
                                        );
                                        // phpcs:enable
                                        ?>
                                    </td>
                                </tr>
                            </table>
                        </div>
                        <div class="wcsf-modal-dialog-footer">
                            <a href="#close" class="button button-primary" data-component="wcsf-modal-close"><?php
                                L10N::e('Save');
                                ?></a>
                        </div>
                    </div>
                </div>
            </td>
            <td>
                <button class="button-secondary"
                    title="<?php echo esc_attr(L10N::r('Remove')); ?>"
                    data-component="wcsf-question-manual-value-remove"
                    data-index="<?php echo esc_attr($args['index']); ?>">&times;</button>
            </td>
        </tr>
        <?php
    }

    /**
     * Get list of an option values
     *
     * @param integer|array $id
     *
     * @return array
     */
    public function getValuesSelectOptions($id)
    {
        $output = [];
        $filterType = self::getSetting($id, 'filter_type');
        $viewType = self::getSetting($id, 'view_type');

        if ($filterType == 'price' || $viewType == 'formula') {
            $output = null;
        } elseif ($filterType == 'manual') {
            foreach ((array) self::getSetting($id, 'manual_values') as $key => $value) {
                $output[$key] = $value['name'];
            }
        } elseif ($filterType == 'attribute') {
            $taxonomy = wc_attribute_taxonomy_name_by_id((int) self::getSetting($id, 'attribute'));
            $valueSource = self::getSetting($id, 'attribute_value_source');
            $terms = get_terms([
                'taxonomy' => $taxonomy,
                'hide_empty' => false
            ]);

            foreach ($terms as $term) {
                switch ($valueSource) {
                    default:
                    case 'name':
                        $termValue = (float) $term->name;
                        break;

                    case 'slug':
                        $termValue = (float) $term->slug;
                }

                $key = strpos($viewType, 'number') !== false && $filterType == 'formula'
                    ? $termValue
                    : $term->term_id;

                $output[$key] = $term->name;
            }
        } elseif (in_array($filterType, ['tag', 'category'])) {
            $taxonomy = $filterType == 'tag' ? 'product_tag' : 'product_cat';
            $terms = Utils::getTermsTree(
                ['taxonomy' => $taxonomy],
                [
                    'flatten' => true,
                    'showLevelInName' => true
                ]
            );

            foreach ($terms as $term) {
                $output[$term->term_id] = $term->name;
            }
        }

        if (in_array($viewType, ['select', 'radio', 'image-radio', 'chips-radio', 'button-radio'])
            && self::getSetting($id, 'add_any_product_value')
        ) {
            $output[0] = L10N::r('Doesn\'t matter');
        }

        if (is_array($output)) {
            asort($output, SORT_NUMERIC);
        }

        return $output;
    }

    /** Get list of an option values via Ajax */
    public function getValuesSelectOptionsAjax()
    {
        Utils::sendJSON($this->getValuesSelectOptions((int) $_REQUEST['value']));
    }

    /**
     * Return "any product" radio html
     *
     * @param array $args
     */
    public function renderAnyProductField($args = [])
    {
        $defaults = [
            'index' => 0,
            'filter-by' => 'category'
        ];

        $args = array_replace($defaults, $args);
        $active = $args['filter-by'] == 'any';
        ?>
        <li class="wcsf-question-manual-value-filter<?php echo !$active ? ' is-closed' : ''; ?>"
            data-component="wcsf-question-manual-value-filter"
            data-id="any">
            <label class="wcsf-question-manual-value-filter-label">
                <input type="radio" data-component="wcsf-question-manual-value-filter-by"
                    data-id="any" name="_manual_values[<?php echo esc_attr($args['index']); ?>][filter-by]"
                    value="any" <?php checked($active); ?>><?php L10N::e('Any product'); ?></label>
        </li>
        <?php
    }

    /**
     * Return filter by terms radio html
     *
     * @param array $args
     */
    public function renderTermsSelectFields($args = [])
    {
        $defaults = [
            'index' => 0,
            'items' => [],
            'type' => 'select',
            'filter-by' => 'term'
        ];

        $args = array_replace($defaults, $args);
        $active = $args['filter-by'] == 'term';
        $attributes = \wc_get_attribute_taxonomies();
        $sources = [
            'product_cat' => L10N::r('Category'),
            'product_tag' => L10N::r('Tag')
        ];

        foreach ($attributes as $attribute) {
            $sources['pa_' . $attribute->attribute_name] = $attribute->attribute_label;
        }
        ?>
        <li class="wcsf-question-manual-value-filter<?php echo !$active ? ' is-closed' : ''; ?>"
            data-component="wcsf-question-manual-value-filter"
            data-id="term">
            <label class="wcsf-question-manual-value-filter-label">
                <input type="radio" data-component="wcsf-question-manual-value-filter-by"
                    data-id="term" name="_manual_values[<?php echo esc_attr($args['index']); ?>][filter-by]"
                    value="term" <?php checked($active); ?>><?php L10N::e('Filter by terms'); ?></label>
            <fieldset class="wcsf-question-manual-value-filter-values"
                data-component="wcsf-question-manual-value-filter-values"><?php
                Admin::settingFieldView(
                    [
                        'label' => L10N::r('Terms'),
                        'key' => "_manual_values[{$args['index']}][items]",
                        'type' => 'data-table',
                        'values' => [
                            'source' => [
                                'label' => L10N::r('Source'),
                                'key' => 'source',
                                'type' => 'select',
                                'values' => $sources,
                                'default' => '',
                                'data-component' => 'wcsf-question-manual-value-filter-values-item-source'
                            ],
                            'terms' => [
                                'label' => L10N::r('Terms'),
                                'key' => 'terms',
                                'type' => 'wc-terms-search',
                                'target-parent' => '[data-component="wcsf-data-table-item"]',
                                'target-selector' => '[data-key="attribute"] select',
                                'action' => 'woocommerce_json_search_category',
                                'default' => '',
                                'queryArgs' => ['taxonomy' => array_keys($sources)],
                                'data-component' => 'wcsf-question-manual-value-filter-values-item-terms'
                            ],
                            'inner_relation' => [
                                'label' => L10N::r('Inner relation'),
                                'key' => 'inner_relation',
                                'type' => 'select',
                                'default' => 'or',
                                'values' => [
                                    'or' => L10N::r('OR'),
                                    'and' => L10N::r('AND')
                                ]
                            ],
                            'outer_relation' => [
                                'label' => L10N::r('Outer relation'),
                                'key' => 'outer_relation',
                                'type' => 'select',
                                'default' => 'or',
                                'values' => [
                                    'or' => L10N::r('OR'),
                                    'and' => L10N::r('AND')
                                ]
                            ]
                        ],
                        'default' => []
                    ],
                    ['values' => ["_manual_values[{$args['index']}][items]" => $active ? $args['items'] : []]]
                );
                ?></fieldset>
        </li>
        <?php
    }

    /**
     * Return filter by product html
     *
     * @param array $args
     */
    public function renderProductSelectFields($args = [])
    {
        $defaults = [
            'index' => 0,
            'items' => [],
            'type' => 'select',
            'filter-by' => 'product',
            'intersect' => true
        ];

        $args = array_replace($defaults, $args);
        $active = $args['filter-by'] == 'product';
        ?>
        <li class="wcsf-question-manual-value-filter<?php echo !$active ? ' is-closed' : ''; ?>"
            data-component="wcsf-question-manual-value-filter"
            data-id="product">
            <label class="wcsf-question-manual-value-filter-label">
                <input type="radio" data-component="wcsf-question-manual-value-filter-by"
                    data-id="product" name="_manual_values[<?php echo esc_attr($args['index']); ?>][filter-by]"
                    value="product" <?php checked($active); ?>><?php L10N::e('Select products manually'); ?></label>
            <fieldset class="wcsf-question-manual-value-filter-values"
                data-component="wcsf-question-manual-value-filter-values">
                <table class="wcsf-settings-table form-table wp-list-table">
                    <tbody>
                        <tr>
                            <th scope="row">
                                <label for="<?php echo esc_attr("_manual_values[{$args['index']}][items]"); ?>"><?php
                                    L10N::e('Products');
                                    ?></label>
                            </th>
                            <td>
                                <?php
                                Admin::settingFieldView(
                                    [
                                        'label' => L10N::r('Products'),
                                        'key' => "_manual_values[{$args['index']}][items]",
                                        'type' => 'wc-product-search',
                                        'action' => 'woocommerce_json_search_products',
                                        'default' => []
                                    ],
                                    [
                                        'values' => [
                                            "_manual_values[{$args['index']}][items]" => $active ? $args['items'] : []
                                        ]
                                    ]
                                );
                                ?>
                            </td>
                        </tr>
                        <tr>
                            <th scope="row">
                                <label for="<?php
                                echo esc_attr("_manual_values[{$args['index']}][intersect]");
                                ?>"><?php
                                    L10N::e('Intersect with the other answers');
                                    ?></label>
                            </th>
                            <td>
                                <?php
                                Admin::settingFieldView(
                                    [
                                        'label' => L10N::r('Intersect with the other answers'),
                                        'key' => "_manual_values[{$args['index']}][intersect]",
                                        'type' => 'checkbox',
                                        'default' => true,
                                        'description' => L10N::r('Filter other manually selected products intersected')
                                    ],
                                    ['values' => ["_manual_values[{$args['index']}][intersect]" => $args['intersect']]]
                                );
                                ?>
                            </td>
                        </tr>
                    </tbody>
                </table>
            </fieldset>
        </li>
        <?php
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Get settings model array
     *
     * @return array
     */
    public static function getSettingsModel()
    {
        return apply_filters(
            'wcsf_question_settings_model',
            [
                'label' => [
                    'label' => L10N::r('Label'),
                    'key' => '_label',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Optional - used instead of the post title')
                ],
                'view_type' => [
                    'label' => L10N::r('View type'),
                    'key' => '_view_type',
                    'type' => 'select',
                    'default' => 'select',
                    'values' => [
                        'select' => L10N::r('Select'),
                        'radio' => L10N::r('Radio'),
                        'checkbox' => L10N::r('Checkbox'),
                        'chips-radio' => L10N::r('Chips radio'),
                        'chips-checkbox' => L10N::r('Chips checkbox'),
                        'button-radio' => L10N::r('Button radio'),
                        'button-checkbox' => L10N::r('Button checkbox'),
                        'image-radio' => L10N::r('Image radio'),
                        'image-checkbox' => L10N::r('Image checkbox'),
                        'number' => L10N::r('Number'),
                        'number-from' => L10N::r('Number from'),
                        'number-to' => L10N::r('Number to'),
                        'number-between' => L10N::r('Number between'),
                        'multi-choice-checkbox' => L10N::r('Multi-choice checkbox'),
                        'multi-choice-range' => L10N::r('Multi-choice range'),
                        'formula' => L10N::r('Formula')
                    ]
                ],
                'filter_type' => [
                    'label' => L10N::r('Filter type'),
                    'key' => '_filter_type',
                    'type' => 'select',
                    'default' => 'select',
                    'values' => [
                        'attribute' => L10N::r('Attribute'),
                        'price' => L10N::r('Price'),
                        'tag' => L10N::r('Tag'),
                        'category' => L10N::r('Category'),
                        'manual' => L10N::r('Manual'),
                        'meta' => L10N::r('Meta')
                    ],
                    'separate' => true
                ],
                'attribute' => [
                    'label' => L10N::r('Attribute'),
                    'key' => '_attribute',
                    'type' => 'select',
                    'default' => 0,
                    'values' => []
                ],
                'attribute_value_source' => [
                    'label' => L10N::r('Attribute value source'),
                    'key' => '_attribute_value_source',
                    'type' => 'select',
                    'default' => 'name',
                    'values' => [
                        'name' => L10N::r('Name'),
                        'slug' => L10N::r('Slug')
                    ],
                    'description' => L10N::r('Select the attribute value property for using'),
                    'separate' => true
                ],
                'meta_key' => [
                    'label' => L10N::r('Meta key'),
                    'key' => '_meta_key',
                    'type' => 'text',
                    'default' => '',
                    'separate' => true
                ],
                'multi_choice_values' => [
                    'label' => L10N::r('Available values'),
                    'key' => '_multi_choice_possible_values',
                    'type' => 'multi-select',
                    'default' => [0, 1, 2, -1],
                    'values' => [
                        0 => L10N::r('Undefined'),
                        1 => L10N::r('Possible'),
                        2 => L10N::r('Required'),
                        -1 => L10N::r('Excluded')
                    ]
                ],
                'multi_choice_undefined_value_text' => [
                    'label' => L10N::r('"Undefined" value text'),
                    'key' => '_multi_choice_undefined_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Undefined')
                ],
                'multi_choice_possible_value_text' => [
                    'label' => L10N::r('"Possible" value text'),
                    'key' => '_multi_choice_possible_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Possible')
                ],
                'multi_choice_required_value_text' => [
                    'label' => L10N::r('"Required" value text'),
                    'key' => '_multi_choice_required_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Required')
                ],
                'multi_choice_excluded_value_text' => [
                    'label' => L10N::r('"Excluded" value text'),
                    'key' => '_multi_choice_excluded_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Excluded'),
                    'separate' => true
                ],
                'from_value_text' => [
                    'label' => L10N::r('"From" value text'),
                    'key' => '_from_value_text',
                    'type' => 'text',
                    'default' => L10N::r('From')
                ],
                'to_value_text' => [
                    'label' => L10N::r('"To" value text'),
                    'key' => '_to_value_text',
                    'type' => 'text',
                    'default' => L10N::r('To')
                ],
                'enable_slider' => [
                    'label' => L10N::r('Enable slider'),
                    'key' => '_enable_slider',
                    'type' => 'checkbox',
                    'default' => true
                ],
                'show_slider_tooltips' => [
                    'label' => L10N::r('Show slider tooltips'),
                    'key' => '_show_slider_tooltips',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'show_inputs' => [
                    'label' => L10N::r('Show inputs'),
                    'key' => '_show_inputs',
                    'type' => 'checkbox',
                    'default' => true
                ],
                'label_as_placeholder' => [
                    'label' => L10N::r('Label as placeholder'),
                    'key' => '_label_as_placeholder',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Move label to input placeholder')
                ],
                'numeric_unit' => [
                    'label' => L10N::r('Numeric unit'),
                    'key' => '_numeric_unit',
                    'type' => 'text',
                    'default' => '$'
                ],
                'numeric_step' => [
                    'label' => L10N::r('Value step'),
                    'key' => '_numeric_step',
                    'type' => 'number',
                    'default' => 1,
                    'step' => 'any',
                    'min' => 0.0001,
                    'separate' => true
                ],
                'alignment' => [
                    'label' => L10N::r('Thumbnails alignment'),
                    'key' => '_alignment',
                    'type' => 'select',
                    'default' => 'select',
                    'values' => [
                        'start' => L10N::r('Start'),
                        'end' => L10N::r('End'),
                        'center' => L10N::r('Center'),
                        'between' => L10N::r('Between'),
                        'around' => L10N::r('Around'),
                        'evenly' => L10N::r('Evenly')
                    ]
                ],
                'thumbnail_size' => [
                    'label' => L10N::r('Thumbnail size'),
                    'key' => '_thumbnail_size',
                    'type' => 'text',
                    'default' => 'thumbnail',
                    // phpcs:disable
                    'description' => L10N::r('Set width and height separated by a comma or use string value. For example thumbnail, medium, large')
                    // phpcs:enable
                ],
                'thumbnail_width' => [
                    'label' => L10N::r('Thumbnail width (px)'),
                    'key' => '_thumbnail_width',
                    'type' => 'number',
                    'default' => '',
                    'min' => 0,
                    'description' => L10N::r('Limit the field best width value'),
                    'separate' => true
                ],
                'category_parent' => [
                    'label' => L10N::r('Category parent'),
                    'key' => '_category_parent',
                    'type' => 'select',
                    'default' => 0,
                    'values' => []
                ],
                'category_parent_by_question' => [
                    'label' => L10N::r('Category parent by a question answer'),
                    'key' => '_category_parent_by_question',
                    'type' => 'select',
                    'default' => '',
                    'values' => [],
                    'description' => L10N::r('Output only children of the category-based question answer')
                ],
                'start_from_categories_level' => [
                    'label' => L10N::r('Start from categories level'),
                    'key' => '_start_from_categories_level',
                    'type' => 'number',
                    'default' => 0,
                    'min' => 0,
                ],
                'max_sub_categories_level' => [
                    'label' => L10N::r('Max sub-categories level'),
                    'key' => '_max_sub_categories_level',
                    'type' => 'number',
                    'default' => 0,
                    'min' => 0,
                    'description' => L10N::r('Must be greater or equal to the "Start from categories level" setting')
                ],
                'hide_empty_terms' => [
                    'label' => L10N::r('Hide empty terms'),
                    'key' => '_hide_empty_terms',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'included_terms' => [
                    'label' => L10N::r('Included terms'),
                    'key' => '_included_terms',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Define terms IDs with comma to use only them in the question')
                ],
                'excluded_terms' => [
                    'label' => L10N::r('Excluded terms'),
                    'key' => '_excluded_terms',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Define terms IDs with comma to not use them in the question')
                ],
                'show_count' => [
                    'label' => L10N::r('Show count'),
                    'key' => '_show_count',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Show term products count')
                ],
                'show_description' => [
                    'label' => L10N::r('Show description'),
                    'key' => '_show_description',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Show terms description'),
                    'separate' => true
                ],
                'formula' => [
                    'label' => L10N::r('Formula'),
                    'key' => '_formula',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Use question description')
                ],
                'formula_operator' => [
                    'label' => L10N::r('Formula operator'),
                    'key' => '_formula_operator',
                    'type' => 'select',
                    'default' => '=',
                    'values' => [
                        '=' => L10N::r('Equal'),
                        '!=' => L10N::r('Not equal'),
                        '<' => L10N::r('Less'),
                        '<=' => L10N::r('Less or equal'),
                        '>' => L10N::r('More'),
                        '>=' => L10N::r('More or equal')
                    ]
                ],
                'formula_variables' => [
                    'label' => L10N::r('Formula variables'),
                    'key' => '_formula_variables',
                    'type' => 'data-table',
                    'default' => [],
                    // phpcs:disable
                    'values' => [
                        'name' => [
                            'label' => L10N::r('Name'),
                            'key' => 'name',
                            'type' => 'text',
                            'default' => '',
                            'placeholder' => 'x'
                        ],
                        'value' => [
                            'label' => L10N::r('Value'),
                            'key' => 'value',
                            'type' => 'number',
                            'step' => 'any',
                            'default' => 0
                        ]
                    ],
                    'description' => L10N::r('"Key => number" pair for static variables. Simple numeric values might work wrong right in the formula. Or define default variable values here.')
                    // phpcs:enable
                ],
                'formula_name_aliases' => [
                    'label' => L10N::r('Formula name aliases'),
                    'key' => '_formula_name_aliases',
                    'type' => 'data-table',
                    'default' => [],
                    'values' => [
                        'name' => [
                            'label' => L10N::r('Name'),
                            'key' => 'name',
                            'type' => 'text',
                            'default' => '',
                            'placeholder' => L10N::r('My custom field')
                        ],
                        'alias' => [
                            'label' => L10N::r('Alias'),
                            'key' => 'alias',
                            'type' => 'text',
                            'default' => '',
                            'placeholder' => 'a'
                        ]
                    ],
                    'description' => L10N::r('Transform input names to variables')
                ],
                'formula_value_aliases' => [
                    'label' => L10N::r('Formula value aliases'),
                    'key' => '_formula_value_aliases',
                    'type' => 'data-table',
                    'default' => [],
                    'values' => [
                        'name' => [
                            'label' => L10N::r('Name'),
                            'key' => 'name',
                            'type' => 'text',
                            'default' => '',
                            'placeholder' => L10N::r('My custom field')
                        ],
                        'value' => [
                            'label' => L10N::r('Value'),
                            'key' => 'value',
                            'type' => 'text',
                            'default' => '',
                            'placeholder' => L10N::r('My value')
                        ],
                        'alias' => [
                            'label' => L10N::r('Alias'),
                            'key' => 'alias',
                            'type' => 'number',
                            'step' => 'any',
                            'default' => '',
                            'placeholder' => ''
                        ]
                    ],
                    'description' => L10N::r('Transform input values to numbers')
                ],
                'output_formula_value' => [
                    'label' => L10N::r('Output calculated formula value'),
                    'key' => '_output_formula_value',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'formula_value_text' => [
                    'label' => L10N::r('Calculated formula value text'),
                    'key' => '_formula_value_text',
                    'type' => 'text',
                    'default' => '<br>Formula result: %value%',
                    'description' => L10N::r('"%value%" - formula value variable'),
                    'separate' => true
                ],
                'relation' => [
                    'label' => L10N::r('Relation'),
                    'key' => '_relation',
                    'type' => 'select',
                    'default' => 'or',
                    'values' => [
                        'or' => L10N::r('OR'),
                        'and' => L10N::r('AND')
                    ],
                    'description' => L10N::r('OR - any of the answers, AND - all of the answers')
                ],
                'action_on_click' => [
                    'label' => L10N::r('Action on click'),
                    'key' => '_action_on_click',
                    'type' => 'select',
                    'default' => 'none',
                    'values' => [
                        'none' => L10N::r('None'),
                        'submit' => L10N::r('Submit'),
                        'submit-and-go-next' => L10N::r('Submit and go next'),
                        'submit-and-go-results' => L10N::r('Submit and go to results')
                    ],
                    'separate' => true
                ],
                'min_values_selected' => [
                    'label' => L10N::r('Minimum values selected'),
                    'key' => '_min_values_selected',
                    'type' => 'number',
                    'default' => '',
                    'min' => 0
                ],
                'max_values_selected' => [
                    'label' => L10N::r('Maximum values selected'),
                    'key' => '_max_values_selected',
                    'type' => 'number',
                    'default' => '',
                    'min' => 0
                ],
                'min_values_selected_message' => [
                    'label' => L10N::r('Minimum values selected message'),
                    'key' => '_min_values_selected_message',
                    'type' => 'text',
                    'default' => L10N::r('Minimum selected values are required: %limit%'),
                    'description' => L10N::r('"%limit%" - values limit')
                        . '<br>'
                        . L10N::r('"%value%" - current values count'),
                ],
                'max_values_selected_message' => [
                    'label' => L10N::r('Maximum values selected message'),
                    'key' => '_max_values_selected_message',
                    'type' => 'text',
                    'default' => L10N::r('Maximum selected values are required: %limit%'),
                    'description' => L10N::r('"%limit%" - values limit')
                        . '<br>'
                        . L10N::r('"%value%" - current values count'),
                    'separate' => true
                ],
                'add_any_product_value' => [
                    'label' => L10N::r('Add "Doesn\'t matter" value'),
                    'key' => '_any_product_value',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'any_product_value_text' => [
                    'label' => L10N::r('"Doesn\'t matter" value text'),
                    'key' => '_any_product_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Doesn\'t matter')
                ],
                'any_product_value_position' => [
                    'label' => L10N::r('"Doesn\'t matter" value position'),
                    'key' => '_any_product_value_position',
                    'type' => 'number',
                    'default' => 0,
                    'min' => 0
                ],
                'any_product_value_checked' => [
                    'label' => L10N::r('"Doesn\'t matter" value is checked'),
                    'key' => '_any_product_value_checked',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'any_product_value_thumbnail' => [
                    'label' => L10N::r('"Doesn\'t matter" value thumbnail'),
                    'key' => '_any_product_value_thumbnail',
                    'type' => 'thumbnail',
                    'default' => '',
                    'separate' => true
                ],
                'add_empty_value' => [
                    'label' => L10N::r('Add empty value'),
                    'key' => '_add_empty_value',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'empty_value_text' => [
                    'label' => L10N::r('Empty value text'),
                    'key' => '_empty_value_text',
                    'type' => 'text',
                    'default' => L10N::r('Choose a value'),
                    'separate' => true
                ],
                'unmet_values_behavior' => [
                    'label' => L10N::r('Unmet values behavior'),
                    'key' => '_unmet_values_behavior',
                    'type' => 'select',
                    'default' => 'none',
                    'values' => [
                        'none' => L10N::r('Keep'),
                        'disable' => L10N::r('Disable'),
                        'hide' => L10N::r('Hide')
                    ],
                    'description' => L10N::r('Values which will lead to zero results will be disabled')
                ],
                'additional_get_param' => [
                    'label' => L10N::r('Additional GET parameter'),
                    'key' => '_additional_get_param',
                    'type' => 'text',
                    'default' => '',
                    'description' =>
                        L10N::r('Add extra GET parameter with selected value on the results page')
                        . '<br>'
                        . L10N::r('For "number between" view set two strings separated by | symbol if needed')
                ],
                'additional_get_param_term_source' => [
                    'label' => L10N::r('Additional GET parameter term source'),
                    'key' => '_additional_get_param_term_source',
                    'type' => 'select',
                    'default' => 'id',
                    'values' => [
                        'id' => L10N::r('ID'),
                        'slug' => L10N::r('Slug'),
                        'name' => L10N::r('Name')
                    ],
                    'separate' => true
                ],
                'dependencies' => [
                    'label' => L10N::r('Dependencies'),
                    'key' => '_dependencies',
                    'type' => 'data-table',
                    'default' => [],
                    'values' => [
                        'question' => [
                            'label' => L10N::r('Question'),
                            'key' => 'question',
                            'type' => 'select',
                            'default' => '',
                            'values' => []
                        ],
                        'value' => [
                            'label' => L10N::r('Value'),
                            'key' => 'value',
                            'type' => 'ajax-select',
                            'default' => '',
                            'target-parent' => '[data-component="wcsf-data-table-item"]',
                            'target-selector' => '[data-key="question"] select',
                            'action' => 'wcsfGetQuestionValuesSelectOptions',
                            // phpcs:disable
                            'description' => L10N::r('Select a presented value or type own for price or formula based questions')
                            // phpcs:enable
                        ],
                        'operator' => [
                            'label' => L10N::r('Operator'),
                            'key' => 'operator',
                            'type' => 'select',
                            'default' => '=',
                            'values' => [
                                '=' => L10N::r('Equal'),
                                '!=' => L10N::r('Not equal'),
                                '<' => L10N::r('Less'),
                                '<=' => L10N::r('Less or equal'),
                                '>' => L10N::r('More'),
                                '>=' => L10N::r('More or equal'),
                                '0' => L10N::r('Undefined'),
                                '1' => L10N::r('Possible'),
                                '2' => L10N::r('Required'),
                                '-1' => L10N::r('Excluded')
                            ]
                        ],
                        'upper_value' => [
                            'label' => L10N::r('Upper value'),
                            'key' => 'upper_value',
                            'type' => 'checkbox',
                            'default' => false,
                            'description' => L10N::r('For the "Number between" type')
                        ],
                        'relation' => [
                            'label' => L10N::r('Relation with the next rule'),
                            'key' => 'relation',
                            'type' => 'select',
                            'default' => 'or',
                            'values' => [
                                'or' => L10N::r('OR'),
                                'and' => L10N::r('AND')
                            ]
                        ]
                    ],
                    'inModal' => true,
                    'description' =>
                        L10N::r('The question will be shown only in the case of the meeting of a dependency')
                ],
                'bond_questions' => [
                    'label' => L10N::r('Bind with questions'),
                    'key' => '_bonded_questions',
                    'type' => 'multi-select',
                    'default' => [],
                    'values' => [],
                    'description' => L10N::r('Question value will be passed to the bonded questions too')
                ],
                'manual_values' => [
                    'label' => L10N::r('Manual values'),
                    'key' => '_manual_values',
                    'type' => 'data-table',
                    'default' => []
                ],
                'ignore_filtering' => [
                    'label' => L10N::r('Ignore for filtering'),
                    'key' => '_ignore_filtering',
                    'type' => 'checkbox',
                    'default' => false,
                    // phpcs:disable
                    'description' => L10N::r('Keep the question to apply dependencies or collect statistic but ignore it while products filtering')
                    // phpcs:enable
                ]
            ]
        );
    }

    /**
     * Get category settings model array
     *
     * @return array
     */
    public static function getCategorySettingsModel()
    {
        return apply_filters(
            'wcsf_question_category_settings_model',
            [
                'layout' => [
                    'label' => L10N::r('Layout'),
                    'key' => '_layout',
                    'type' => 'select',
                    'default' => 'list',
                    'values' => [
                        'list' => L10N::r('List'),
                        'grid' => L10N::r('Grid')
                    ]
                ],
                'grid_columns' => [
                    'label' => L10N::r('Grid columns number'),
                    'key' => '_grid_columns',
                    'type' => 'group',
                    'default' => [],
                    'values' => [
                        [
                            'label' => L10N::r('XXS'),
                            'key' => 'xxs',
                            'type' => 'number',
                            'default' => 1,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('XS'),
                            'key' => 'xs',
                            'type' => 'number',
                            'default' => 1,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('S'),
                            'key' => 'sm',
                            'type' => 'number',
                            'default' => 2,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('M'),
                            'key' => 'md',
                            'type' => 'number',
                            'default' => 2,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('L'),
                            'key' => 'lg',
                            'type' => 'number',
                            'default' => 2,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('XL'),
                            'key' => 'xl',
                            'type' => 'number',
                            'default' => 3,
                            'min' => 1,
                            'max' => 12
                        ],
                        [
                            'label' => L10N::r('XXL'),
                            'key' => 'xxl',
                            'type' => 'number',
                            'default' => 3,
                            'min' => 1,
                            'max' => 12
                        ]
                    ],
                    'showHeader' => true,
                    'description' => L10N::r('Number of columns from 1 to 12 according to the screen size'),
                    'group' => L10N::r('View')
                ],
                'thumbnail' => [
                    'label' => L10N::r('Thumbnail'),
                    'key' => '_thumbnail',
                    'type' => 'thumbnail',
                    'default' => ''
                ]
            ]
        );
    }
    // </editor-fold>

    // <editor-fold desc="Get">
    /**
     * Return question static data object
     *
     * @param \WP_Post $question
     *
     * @return object|null
     */
    public static function getData($question)
    {
        if (!$question || !property_exists($question, 'ID')) {
            return apply_filters('wcsf_question_data', null, $question);
        }

        static $cache = [];

        if (isset($cache[$question->ID])) {
            return apply_filters('wcsf_question_data', $cache[$question->ID], $question);
        }

        $viewType = self::getSetting($question->ID, 'view_type');
        $label = self::getSetting($question->ID, 'label');
        $thumbnailSize = self::getSetting($question->ID, 'thumbnail_size');

        if (strpos($thumbnailSize, ',') !== false) {
            $thumbnailSize = explode(',', $thumbnailSize);
        }

        $output = [
            'id' => (int) $question->ID,
            'name' => $label ? $label : $question->post_title,
            'description' => $question->post_content,
            'thumbnail' => get_post_thumbnail_id($question->ID),
            'thumbnailSize' => $thumbnailSize,
            'thumbnailWidth' => self::getSetting($question->ID, 'thumbnail_width'),
            'viewType' => $viewType,
            'filterType' => self::getSetting($question->ID, 'filter_type'),
            'attribute' => self::getSetting($question->ID, 'attribute'),
            'alignment' => self::getSetting($question->ID, 'alignment'),
            'manualValues' => self::getSetting($question->ID, 'manual_values'),
            'relation' => self::getSetting($question->ID, 'relation'),
            'metaKey' => self::getSetting($question->ID, 'meta_key'),
            'ignoreFiltering' => self::getSetting($question->ID, 'ignore_filtering')
        ];

        if (strpos($viewType, 'multi-choice') !== false) {
            $multiChoiceValuesIncluded = (array) self::getSetting($question->ID, 'multi_choice_values');
            $multiChoiceValues = [
                0 => [
                    'label' => self::getSetting($question->ID, 'multi_choice_undefined_value_text'),
                    'key' => 'undefined',
                    'value' => 0
                ],
                1 => [
                    'label' => self::getSetting($question->ID, 'multi_choice_possible_value_text'),
                    'key' => 'possible',
                    'value' => 1
                ],
                2 => [
                    'label' => self::getSetting($question->ID, 'multi_choice_required_value_text'),
                    'key' => 'required',
                    'value' => 2
                ],
                -1 => [
                    'label' => self::getSetting($question->ID, 'multi_choice_excluded_value_text'),
                    'key' => 'excluded',
                    'value' => -1
                ]
            ];

            foreach ($multiChoiceValues as $key => $multiChoiceValue) {
                if (in_array($multiChoiceValue['value'], $multiChoiceValuesIncluded)) {
                    continue;
                }

                unset($multiChoiceValues[$key]);
            }

            $multiChoiceValues = array_replace(array_flip($multiChoiceValuesIncluded), $multiChoiceValues);
            $output['multiChoiceValues'] = array_values(array_filter($multiChoiceValues));
        } elseif (strpos($viewType, 'number') !== false) {
            $output['enableSlider'] = self::getSetting($question->ID, 'enable_slider');
            $output['showSliderTooltips'] = self::getSetting($question->ID, 'show_slider_tooltips');
            $output['showInputs'] = self::getSetting($question->ID, 'show_inputs');
            $output['labelAsPlaceholder'] = self::getSetting($question->ID, 'label_as_placeholder');
            $output['unit'] = self::getSetting($question->ID, 'numeric_unit');
        } elseif ($viewType == 'formula') {
            $output['formula'] = self::getSetting($question->ID, 'formula');
            $output['formulaVariables'] = self::getSetting($question->ID, 'formula_variables');
            $output['formulaNameAliases'] = self::getSetting($question->ID, 'formula_name_aliases');
            $output['formulaValueAliases'] = self::getSetting($question->ID, 'formula_value_aliases');
            $output['formulaOperator'] = self::getSetting($question->ID, 'formula_operator');
        }

        $output = $cache[$question->ID] = (object) $output;

        return apply_filters('wcsf_question_data', $output, $question);
    }

    /**
     * Return a question by id
     *
     * @param integer $id
     *
     * @return object|null
     */
    public static function get($id)
    {
        $output = null;

        if (!is_numeric($id)) {
            return apply_filters('wcsf_question', $output, $id);
        }

        static $cache = [];

        if (isset($cache[$id])) {
            return apply_filters('wcsf_question', $cache[$id], $id);
        }

        $output = $cache[$id] = self::getData(get_post($id));

        return apply_filters('wcsf_question', $output, $id);
    }

    /**
     * Return array of questions
     *
     * @param integer $filterId
     * @param array $ids
     *
     * @return array
     */
    public static function getList($filterId, $ids)
    {
        global $wpdb;

        static $cache = [];

        $ids = array_filter($ids);
        $idsCopy = $ids;

        sort($idsCopy);

        $hash = md5(serialize($idsCopy));
        $output = [];

        if (empty($filterId) || empty($ids)) {
            return apply_filters('wcsf_questions_list', $output, $filterId, $ids);
        }

        if (isset($cache[$hash])) {
            return apply_filters('wcsf_questions_list', $cache[$hash], $filterId, $ids);
        }

        // using of an SQL query avoids cross wp_queries problems
        $value = Filter::getValue($filterId);
        $output = [];
        $questionsIdsString = esc_sql(implode(', ', $ids));
        $queryString = "SELECT $wpdb->posts.* 
            FROM $wpdb->posts
            WHERE $wpdb->posts.post_status = 'publish' 
            AND $wpdb->posts.post_type = '" . self::$postType . "'
            AND $wpdb->posts.ID IN ($questionsIdsString)
            ORDER BY FIELD(ID, $questionsIdsString) ASC";

        $questions = $wpdb->get_results($queryString, OBJECT);

        foreach ($questions as $question) {
            if (!$question || !property_exists($question, 'ID')) {
                continue;
            }

            $output[$question->ID] = self::getData($question);
            $output[$question->ID]->hasValue = isset($value[$question->ID]);
        }

        $output = $cache[$hash]
            = self::filterListAccordingDependencies($filterId, $output, Filter::getValue($filterId));

        return apply_filters('wcsf_questions_list', $output, $filterId, $ids);
    }

    /**
     * Returns array of ids and titles of categories
     *
     * @param array $args
     *
     * @return array
     */
    public static function getCategoriesIdsList($args = [])
    {
        $defaults = [
            'taxonomy' => self::$taxonomy,
            'fields' => 'id=>name',
            'hide_empty' => false,
        ];

        $args = array_replace($defaults, $args);
        $output = get_terms($args);

        return apply_filters('wcsf_questions_categories_list', $output);
    }

    /**
     * Remove question which are not met their dependencies
     *
     * @param integer $filterId
     * @param array $questions
     * @param array $values
     *
     * @return array
     */
    public static function filterListAccordingDependencies($filterId, $questions, $values)
    {
        foreach ($questions as $questionId => $question) {
            $dependencies = (array) self::getSetting($questionId, 'dependencies');
            $met = false;
            $localMet = false;
            $previous = null;

            if (empty($dependencies) || empty(reset($dependencies)['question'])) {
                continue;
            }

            foreach ($dependencies as $dependency) {
                if (empty($dependency['question'])) {
                    continue;
                }

                if (!isset($values[$dependency['question']])) {
                    $localMet = false;
                } else {
                    $upperValue = isset($dependency['upper_value'])
                        && filter_var($dependency['upper_value'], FILTER_VALIDATE_BOOLEAN);

                    $numericValue = $upperValue
                        ? (float) end($values[$dependency['question']])
                        : (float) reset($values[$dependency['question']]);

                    switch ($dependency['operator']) {
                        case '=':
                            $localMet = in_array($dependency['value'], $values[$dependency['question']]);
                            break;

                        case '!=':
                            $localMet = !in_array($dependency['value'], $values[$dependency['question']]);
                            break;

                        case '<':
                            $localMet = $numericValue < (float) $dependency['value'];
                            break;

                        case '<=':
                            $localMet = $numericValue <= (float) $dependency['value'];
                            break;

                        case '>':
                            $localMet = $numericValue > (float) $dependency['value'];
                            break;

                        case '>=':
                            $localMet = $numericValue >= (float) $dependency['value'];
                            break;

                        case '0':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 0;

                            break;
                        case '1':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 1;

                            break;
                        case '2':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 2;

                            break;
                        case '-1':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == -1;
                    }
                }

                // look at the previous value
                if (!is_null($previous)) {
                    $localMet = $localMet && $previous;
                }

                if ($dependency['relation'] == 'and' && end($dependencies) != $dependency) {
                    $previous = $localMet;
                } else {
                    $previous = null;
                    $met = $met || $localMet;
                }
            }

            if (!$met) {
                Filter::resetState($filterId, $questionId);
                unset($questions[$questionId]);
            }
        }

        return apply_filters('wcsf_filtered_questions_list', $questions, $filterId);
    }
    // </editor-fold>

    // <editor-fold desc="Fields">
    /**
     * Remove manual values which are not met their dependencies
     *
     * @param integer $filterId
     * @param array $items
     * @param array $values
     *
     * @return array
     */
    public static function filterManualValuesAccordingDependencies($filterId, $items, $values)
    {
        foreach ($items as $key => $item) {
            if (!isset($item['dependencies'])) {
                continue;
            }

            $dependencies = $item['dependencies'];
            $met = false;
            $localMet = false;
            $previous = null;

            if (empty($dependencies) || empty(reset($dependencies)['question'])) {
                continue;
            }

            foreach ($dependencies as $dependency) {
                if (empty($dependency['question'])) {
                    continue;
                }

                if (!isset($values[$dependency['question']])) {
                    $localMet = false;
                } else {
                    $upperValue = isset($dependency['upper_value'])
                        && filter_var($dependency['upper_value'], FILTER_VALIDATE_BOOLEAN);

                    $numericValue = $upperValue
                        ? (float) end($values[$dependency['question']])
                        : (float) reset($values[$dependency['question']]);

                    switch ($dependency['operator']) {
                        case '=':
                            $localMet = in_array($dependency['value'], $values[$dependency['question']]);
                            break;

                        case '!=':
                            $localMet = !in_array($dependency['value'], $values[$dependency['question']]);
                            break;

                        case '<':
                            $localMet = $numericValue < (float) $dependency['value'];
                            break;

                        case '<=':
                            $localMet = $numericValue <= (float) $dependency['value'];
                            break;

                        case '>':
                            $localMet = $numericValue > (float) $dependency['value'];
                            break;

                        case '>=':
                            $localMet = $numericValue >= (float) $dependency['value'];
                            break;

                        case '0':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 0;

                            break;
                        case '1':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 1;

                            break;
                        case '2':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == 2;

                            break;
                        case '-1':
                            $localMet = isset($values[$dependency['question']][$dependency['value']])
                                && $values[$dependency['question']][$dependency['value']] == -1;
                    }
                }

                // look at the previous value
                if (!is_null($previous)) {
                    $localMet = $localMet && $previous;
                }

                if ($dependency['relation'] == 'and' && end($dependencies) != $dependency) {
                    $previous = $localMet;
                } else {
                    $previous = null;
                    $met = $met || $localMet;
                }
            }

            if (!$met) {
                unset($items[$key]);
            }
        }

        return apply_filters('wcsf_filtered_manual_values', $items, $filterId);
    }

    /**
     * Get fields according the question settings
     *
     * @param integer $filterId
     * @param integer $questionId
     * @param array $extraQueryArgs - additional WP_Query arguments
     *
     * @return array
     */
    public static function getFields($filterId, $questionId, $extraQueryArgs = [])
    {
        $output = [];

        if (!is_numeric($questionId)) {
            return apply_filters('wcsf_question_fields', $output, $filterId, $questionId);
        }

        $question = self::get($questionId);

        if (!is_object($question) || !property_exists($question, 'viewType') || $question->viewType == 'formula') {
            return apply_filters('wcsf_question_fields', $output, $filterId, $questionId);
        }

        $output = [];
        $boundary = [];
        $products = [];
        $component = 'wcsf-field-input';
        $value = Filter::getValue($filterId);
        $actionOnClick = self::getSetting($questionId, 'action_on_click');
        $currentValue = isset($value[$questionId]) ? $value[$questionId] : [];
        $viewsWithAnyProductValue = ['select', 'radio', 'image-radio', 'chips-radio', 'button-radio'];
        $viewsWithDisableUnmet = array_merge(
            $viewsWithAnyProductValue,
            ['checkbox', 'chips-checkbox', 'button-checkbox', 'image-checkbox'],
            ['number-from', 'number-to', 'number', 'number-between']
        );

        $unmetValuesBehavior = self::getSetting($questionId, 'unmet_values_behavior');
        $showPreliminaryResults = Filter::getSetting($filterId, 'show_preliminary_results_count')
            || Filter::getSetting($filterId, 'show_header_preliminary_results')
            || Filter::getSetting($filterId, 'show_footer_preliminary_results');

        $disableUnmet = in_array($unmetValuesBehavior, ['disable', 'hide'])
            && in_array($question->viewType, $viewsWithDisableUnmet);

        if ($disableUnmet) {
            // don't filter products by this question to make all values available
            $valueCopy = $value;

            if (isset($valueCopy[$questionId])) {
                unset($valueCopy[$questionId]);
            }

            $products = get_posts(array_replace(
                [
                    'post_type' => 'product',
                    'fields' => 'ids',
                    'no_found_rows' => true,
                    'nopaging' => true
                ],
                Core::getQueryArgsByFilter($filterId, $valueCopy, $extraQueryArgs),
                $extraQueryArgs
            ));
        }

        if ($actionOnClick != 'none') {
            $component .= ' wcsf-' . $actionOnClick;
        } elseif ($showPreliminaryResults) {
            $component .= ' wcsf-get-preliminary-results-count';
        }

        if (!isset($question->viewType)) {
            return apply_filters('wcsf_question_fields', $output, $filterId, $questionId);
        }

        // check dependencies
        if ($question->filterType == 'manual') {
            $question->manualValues =
                self::filterManualValuesAccordingDependencies($filterId, $question->manualValues, $value);
        }

        if (strpos($question->viewType, 'number') !== false) {
            $unit = self::getSetting($questionId, 'numeric_unit');
            $numericStep = self::getSetting($questionId, 'numeric_step');

            // get min/max values
            switch ($question->filterType) {
                case 'price':
                    $boundary = Utils::getBoundaryPrices($products);
                    break;

                case 'attribute':
                    $boundary = Utils::getBoundaryAttributeValues($question->attribute, $products);
                    break;

                case 'meta':
                    $boundary = Utils::getBoundaryMetaValues($question->metaKey, $products);
            }

            if ($question->viewType == 'number') {
                $output[] = (object) [
                    'label' => $unit,
                    'name' => "value[$questionId][]",
                    'unit' => $unit,
                    'step' => $numericStep,
                    'value' => !empty($currentValue) ? reset($currentValue) : $boundary['min'],
                    'min' => isset($boundary['min']) ? $boundary['min'] : '',
                    'max' => isset($boundary['max']) ? $boundary['max'] : '',
                    'component' => $component
                ];
            }

            if (in_array($question->viewType, ['number-from', 'number-between'])) {
                $output['from'] = (object) [
                    'label' => self::getSetting($questionId, 'from_value_text'),
                    'name' => "value[$questionId][from]",
                    'unit' => $unit,
                    'step' => $numericStep,
                    'value' => isset($currentValue['from']) ? $currentValue['from'] : $boundary['min'],
                    'min' => isset($boundary['min']) ? $boundary['min'] : '',
                    'max' => isset($boundary['max']) ? $boundary['max'] : '',
                    'component' => $component
                ];
            }

            if (in_array($question->viewType, ['number-to', 'number-between'])) {
                $output['to'] = (object) [
                    'label' => self::getSetting($questionId, 'to_value_text'),
                    'name' => "value[$questionId][to]",
                    'unit' => $unit,
                    'step' => $numericStep,
                    'value' => isset($currentValue['to']) ? $currentValue['to'] : $boundary['max'],
                    'min' => isset($boundary['min']) ? $boundary['min'] : '',
                    'max' => isset($boundary['max']) ? $boundary['max'] : '',
                    'component' => $component
                ];
            }
        } else {
            if (in_array($question->filterType, ['tag', 'category', 'attribute'])) {
                $maxSubCategoriesLevel = 0;
                $startFromLevel = 0;
                $parentsOnly = [];
                $exclude = self::getSetting($questionId, 'excluded_terms');
                $include = self::getSetting($questionId, 'included_terms');
                $showCount = self::getSetting($questionId, 'show_count');
                $showDescription = self::getSetting($questionId, 'show_description');

                switch ($question->filterType) {
                    case 'tag':
                        $taxonomy = 'product_tag';
                        $thumbnailMetaKey = '_wcsf_thumbnail';
                        break;

                    case 'attribute':
                        $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                        $thumbnailMetaKey = '_wcsf_thumbnail';
                        break;

                    default:
                    case 'category':
                        $taxonomy = 'product_cat';
                        $thumbnailMetaKey = 'thumbnail_id';
                }

                $queryArgs = [
                    'taxonomy' => $taxonomy,
                    'hide_empty' => self::getSetting($questionId, 'hide_empty_terms')
                ];

                if ($exclude) {
                    $queryArgs['exclude'] = wp_parse_id_list($exclude);
                }

                if ($include) {
                    $queryArgs['include'] = wp_parse_id_list($include);
                    $queryArgs['orderby'] = 'include';
                }

                if ($question->filterType == 'category') {
                    $maxSubCategoriesLevel = self::getSetting($questionId, 'max_sub_categories_level');
                    $startFromLevel = self::getSetting($questionId, 'start_from_categories_level');
                    $queryArgs['parent'] = (int) self::getSetting($questionId, 'category_parent');
                    $categoryParentByQuestion = self::getSetting($questionId, 'category_parent_by_question');
                    $parentsOnly = isset($value[$categoryParentByQuestion])
                    && is_array($value[$categoryParentByQuestion])
                        ? array_map('intval', $value[$categoryParentByQuestion])
                        : [];
                }

                $queryArgs = apply_filters('wcsf_question_fields_query_args', $queryArgs, $filterId, $questionId);
                $terms = Utils::getTermsTree(
                    $queryArgs,
                    [
                        'flatten' => !in_array($question->viewType, ['radio', 'checkbox']),
                        'showLevelInName' => $question->viewType == 'select',
                        'maxLevel' => $maxSubCategoriesLevel,
                        'startFromLevel' => $startFromLevel,
                        'parentsOnly' => $parentsOnly
                    ]
                );

                if (self::getSetting($questionId, 'add_empty_value') && $question->viewType == 'select') {
                    $output[] = (object) [
                        'checked' => empty($currentValue),
                        'disabled' => false,
                        'name' => self::getSetting($questionId, 'empty_value_text'),
                        'value' => '',
                        'thumbnail' => null,
                        'component' => $component
                    ];
                }

                $output = array_merge(
                    $output,
                    self::getTermsFields([
                        'question' => $question,
                        'filterId' => $filterId,
                        'terms' => $terms,
                        'taxonomy' => $taxonomy,
                        'thumbnailMetaKey' => $thumbnailMetaKey,
                        'currentValue' => $currentValue,
                        'component' => $component,
                        'unmetValuesBehavior' => $unmetValuesBehavior,
                        'disableUnmet' => $disableUnmet,
                        'showCount' => $showCount,
                        'showDescription' => $showDescription,
                        'products' => $products,
                        'extraQueryArgs' => $extraQueryArgs
                    ])
                );
            } elseif ($question->filterType == 'manual') {
                // prepare manual values
                foreach ($question->manualValues as $valueKey => $value) {
                    $enabled = !$disableUnmet;
                    $isChecked = $currentValue
                        ? in_array($valueKey, $currentValue)
                        : (isset($value['pre-checked']) ? (bool) $value['pre-checked'] : false);

                    if ($disableUnmet) {
                        switch ($value['filter-by']) {
                            case 'category':
                                // check is term disabled
                                foreach ($products as $product) {
                                    if (has_term($value['items'], 'product_cat', $product)) {
                                        $enabled = true;

                                        break;
                                    }
                                }

                                break;
                            case 'product':
                                $enabled = !empty(array_intersect($products, $value['items']));
                                break;

                            case 'any':
                                $enabled = true;
                        }
                    }

                    if ($unmetValuesBehavior == 'hide' && !$isChecked && !$enabled) {
                        continue;
                    }

                    $output[] = (object) [
                        'checked' => $isChecked,
                        'disabled' => !$isChecked && !$enabled,
                        'name' => isset($value['name']) ? $value['name'] : '',
                        'value' => $valueKey,
                        'thumbnail' => isset($value['thumbnail']) ? $value['thumbnail'] : '',
                        'component' => $component
                    ];
                }
            } elseif ($question->filterType == 'meta') {
                // get meta values
                $values = Utils::getMetaValues($question->metaKey);

                foreach ($values as $productId => $value) {
                    $enabled = !$disableUnmet;
                    $isChecked = $currentValue ? in_array($value, $currentValue) : false;

                    if ($disableUnmet) {
                        $enabled = in_array($productId, $products);
                    }

                    if ($unmetValuesBehavior == 'hide' && !$isChecked && !$enabled) {
                        continue;
                    }

                    $output[] = (object) [
                        'checked' => $isChecked,
                        'disabled' => !$isChecked && !$enabled,
                        'name' => $value,
                        'value' => $value,
                        'thumbnail' => null,
                        'component' => $component
                    ];
                }
            }

            // add any product value
            if (self::getSetting($questionId, 'add_any_product_value')
                && in_array($question->viewType, $viewsWithAnyProductValue)
            ) {
                $position = self::getSetting($questionId, 'any_product_value_position');
                $anyProductValue = (object) [
                    'checked' => $currentValue
                        ? in_array(0, $currentValue)
                        : self::getSetting($questionId, 'any_product_value_checked'),
                    'disabled' => false,
                    'name' => self::getSetting($questionId, 'any_product_value_text'),
                    'value' => 0,
                    'thumbnail' => self::getSetting($questionId, 'any_product_value_thumbnail'),
                    'component' => $component
                ];

                if (isset($output[$position])) {
                    $output = array_merge(
                        array_slice($output, 0, $position, true),
                        [$anyProductValue],
                        array_slice($output, $position, count($output) - $position, true)
                    );
                } else {
                    $output[$position] = $anyProductValue;
                }
            }
        }

        return apply_filters('wcsf_question_fields', $output, $filterId, $questionId);
    }

    /**
     * Get fields by terms array
     *
     * @param array $args
     *
     * @return array
     */
    public static function getTermsFields($args = [])
    {
        $defaults = [
            'question' => null,
            'filterId' => null,
            'terms' => [],
            'taxonomy' => 'product_cat',
            'thumbnailMetaKey' => 'thumbnail_id',
            'currentValue' => null,
            'component' => 'wcsf-field-input',
            'unmetValuesBehavior' => 'disable',
            'disableUnmet' => false,
            'showCount' => false,
            'showDescription' => false,
            'products' => [],
            'extraQueryArgs' => []
        ];

        $args = array_replace($defaults, $args);
        $output = [];

        foreach ($args['terms'] as $term) {
            if (!is_object($term)) {
                continue;
            }

            $enabled = !$args['disableUnmet'];
            $isPreChecked = (bool) get_term_meta($term->term_id, '_wcsf_pre_checked', true);
            $isChecked = $args['currentValue'] ? in_array($term->term_id, $args['currentValue']) : $isPreChecked;
            $thumbnail = $args['thumbnailMetaKey']
                ? get_term_meta($term->term_id, $args['thumbnailMetaKey'], true)
                : '';

            if ($args['disableUnmet']) {
                $children = get_term_children($term->term_id, $args['taxonomy']);

                // check is term disabled
                foreach ($args['products'] as $product) {
                    if (has_term(array_merge($children, [$term->term_id]), $args['taxonomy'], $product)) {
                        $enabled = true;

                        break;
                    }
                }
            }

            if ($args['unmetValuesBehavior'] == 'hide' && !$isChecked && !$enabled) {
                continue;
            }

            $outputItem = [
                'checked' => $isChecked,
                'currentValue' => $args['currentValue'],
                'disabled' => !$isChecked && !$enabled,
                'name' => $term->name,
                'value' => $term->term_id,
                'thumbnail' => $thumbnail,
                'component' => $args['component']
            ];

            if ($args['showCount']) {
                if (strpos($args['question']->viewType, 'multi-choice') !== false) {
                    $questionValue = [$args['question']->id => [$outputItem['value'] => 1]];
                } else {
                    $questionValue = [$args['question']->id => [$outputItem['value']]];
                }

                $outputItem['count'] = Filter::getPreliminaryResultsCount($args['filterId'], $questionValue, $args['extraQueryArgs']);
            }

            if ($args['showDescription']) {
                $outputItem['description'] = $term->description;
            }

            $outputItem = (object) $outputItem;

            if (isset($term->children) && !empty($term->children)) {
                $outputItem->children = self::getTermsFields(array_replace($args, ['terms' => $term->children]));
            }

            $output[] = $outputItem;
        }

        return $output;
    }
    // </editor-fold>

    /**
     * Pass the value to all bond questions in filters using them
     *
     * @param integer $filterId
     * @param integer $questionId
     * @param mixed $value
     */
    public static function passValueToBond($filterId, $questionId, $value)
    {
        $bondQuestions = array_filter((array) self::getSetting($questionId, 'bond_questions'));

        if (empty($bondQuestions)) {
            return;
        }

        $filtersIds = Filter::getPostsIds(['post__not_in' => [$filterId]]);

        foreach (array_keys($filtersIds) as $id) {
            $questions = Filter::getQuestionsList($id);

            foreach ($bondQuestions as $bondQuestion) {
                if (!in_array($bondQuestion, array_keys($questions)) || $bondQuestion == $questionId) {
                    continue;
                }

                Storage::set(Filter::$valuesSessionKey, $id, $value, $bondQuestion);
            }
        }
    }
}
