/* WooCommerce Step Filter Question
 * Original author: Alex Troll
 * Further changes, comments: mail@troll-winner.com
 */

(function (root, factory) {
    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['jquery'], factory);
    } else if (typeof exports === 'object'
        && typeof module !== 'undefined'
        && typeof require === 'function'
    ) {
        module.exports = factory(require('jquery'));
    } else {
        factory(root.jQuery);
    }
})(this, function ($) {
    'use strict';

    const pluginName = 'wcsfQuestion';
    const defaults = {};
    const $document = $(document);

    const Plugin = function (element, options) {
        this.element = element;
        this.options = $.extend({}, defaults, options);

        return this.init();
    };

    /**
     * Init the instance
     * @returns {this} self instance
     */
    Plugin.prototype.init = function () {
        this.$element = $(this.element);
        this.$settings = this.$element.find('[data-component~="wcsf-setting"]');
        this.$manualValues = this.$element.find('[data-component~="wcsf-question-manual-values"]');
        this.$filterType = this.$element.find('#_filter_type');
        this.$filterTypes = this.$filterType.find('option');
        this.$viewType = this.$element.find('#_view_type');
        this.ajaxUrl = this.$element.data('ajax-url');

        this.toggleFields();

        if ($.fn.sortable) {
            // init sortable
            this.$manualValues.sortable({
                items: '[data-component~="wcsf-question-manual-value"]',
                distance: 50,
                update: () => this.recalculateIds(),
                cancel: 'input,textarea,button,select,option,.wcsf-modal'
            });
        }

        // disable not active manual source inputs
        this.$element
            .find('[data-component~="wcsf-question-manual-value-filter"].is-closed')
            .find('[data-component="wcsf-question-manual-value-filter-values"]')
            .attr('disabled', true);

        this.$element.find('[data-component~="wcsf-question-manual-value-filter-values-item-source"]').trigger('change'); // eslint-disable-line

        return this.initEventListeners();
    };

    /**
     * Add required event listeners
     * @returns {this} self instance
     */
    Plugin.prototype.initEventListeners = function () {
        const _this = this;

        this.$element.on('change', '#_view_type, #_filter_type', () => this.toggleFields());

        // add manual filter row
        this.$element.on('click', '[data-component~="wcsf-question-add-value"]', function (event) {
            event.preventDefault();

            this.setAttribute('disabled', 'disabled');

            return _this.addValue().then(() => this.removeAttribute('disabled'));
        });

        // remove manual filter row
        this.$element.on('click', '[data-component~="wcsf-question-manual-value-remove"]', function (event) {
            event.preventDefault();

            const $value = _this.$element.find('[data-component~="wcsf-question-manual-value"]');
            const $valueToRemove = $value.filter(`[data-index="${this.getAttribute('data-index')}"]`);

            if ($value.length !== 1) {
                $valueToRemove.remove();

                _this.recalculateIds();
            }
        });

        // toggle manual filter source sections
        this.$element.on('change', '[data-component~="wcsf-question-manual-value-filter-by"]', function () {
            const $section = $(this)
                .closest('[data-component~="wcsf-modal"]')
                .find('[data-component~="wcsf-question-manual-value-filter"]');

            $section.addClass('is-closed')
                .find('[data-component="wcsf-question-manual-value-filter-values"]')
                .attr('disabled', true);

            $section.filter(`[data-id="${this.getAttribute('data-id')}"]`).removeClass('is-closed')
                .find('[data-component="wcsf-question-manual-value-filter-values"]')
                .removeAttr('disabled');
        });

        // manual value item source change
        this.$element.on('change', '[data-component~="wcsf-question-manual-value-filter-values-item-source"]', function () { // eslint-disable-line
            const $source = $(this);
            const $row = $source.closest('[data-component="wcsf-data-table-item"]');
            const $search = $row.find('[data-component~="wcsf-question-manual-value-filter-values-item-terms"]');
            const value = $source.val();

            switch (value) {
                case 'product_cat':
                    $search.data('action', 'woocommerce_json_search_category');
                    break;

                case 'product_tag':
                    $search.data('action', 'woocommerce_json_search_tags');
                    break;

                default:
                    $search.data('action', `woocommerce_json_search_${value}`);
            }
        });
    };

    /**
     * Show/hide fields
     * @returns {this} self instance
     */
    Plugin.prototype.toggleFields = function () {
        let filterType = this.$filterType.val();
        const viewType = this.$viewType.val();
        const numericViews = ['number', 'number-from', 'number-to', 'number-between'];
        const checkboxViews = ['checkbox', 'image-checkbox', 'chips-checkbox', 'button-checkbox'];
        const multiChoiceViews = ['multi-choice-checkbox', 'multi-choice-range'];
        const imagedViews = ['image-radio', 'image-checkbox'];
        const singleValueViews = ['select', 'radio', 'image-radio', 'chips-radio', 'button-radio'];
        const terms = ['attribute', 'tag', 'category'];

        const typesVisibilityRules = {
            price: ['formula', ...numericViews],
            manual: [...singleValueViews, ...checkboxViews],
            category: [...checkboxViews, ...singleValueViews, ...multiChoiceViews],
            tag: [...checkboxViews, ...singleValueViews, ...multiChoiceViews],
            meta: [
                'formula',
                'select',
                'radio',
                'checkbox',
                'chips-radio',
                'chips-checkbox',
                'button-radio',
                'button-checkbox',
                ...multiChoiceViews,
                ...numericViews
            ]
        };

        const settingsVisibilityRulesByView = {
            formula: ['formula'],
            formula_operator: ['formula'],
            formula_variables: ['formula'],
            formula_name_aliases: ['formula'],
            formula_value_aliases: ['formula'],
            output_formula_value: ['formula'],
            formula_value_text: ['formula'],
            from_value_text: ['number-from', 'number-between'],
            to_value_text: ['number-to', 'number-between'],
            numeric_unit: numericViews,
            numeric_step: numericViews,
            enable_slider: numericViews,
            show_slider_tooltips: numericViews,
            show_inputs: numericViews,
            label_as_placeholder: numericViews,
            attribute_value_source: ['formula', ...numericViews],
            add_any_product_value: singleValueViews,
            any_product_value_text: singleValueViews,
            any_product_value_checked: singleValueViews,
            any_product_value_position: singleValueViews,
            any_product_value_thumbnail: ['image-radio'],
            add_empty_value: ['select'],
            empty_value_text: ['select'],
            alignment: imagedViews,
            thumbnail_size: imagedViews,
            thumbnail_width: imagedViews,
            multi_choice_values: multiChoiceViews,
            multi_choice_undefined_value_text: multiChoiceViews,
            multi_choice_possible_value_text: multiChoiceViews,
            multi_choice_required_value_text: multiChoiceViews,
            multi_choice_excluded_value_text: multiChoiceViews,
            hide_empty_terms: [...checkboxViews, ...singleValueViews],
            included_terms: [...checkboxViews, ...singleValueViews],
            excluded_terms: [...checkboxViews, ...singleValueViews],
            unmet_values_behavior: [...checkboxViews, ...singleValueViews, ...numericViews],
            show_count: [...checkboxViews, ...singleValueViews, ...multiChoiceViews],
            show_description: [...checkboxViews, ...singleValueViews, ...multiChoiceViews],
            action_on_click: [...checkboxViews, ...singleValueViews, ...multiChoiceViews],
            additional_get_param: [...checkboxViews, ...singleValueViews, ...numericViews],
            additional_get_param_term_source: [...checkboxViews, ...singleValueViews],
            min_values_selected: checkboxViews,
            min_values_selected_message: checkboxViews,
            max_values_selected: checkboxViews,
            max_values_selected_message: checkboxViews,
            relation: checkboxViews
        };

        const settingsVisibilityRulesByType = {
            attribute: ['attribute'],
            attribute_value_source: ['attribute'],
            add_any_product_value: [...terms, 'meta'],
            any_product_value_text: [...terms, 'meta'],
            any_product_value_checked: [...terms, 'meta'],
            any_product_value_position: [...terms, 'meta'],
            any_product_value_thumbnail: [...terms, 'meta'],
            additional_get_param_term_source: terms,
            category_parent: ['category'],
            category_parent_by_question: ['category'],
            start_from_categories_level: ['category'],
            max_sub_categories_level: ['category'],
            hide_empty_terms: terms,
            included_terms: terms,
            excluded_terms: terms,
            show_count: terms,
            show_description: terms,
            manual: ['manual'],
            meta_key: ['meta']
        };

        // toggle types according to the selected view
        this.$filterTypes.each(function () {
            if (!typesVisibilityRules.hasOwnProperty(this.getAttribute('value'))) {
                return null;
            }

            const hide = $.inArray(viewType, typesVisibilityRules[this.getAttribute('value')]) === -1;

            if (hide) {
                this.setAttribute('disabled', 'disabled');
            } else {
                this.removeAttribute('disabled');
            }

            return this;
        });

        // reset select if disabled option was selected
        if (!this.$filterType.val()) {
            this.$filterType.val(this.$filterType.find('option:not(disabled):eq(0)').attr('value'));
            filterType = this.$filterType.val();
        }

        // toggle settings according to the selected view
        this.$settings.each(function () {
            const $element = $(this);

            if (!settingsVisibilityRulesByView.hasOwnProperty(this.getAttribute('data-key'))) {
                return null;
            }

            const hide = $.inArray(viewType, settingsVisibilityRulesByView[this.getAttribute('data-key')]) === -1;

            $element[hide ? 'addClass' : 'removeClass']('is-hidden-by-view');

            return this;
        });

        // toggle settings according to the selected type
        this.$settings.each(function () {
            const $element = $(this);

            if (!settingsVisibilityRulesByType.hasOwnProperty(this.getAttribute('data-key'))) {
                return null;
            }

            const hide = $.inArray(filterType, settingsVisibilityRulesByType[this.getAttribute('data-key')]) === -1;

            $element[hide ? 'addClass' : 'removeClass']('is-hidden-by-type');

            return this;
        });

        return this;
    };

    /**
     * Makes the ajax-request and add manual value fields row
     * @returns {Promise} ajax request
     */
    Plugin.prototype.addValue = function () {
        const data = {
            action: 'wcsfAddQuestionValue',
            postId: this.$element.data('post-id'),
            index: this.$manualValues.find('[data-component~="wcsf-question-manual-value"]').length
        };

        return $.get(
            this.ajaxUrl,
            data,
            (response) => {
                this.$manualValues.append(response);

                const $element = this.$manualValues.find('[data-component~="wcsf-question-manual-value"]:last');

                $document.trigger('added.value.question.wcsf', [this, $element]);
            }
        );
    };

    /**
     * Recalculate input names indexes
     * @returns {Object} jQuery elements
     */
    Plugin.prototype.recalculateIds = function () {
        return this.$element.find('[data-component~="wcsf-question-manual-value"]').each(function (index) {
            const $element = $(this);

            // change item index
            this.setAttribute('data-index', index);

            // change item children index
            $element.find('[data-index]').each(function () {
                return this.setAttribute('data-index', index);
            });

            return $element.find(':input').each(function () {
                const name = this.getAttribute('name');

                if (name) {
                    const number = index.toString().split('').reverse().join(''); // eslint-disable-line

                    // replace the first array key from the end
                    this.setAttribute(
                        'name',
                        name
                            .split('')
                            .reverse()
                            .join('')
                            .replace(/]\d+\[/, `]${number}[`)
                            .split('')
                            .reverse()
                            .join('')
                    );
                }
            });
        });
    };

    $.fn[pluginName] = function (options) {
        return this.each(function () {
            if (!$.data(this, pluginName)) {
                $.data(this, pluginName, new Plugin(this, options));
            }
        });
    };

    const init = () => $('[data-component~="wcsf-question"]').each(function () {
        return $(this).wcsfQuestion();
    });

    $document.ready(() => init());
    $document.on('init.wcsfQuestion', () => init());
});
