/* WooCommerce Step Filter global instance and main event handlers
 * Original author: mail@troll-winner.ru
 * Further changes, comments: mail@troll-winner.ru
 */

(function (root, factory) {
    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['jquery'], factory);
    } else if (typeof exports === 'object' && typeof module !== 'undefined' && typeof require === 'function') {
        module.exports = factory(require('jquery'));
    } else {
        factory(root.jQuery);
    }
})(this, function ($) {
    'use strict';

    const wcsf = {
        windowNode: window,
        documentNode: document,
        $window: $(window),
        $document: $(document),
        stickyObserverElements: null,

        init: function () {
            if (typeof $.fn.wcsf === 'undefined') {
                this.windowNode.console.error('jQuery.fn.wcsf is not exist');

                return this;
            }

            this.$document.find('[data-component~="wcsf"]').each(function () {
                const $element = $(this);

                return $element.wcsf($element.data('options') || {});
            });

            return this;
        },

        tooltip: (function () {
            let hideTimeout = null;

            return {
                template: '<span class="wcsf-tooltip" data-component="wcsf-tooltip" role="tooltip" aria-hidden="false">%text%</span>', // eslint-disable-line
                fadeDuration: 200,
                hideDelay: 2500,
                hideAll: function () {
                    for (let element of wcsf.documentNode.querySelectorAll('[data-component~="wcsf-tooltip"]')) {
                        this.hide(element);
                    }

                    return this;
                },
                hide: function (element) {
                    if (!element) {
                        return this;
                    }

                    element.style.transition = this.fadeDuration + 'ms';
                    element.style.opacity = 0;

                    setTimeout(() => element.remove(), this.fadeDuration);

                    return this;
                },
                show: function (element, text) {
                    this.hideAll();

                    clearTimeout(hideTimeout);

                    let tooltip = wcsf.documentNode.createElement('span');

                    tooltip.innerHTML = this.template.replace('%text%', text);
                    tooltip = tooltip.querySelector(':first-child');

                    element.before(tooltip);

                    hideTimeout = setTimeout(() => this.hide(tooltip), this.hideDelay);

                    return this;
                }
            };
        })(),

        updateStickyObserverElementsState: function () {
            if (!this.stickyObserverElements || this.stickyObserverElements.length <= 0) {
                return;
            }

            for (let element of this.stickyObserverElements) {
                if (!element.parentElement) {
                    continue;
                }

                const parentTop = element.parentElement.getBoundingClientRect().top;
                const top = Math.round(element.getBoundingClientRect().top - parentTop);

                element.style.position = 'static';
                element.toggleAttribute(
                    'stuck',
                    top !== Math.round(element.getBoundingClientRect().top - parentTop)
                );

                element.style.position = '';
                element.dataset.prevClientTop = String(top);
            }
        }
    };

    if (typeof document.wcsf === 'undefined') {
        document.wcsf = wcsf;
    }

    wcsf.$window.on('load', () => {
        // init sticky observer elements
        wcsf.stickyObserverElements = wcsf.documentNode.querySelectorAll('[data-component~="wcsf-sticky-observer"]');

        if (!wcsf.stickyObserverElements || wcsf.stickyObserverElements.length <= 0) {
            return;
        }

        wcsf.updateStickyObserverElementsState();
        wcsf.windowNode.addEventListener('resize', () => setTimeout(() => wcsf.updateStickyObserverElementsState(), 0));
        wcsf.windowNode.addEventListener('scroll', () => {
            for (let element of wcsf.stickyObserverElements) {
                if (!element.parentElement) {
                    continue;
                }

                const top = Math
                    .round(element.getBoundingClientRect().top - element.parentElement.getBoundingClientRect().top);

                element.toggleAttribute('stuck', Number(element.dataset.prevClientTop) !== top);
                element.dataset.prevClientTop = String(top);
            }
        }, false);
    });

    // init plugin
    wcsf.$document.ready(() => wcsf.init());
    wcsf.$document.on('init.wcsf', () => wcsf.init());

    // get preliminary results count manual trigger
    wcsf.$document.on('getPreliminaryResultsCount.wcsf', (event, instance, data, element) => {
        if (data.tooltip) {
            wcsf.tooltip.show(element, data.tooltip);
        }
    });

    // ajax event is completed
    wcsf.$document.on('ajaxCompleted.wcsf', (event, instance, response) => {
        // re-init sticky observer elements
        wcsf.stickyObserverElements = instance.element.querySelectorAll('[data-component~="wcsf-sticky-observer"]');
        wcsf.updateStickyObserverElementsState();

        // noUi slider init
        wcsf.documentNode.dispatchEvent(new CustomEvent('init.nouislider.wcsf'));

        // product wizard variations init
        wcsf.$document.trigger('init.variationForm.wcpw');

        if (response.hasError) {
            const message = instance.element.querySelectorAll('[data-component~="wcsf-message"]');

            // scroll to the message
            if (message && !instance.isScrolledIntoView(message)) {
                instance.scrollToElement(message, instance.options.scrollingUpGap);
            }

            // vibration signal
            instance.vibrate();
        }
    });

    // hide tooltips on click
    wcsf.$document.on('click.hideAll.tooltip.wcsf', '[data-component~="wcsf"]', () => wcsf.tooltip.hideAll());

    // contact form 7 hooks
    if (typeof wpcf7 !== 'undefined') {
        wcsf.$document.on('ajaxCompleted.wcsf', (event, instance) => {
            const form = instance.element.querySelector('.wpcf7-form');

            if (form) {
                if (typeof wpcf7.initForm !== 'undefined') {
                    wpcf7.initForm($(form));
                } else if (typeof wpcf7.init !== 'undefined') {
                    wpcf7.init(form);
                }
            }
        });
    }
});
