/* WooCommerce Step Filter global instance and main event handlers
 * Original author: mail@troll-winner.ru
 * Further changes, comments: mail@troll-winner.ru
 */

import { WCSF } from 'wcsf';

const wcsf = {
    windowNode: window,
    documentNode: document,
    stickyObserverElements: null,

    init: function () {
        if (typeof WCSF === 'undefined') {
            this.windowNode.console.error('WCSF class is not exist');

            return this;
        }

        for (let element of this.documentNode.querySelectorAll('[data-component~="wcsf"]')) {
            element.wcsf = new WCSF(element, JSON.parse(element.getAttribute('data-options')) || {});
            element.wcsf.init();
        }

        return this;
    },

    tooltip: (function () {
        let hideTimeout = null;

        return {
            template: '<span class="wcsf-tooltip" data-component="wcsf-tooltip" role="tooltip" aria-hidden="false">%text%</span>', // eslint-disable-line
            fadeDuration: 200,
            hideDelay: 2500,
            hideAll: function () {
                for (let element of wcsf.documentNode.querySelectorAll('[data-component~="wcsf-tooltip"]')) {
                    this.hide(element);
                }

                return this;
            },
            hide: function (element) {
                if (!element) {
                    return this;
                }

                element.style.transition = wcsf.fadeDuration + 'ms';
                element.style.opacity = 0;

                setTimeout(() => element.remove(), this.fadeDuration);

                return this;
            },
            show: function (element, text) {
                this.hideAll();

                clearTimeout(hideTimeout);

                let tooltip = wcsf.documentNode.createElement('span');

                tooltip.innerHTML = this.template.replace('%text%', text);
                tooltip = tooltip.querySelector(':first-child');

                element.before(tooltip);

                hideTimeout = setTimeout(() => this.hide(tooltip), this.hideDelay);

                return this;
            }
        };
    })(),

    updateStickyObserverElementsState: function () {
        if (!this.stickyObserverElements || this.stickyObserverElements.length <= 0) {
            return;
        }

        for (let element of this.stickyObserverElements) {
            if (!element.parentElement) {
                continue;
            }

            const parentTop = element.parentElement.getBoundingClientRect().top;
            const top = Math.round(element.getBoundingClientRect().top - parentTop);

            element.style.position = 'static';
            element.toggleAttribute(
                'stuck',
                top !== Math.round(element.getBoundingClientRect().top - parentTop)
            );

            element.style.position = '';
            element.dataset.prevClientTop = String(top);
        }
    }
};

wcsf.windowNode.addEventListener('load', () => {
    // init sticky observer elements
    wcsf.stickyObserverElements = wcsf.documentNode.querySelectorAll('[data-component~="wcsf-sticky-observer"]');

    if (!wcsf.stickyObserverElements || wcsf.stickyObserverElements.length <= 0) {
        return;
    }

    wcsf.updateStickyObserverElementsState();
    wcsf.windowNode.addEventListener('resize', () => setTimeout(() => wcsf.updateStickyObserverElementsState(), 0));
    wcsf.windowNode.addEventListener('scroll', () => {
        for (let element of wcsf.stickyObserverElements) {
            if (!element.parentElement) {
                continue;
            }

            const top = Math
                .round(element.getBoundingClientRect().top - element.parentElement.getBoundingClientRect().top);

            element.toggleAttribute('stuck', Number(element.dataset.prevClientTop) !== top);
            element.dataset.prevClientTop = String(top);
        }
    }, false);
});

// init plugin
wcsf.documentNode.addEventListener('DOMContentLoaded', () => wcsf.init());
wcsf.documentNode.addEventListener('init.wcsf', () => wcsf.init());

// get preliminary results count manual trigger
wcsf.documentNode.addEventListener('getPreliminaryResultsCount.wcsf', (event) => {
    if (event.detail.data && event.detail.data.tooltip) {
        wcsf.tooltip.show(event.detail.element, event.detail.data.tooltip);
    }
});

// ajax event is completed
wcsf.documentNode.addEventListener('ajaxCompleted.wcsf', (event) => {
    const instance = event.detail.instance;

    // re-init sticky observer elements
    wcsf.stickyObserverElements = instance.element.querySelectorAll('[data-component~="wcsf-sticky-observer"]');
    wcsf.updateStickyObserverElementsState();

    // noUi slider init
    wcsf.documentNode.dispatchEvent(new CustomEvent('init.nouislider.wcsf'));

    if (event.detail.response.hasError) {
        const message = instance.element.querySelectorAll('[data-component~="wcsf-message"]');

        // scroll to the message
        if (message && !instance.isScrolledIntoView(message)) {
            instance.scrollToElement(message, instance.options.scrollingUpGap);
        }

        // vibration signal
        instance.vibrate();
    }
});

// hide tooltips on click
wcsf.documentNode.addEventListener('click', () => wcsf.tooltip.hideAll());

// contact form 7 hooks
if (typeof wpcf7 !== 'undefined') {
    wcsf.documentNode.addEventListener('ajaxCompleted.wcsf', (event) => {
        const form = event.detail.instance.element.querySelector('.wpcf7-form');

        if (form) {
            if (typeof wpcf7.initForm !== 'undefined') {
                wpcf7.initForm($(form));
            } else if (typeof wpcf7.init !== 'undefined') {
                wpcf7.init(form);
            }
        }
    });
}

export default wcsf;
