/* WooCommerce Step Filter
 * Original author: Alex Troll
 * Further changes, comments: mail@troll-winner.com
 */

class WCSF {
    // <editor-fold desc="Core">
    constructor(element, options = {}) {
        const defaults = {
            documentNode: document,
            windowNode: window,
            rootSelector: 'html, body',
            preliminaryResultsCountRequestInterval: 1000,
            ajaxActions: {
                submit: 'WCSFSubmit',
                submitAndGoNext: 'WCSFSubmitAndGoNext',
                submitAndGoResults: 'WCSFSubmitAndGoResults',
                submitAndGoTo: 'WCSFSubmitAndGoTo',
                getStep: 'WCSFGetStep',
                reset: 'WCSFReset',
                getPreliminaryResultsCount: 'WCSFGetPreliminaryResultsCount'
            }
        };

        this.element = element;
        this.customOptions = options;
        this.options = Object.assign({}, defaults, options);
    }

    /**
     * Init the instance
     * @returns {this} self instance
     */
    init() {
        this.controlsPreliminaryResults = this.element.querySelector('[data-component~="wcsf-controls-preliminary-results"]'); // eslint-disable-line max-len
        this.getPreliminaryResultsCountTimeout = null;
        this.getPreliminaryResultsCountPromise = null;
        this.preventAjaxRequest = false;
        this.eventListeners = [];

        return this.initEventListeners().triggerEvent('launched.wcsf', {instance: this});
    }

    /**
     * Delegate an event listener to a target
     * @param {String} action - event action name
     * @param {String} selector - target element selector
     * @param {Function} callback - function to fire
     * @param {Object} options - listener options
     * @returns {this} self instance
     */
    delegateEventListener(action, selector, callback, options = {capture: false}) {
        const actionName = action.split('.')[0];
        const handler = function (event) {
            let target = event.target;

            while (target && !target.matches(selector) && target !== this) {
                target = target.parentElement;
            }

            if (target && target.matches(selector)) {
                callback.call(target, event);
            }

            return this;
        };

        this.eventListeners.push({
            action,
            actionName,
            selector,
            handler
        });

        this.element.addEventListener(actionName, handler, options);

        return this;
    }

    /**
     * Un-delegate an event listener from the target
     * @param {String} action - event action name
     * @param {String} selector - target element selector
     * @param {Object} options - listener options
     * @returns {this} self instance
     */
    unDelegateEventListener(action, selector = '', options = {capture: false}) {
        const events = this.eventListeners
            .filter((item) => item.action === action && (!selector || item.selector === selector));

        if (events[0]) {
            this.element.removeEventListener(events[0].actionName, events[0].handler, options);
        }

        return this;
    }

    /**
     * Add required event listeners
     * @returns {this} self instance
     */
    initEventListeners() {
        const _this = this;

        // browser history handlers
        this.options.windowNode.addEventListener('popstate', (event) => this.popState(event), false);

        // reset click
        this.delegateEventListener('click.reset.wcsf', '[data-component~="wcsf-reset"]', (event) => {
            event.preventDefault();

            return this.reset();
        });

        // value item remove click
        this.delegateEventListener('click.remove.value.wcsf', '[data-component~="wcsf-value-item-remove"]', function (event) { // eslint-disable-line
            event.preventDefault();

            return _this.reset({resetStepId: this.getAttribute('data-id')});
        });

        // submit and go to click
        this.delegateEventListener(
            'click.submitAndGoTo.wcsf',
            '[data-component~="wcsf-submit-and-go-to"]:not(input), [data-component~="wcsf-submit-and-go-to"]:not(select)', // eslint-disable-line
            function (event) {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                event.preventDefault();

                return _this.submitAndGoTo({stepId: this.getAttribute('data-id')});
            }
        );

        // submit and go to change
        this.delegateEventListener(
            'change.submitAndGoTo.wcsf',
            'input[data-component~="wcsf-submit-and-go-to"], select[data-component~="wcsf-submit-and-go-to"]', // eslint-disable-line
            function () {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                return _this.submitAndGoTo({stepId: this.getAttribute('data-id')});
            }
        );

        // submit and go next click
        this.delegateEventListener(
            'click.submitAndGoNext.wcsf',
            '[data-component~="wcsf-submit-and-go-next"]:not(input), [data-component~="wcsf-submit-and-go-next"]:not(select)', // eslint-disable-line
            function (event) {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                event.preventDefault();

                return _this.submitAndGoNext({stepId: this.getAttribute('data-id')});
            }
        );

        // submit and go next change
        this.delegateEventListener(
            'change.submitAndGoNext.wcsf',
            'input[data-component~="wcsf-submit-and-go-next"], select[data-component~="wcsf-submit-and-go-next"]',
            function () {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                return _this.submitAndGoNext({stepId: this.getAttribute('data-id')});
            }
        );

        // submit and go to results click
        this.delegateEventListener(
            'click.submitAndGoResults.wcsf',
            '[data-component~="wcsf-submit-and-go-results"]:not(input), [data-component~="wcsf-submit-and-go-results"]:not(select)', // eslint-disable-line
            function (event) {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                event.preventDefault();

                return _this.submitAndGoResults({stepId: this.getAttribute('data-id')});
            }
        );

        // submit and go to results change
        this.delegateEventListener(
            'change.submitAndGoResults.wcsf',
            'input[data-component~="wcsf-submit-and-go-results"], select[data-component~="wcsf-submit-and-go-results"]',
            function () {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                return _this.submitAndGoResults({stepId: this.getAttribute('data-id')});
            }
        );

        // submit click
        this.delegateEventListener(
            'click.submit.wcsf',
            '[data-component~="wcsf-submit"]:not(input), [data-component~="wcsf-submit"]:not(select)',
            function (event) {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                event.preventDefault();

                return _this.submit({stepId: this.getAttribute('data-id')});
            }
        );

        // submit change
        this.delegateEventListener(
            'change.submit.wcsf',
            'input[data-component~="wcsf-submit"], select[data-component~="wcsf-submit"]',
            function () {
                if (!_this.checkFromValidity()) {
                    return this;
                }

                return _this.submit({stepId: this.getAttribute('data-id')});
            }
        );

        // results click
        this.delegateEventListener('click.results.wcsf', '[data-component~="wcsf-results"]', function () {
            if (_this.checkFromValidity()) {
                // need to submit the form via the usual request (without js)
                _this.element.querySelector('[data-component~="wcsf-form"]')
                    .setAttribute('data-component', 'wcsf-form-off');
            }
        });

        // none action button click
        this.delegateEventListener('click.results.wcsf', '[data-component~="wcsf-none"]', () => false);

        // submit form
        this.delegateEventListener('submit.form.wcsf', '[data-component~="wcsf-form"]', function (event) {
            if (!_this.checkFromValidity()) {
                return this;
            }

            event.preventDefault();

            return _this.submitAndGoNext();
        });

        // get step click
        this.delegateEventListener('click.getStep.wcsf', '[data-component~="wcsf-get-step"]', function (event) {
            event.preventDefault();

            return _this.getStep({
                stepId: this.getAttribute('data-id'),
                extraQueryArgs: {paged: this.getAttribute('data-page') || 1}
            });
        });

        // products order change
        this.delegateEventListener('change.order.wcsf', '[data-component~="wcsf-ordering-select"]', function (event) {
            event.preventDefault();

            return _this.getStep({extraQueryArgs: {orderby: this.value}});
        });

        // get preliminary results
        this.delegateEventListener(
            'change.getPreliminaryResults.wcsf changed.getPreliminaryResults.wcsf',
            '[data-component~="wcsf-get-preliminary-results-count"]',
            function (event) {
                event.preventDefault();

                return _this.getPreliminaryResultsCount(this);
            }
        );

        return this;
    }

    /**
     * Makes an ajax-request
     * @param {Object} requestData - object of data
     * @param {Object} options - object of method options
     * @returns {Promise} ajax request
     */
    ajaxRequest(requestData, options = {}) {
        this.preventAjaxRequest = false;

        const instance = this;
        const formData = requestData instanceof FormData ? requestData : new FormData();
        const defaultOptions = {
            updateQueryArgs: true,
            queryArgs: {},
            processData: false,
            contentType: false,
            cache: 'default',
            method: 'post',
            scrollingTopOnUpdate: Boolean(this.options.scrollingTopOnUpdate),
            scrollingGap: Number(this.options.scrollingUpGap),
            scrollingBehavior: 'smooth',
            scrollingToElement: null,
            errorsAlerting: true,
            errorsLogging: true
        };

        options = this.extendObject(defaultOptions, options);

        // pass extra args
        if (!(requestData instanceof FormData)) {
            for (let key in requestData) {
                if (!formData.has(key) && requestData.hasOwnProperty(key) && requestData[key] !== null) {
                    formData.append(
                        key, typeof requestData[key] !== 'string' ? JSON.stringify(requestData[key]) : requestData[key]
                    );
                }
            }
        }

        // save extra parameters
        for (let key in this.customOptions) {
            if (!formData.has(key) && this.customOptions.hasOwnProperty(key)) {
                formData.append(
                    key,
                    typeof this.customOptions[key] !== 'string'
                        ? JSON.stringify(this.customOptions[key])
                        : this.customOptions[key]
                );
            }
        }

        // save passed query args to request
        if (options.queryArgs) {
            for (let key in options.queryArgs) {
                if (options.queryArgs.hasOwnProperty(key)) {
                    formData.append(key, options.queryArgs[key]);
                }
            }
        }

        clearTimeout(this.getPreliminaryResultsCountTimeout);

        if (this.getPreliminaryResultsCountPromise && this.getPreliminaryResultsCountPromise.abort) {
            this.getPreliminaryResultsCountPromise.abort();
        }

        this.triggerEvent('ajaxRequest.wcsf', {instance, formData, options});

        if (this.preventAjaxRequest) {
            this.triggerEvent('ajaxPrevent.wcsf', {instance, formData, options});

            return Promise.resolve();
        }

        this.element.classList.add('is-loading');
        this.element.setAttribute('aria-live', 'polite');
        this.element.setAttribute('aria-busy', 'true');

        return this.options.windowNode
            .fetch(
                this.options.ajaxUrl,
                {
                    method: options.method,
                    body: formData,
                    cache: options.cache
                })
            .then((response) => response.json())
            .then((response) => {
                this.triggerEvent('ajaxSuccess.wcsf', {instance, response, formData, options});

                // make redirect only
                if (response.finalRedirectUrl) {
                    this.triggerEvent('finalRedirect.wcsf', [this, formData, response]);

                    this.options.documentNode.location = response.finalRedirectUrl;

                    return this;
                }

                this.element.classList.remove('is-loading');
                this.element.setAttribute('aria-busy', 'false');

                if (response.content) {
                    this.setElementHTML(this.element, response.content);
                }

                // scroll navs
                for (const nav of this.element.querySelectorAll('[data-component~="wcsf-nav"]')) {
                    const navList = nav.querySelector('[data-component~="wcsf-nav-list"]');

                    if (navList) {
                        navList.scrollLeft = nav.querySelector('.active').offsetLeft;
                    }
                }

                // scroll window
                if (options.scrollingTopOnUpdate) {
                    const element = this.element
                        .querySelector(options.scrollingToElement || '[data-component~="wcsf-step"].is-active');

                    if (element && !this.isScrolledIntoView(element)) {
                        this.scrollToElement(element, Number(options.scrollingGap), options.scrollingBehavior);
                    }
                }

                // preliminary results value
                this.controlsPreliminaryResults = this.element.querySelector('[data-component~="wcsf-controls-preliminary-results"]'); // eslint-disable-line

                if (options.updateQueryArgs && options.queryArgs) {
                    if (this.options && this.options.resetFilterOnShowing) {
                        options.queryArgs['wcsf-submit'] = this.options.id;
                    }

                    if (response.hasOwnProperty('stateHash')) {
                        options.queryArgs['wcsf-state-hash'] = response.stateHash;
                    }

                    if (response.hasOwnProperty('stepId')) {
                        options.queryArgs['wcsf-step'] = response.stepId;
                    }

                    if (Object.keys(options.queryArgs)
                        .filter((item) => options.queryArgs[item] !== false).length !== 0
                    ) {
                        this.setQueryArg(options.queryArgs);
                    }
                }

                this.triggerEvent('ajaxCompleted.wcsf', {instance, response, formData, options});

                return response;
            })
            .catch((error) => {
                this.triggerEvent('ajaxError.wcsf', {instance, error, formData, options});
                this.element.classList.remove('is-loading');
                this.element.setAttribute('aria-busy', 'false');

                if (options.errorsLogging) {
                    this.options.windowNode.console.error(error);
                }

                if (options.errorsAlerting) {
                    this.options.windowNode.alert(`Unexpected error occurred: ${error}`);
                }

                return error;
            });
    }

    /**
     * Dispatch an event
     * @param {String} name - event name
     * @param {Object} options - object of arguments
     * @returns {this} self instance
     */
    triggerEvent(name, options = {}) {
        this.element.dispatchEvent(new CustomEvent(name, {bubbles: true, detail: options}));

        return this;
    }

    /**
     * Get a step by the previous state
     * @param {PopStateEvent} event - window history pop event
     * @returns {Promise} ajax request
     */
    popState(event) {
        const requestArgs = {action: this.options.ajaxActions.getStep};
        const openingPath = event && event.state && event.state.path;

        if (!openingPath) {
            return $.when();
        }

        const queryArgs = new URL(openingPath).searchParams;

        if (queryArgs.has('wcsf-step')) {
            requestArgs.stepId = queryArgs.get('wcsf-step');
        }

        return this.ajaxRequest(requestArgs, {updateQueryArgs: false});
    }

    /**
     * Get current filter form data
     * @param {Object} data - extra data to pass
     * @returns {FormData} value
     */
    getFormData(data = {}) {
        const formData = new FormData(this.element.querySelector('[data-component~="wcsf-form"]'));

        for (let key in data) {
            if (data.hasOwnProperty(key) && data[key] !== null) {
                formData.append(key, typeof data[key] !== 'string' ? JSON.stringify(data[key]) : data[key]);
            }
        }

        return formData;
    }
    // </editor-fold>

    // <editor-fold desc="Nav">
    /**
     * Submit step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    submit(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.submit}, args);

        return this.ajaxRequest(this.getFormData(args));
    }

    /**
     * Submit step and go next
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    submitAndGoNext(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.submitAndGoNext}, args);

        return this.ajaxRequest(this.getFormData(args));
    }

    /**
     * Submit step and go to results
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    submitAndGoResults(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.submitAndGoResults}, args);

        return this.ajaxRequest(this.getFormData(args));
    }

    /**
     * Submit step and get another one by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    submitAndGoTo(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.submitAndGoTo}, args);

        return this.ajaxRequest(this.getFormData(args));
    }

    /**
     * Get step by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    getStep(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.getStep}, args);

        return this.ajaxRequest(this.getFormData(args));
    }

    /**
     * Reset form to the first step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    reset(args = {}) {
        args = this.extendObject({action: this.options.ajaxActions.reset}, args);

        return this.ajaxRequest(this.getFormData(args));
    }
    // </editor-fold>

    /**
     * Make a preliminary results query
     * @param {HTMLElement} element - HTML element to work with
     * @returns {this} self instance
     */
    getPreliminaryResultsCount(element) {
        const formData = this.getFormData({action: this.options.ajaxActions.getPreliminaryResultsCount});

        clearTimeout(this.getPreliminaryResultsCountTimeout);

        if (this.getPreliminaryResultsCountPromise && this.getPreliminaryResultsCountPromise.abort) {
            this.getPreliminaryResultsCountPromise.abort();
        }

        // save extra parameters
        for (let key in this.customOptions) {
            if (this.customOptions.hasOwnProperty(key)) {
                formData.append(
                    key,
                    typeof this.customOptions[key] !== 'string'
                        ? JSON.stringify(this.customOptions[key])
                        : this.customOptions[key]
                );
            }
        }

        // pass cart to the server
        this.getPreliminaryResultsCountTimeout = setTimeout(
            () => {
                this.getPreliminaryResultsCountPromise = this.options.windowNode
                    .fetch(
                        this.options.ajaxUrl,
                        {
                            method: 'post',
                            cache: 'default',
                            body: formData
                        }
                    )
                    .then((response) => response.json())
                    .then((response) => {
                        this.triggerEvent('getPreliminaryResultsCount.wcsf', {instance: this, response, element});

                        if (response.controls && this.controlsPreliminaryResults) {
                            this.setElementHTML(this.controlsPreliminaryResults, response.controls);
                        }
                    })
                    .catch((error) => {
                        if (status !== 'abort') {
                            this.options.windowNode.console.error(`Unexpected error occurred: ${error}`);
                        }
                    });
            },
            this.options.preliminaryResultsCountRequestInterval
        );

        return this;
    }

    // <editor-fold desc="Utils">
    /**
     * Check HTML form validity
     * @param {HTMLFormElement} form - jQuery object
     * @returns {Boolean} is valid
     */
    checkFromValidity(form = null) {
        let isValid = true;

        if (!form) {
            form = this.options.documentNode.querySelector('[data-component~="wcsf-form"]');

            if (typeof form.checkValidity !== 'undefined') {
                return form.checkValidity();
            }
        }

        // fallback
        for (let element of form.querySelectorAll('input, select, textarea')) {
            if (element.required && element.value === '') {
                isValid = false;
            }
        }

        return isValid;
    }

    /**
     * Get current URL search params
     * @param {String} search - GET string to parse
     * @returns {object} URLSearchParams
     */
    getQueryArgs(search = this.options.windowNode.location.search) {
        if (typeof URLSearchParams === 'undefined') {
            return {};
        }

        return new URLSearchParams(search);
    }

    /**
     * Set URL request parameter value
     * @param {Object} args - key pair of params
     * @returns {this} self instance
     */
    setQueryArg(args) {
        if (!this.options.windowNode.history || !this.options.windowNode.history.pushState) {
            return this;
        }

        const queryArgs = this.getQueryArgs();

        if (!queryArgs.get) {
            return this;
        }

        for (let key in args) {
            if (args.hasOwnProperty(key)) {
                if (typeof args[key] === 'boolean' && !args[key]) {
                    queryArgs.delete(key);
                } else {
                    queryArgs.set(key, args[key]);
                }
            }
        }

        const path = this.options.windowNode.location.protocol + '//' + this.options.windowNode.location.host
            + this.options.windowNode.location.pathname + '?' + queryArgs.toString();

        this.options.windowNode.history.pushState({path}, '', path);

        return this;
    }

    /**
     * Send vibration signal
     * @param {Array} args - vibration pattern as duration, pause, duration,..
     * @returns {this} self instance
     */
    vibrate(args = [200]) {
        if ('vibrate' in this.options.windowNode.navigator) {
            this.options.windowNode.navigator.vibrate(args);
        } else if ('oVibrate' in this.options.windowNode.navigator) {
            this.options.windowNode.navigator.oVibrate(args);
        } else if ('mozVibrate' in this.options.windowNode.navigator) {
            this.options.windowNode.navigator.mozVibrate(args);
        } else if ('webkitVibrate' in this.options.windowNode.navigator) {
            this.options.windowNode.navigator.webkitVibrate(args);
        }

        return this;
    }

    /**
     * Is element on the screen
     * @param {HTMLElement} element - element to check
     * @param {Boolean} strict - check element bottom position also
     * @returns {Boolean} function result
     */
    isScrolledIntoView(element, strict = false) {
        const rect = element.getBoundingClientRect();

        return !strict && rect.top >= 0 && rect.top <= this.options.windowNode.innerHeight
            || strict && rect.top >= 0 && rect.bottom <= this.options.windowNode.innerHeight;
    }

    /**
     * Scroll window screen to element
     * @param {HTMLElement} element - scroll to element
     * @param {Number} gap - top space gap
     * @param {String} behavior - animation behavior
     * @returns {this} self instance
     */
    scrollToElement(element, gap = 0, behavior = 'smooth') {
        const reduceMotion = this.options.windowNode.matchMedia('(prefers-reduced-motion: reduce)') === true
            || this.options.windowNode.matchMedia('(prefers-reduced-motion: reduce)').matches === true;

        this.options.windowNode.scrollTo({
            top: element.getBoundingClientRect().top - Number(gap),
            behavior: reduceMotion ? 'instant' : behavior
        });

        return this;
    }

    /**
     * Extend object properties by other objects
     * @param {Object} args - object to extend
     * @returns {Object} new extended object
     */
    extendObject(...args) {
        return Object.assign({}, ...args);
    };

    /**
     * Set element HTML content
     * @param {Element} element - html object to work with
     * @param {String} html - html to insert
     * @returns {Object} new extended object
     */
    setElementHTML(element, html) {
        if (typeof element.setHTML !== 'undefined') {
            element.setHTML(html);
        } else {
            element.innerHTML = html;
        }
    };
}

export default WCSF;
